(**
  Copyright The MITRE Corporation 2005.  All rights reserved.
*)

(** The Feature Manager module simply associates a feature set ID with a set of feature functions *)

open TextFeatures
open FeatureGen

type t = (string * string) list


let of_string s = [(s,s)]
let of_assoc_list s = s

let r1 = Str.regexp "[ \n\t]+"

let rec prefixes_of c l =
  match l with
      [] -> c
    | a :: b -> prefixes_of ((a :: (List.hd c)) :: c) b


let affix_of dseq p fn inc =
  let np = ref p in
  let len = Array.length dseq in
  let affix = ref [] in
  let str_of pos = match FeatureGen.Source.get_obs dseq.(pos) with Some s -> s | None -> " " in
  let cond_fn pos = 
    let s = str_of pos in
      not (Str.string_match r1 s 0) in
    while !np >= 0 && !np < len && (cond_fn !np) do
      affix := (str_of !np) :: !affix;
      np := fn !np;
    done;
    if inc then List.rev !affix,(List.hd !affix) else !affix,(List.hd !affix)


(* all character-based feature extraction sets *)
let korean_ne_fns lex =
  let main_fns = 
    [wd_f;
     edge_f;
     node_f;
     window_fns [wd_f] [-4;-3;-2;-1;1;2;3;4]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns

let korean_ne_fns_2 lex =
  let main_fns =
  [wd_f;
   edge_f;
   node_f;
   regexp_f_gen ":WS:" (Str.regexp "[ \n\t]+");
   regexp_f_gen "ALLCAP" (Str.regexp "^[A-Z]+$");
   regexp_f_gen "INITCAP" (Str.regexp "^[A-Z].*$");
   regexp_f_gen "NAT_NUM" (Str.regexp "^[0-9]+$");
   n_gram_fn (regexp_f_gen "NAT_NUM" (Str.regexp "[0-9]+$")) [-1;0];
   window_fns [(regexp_f_gen ":WS:" (Str.regexp "[ \n\t]+"))] [-1;1];
   n_gram_fn wd_f [-1;0];
   n_gram_fn wd_f [-2;-1;0];
   n_gram_fn wd_f [-3;-2;-1;0];
   window_fns [wd_f] [-4;-3;-2;-1;1;2;3;4]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns

let korean_ne_fns_3 lex =
  [wd_f;
   edge_f;
   node_f;
   regexp_f_gen "INITCAP" (Str.regexp "[A-Z].*$");
   regexp_f_gen "NAT_NUM" (Str.regexp "[0-9]+$");
   regexp_f_gen "INITCAP_ALPHA" (Str.regexp "[A-Z][a-z]*$");
   regexp_f_gen "ALPHA_NUM" (Str.regexp "[0-9A-Za-z]+$");
   n_gram_fn (regexp_f_gen "NAT_NUM" (Str.regexp "[0-9]+$")) [-1;0];
   n_gram_fn (regexp_f_gen "ALPHA_NUM" (Str.regexp "[0-9A-Za-z]+$")) [-1;0];
   n_gram_fn wd_f [-1;0];
   n_gram_fn wd_f ~edge:true [-1;0];
   n_gram_fn wd_f [-2;-1;0];
   n_gram_fn wd_f ~edge:true [-2;-1;0];
   n_gram_fn wd_f [-3;-2;-1;0];
   n_gram_fn wd_f [-1;0;1];
   n_gram_fn wd_f ~edge:true [-1;0;1];
   n_gram_fn wd_f [-2;-1;0;1];
   n_gram_fn wd_f [-2;-1;0;1;2];
   n_gram_fn wd_f [-3;-2;-1;0;1;2];
   n_gram_fn wd_f [-4;-3;-2;-1;0];
   window_fns [wd_f] [-4;-3;-2;-1;1;2;3;4]]

let korean_ne_fns_4 lex =
  let main_fns = 
    [
     edge_f;
     node_f;
     (* char_level_affixes; *)
     n_gram_fn wd_f [-1;0];
     n_gram_fn wd_f [-2;-1;0];
     n_gram_fn wd_f ~edge:true [-1;0];
     n_gram_fn wd_f ~edge:true [-2;-1;0];
     window_fns [wd_f] [-4;-3;-2;-1;0;1;2;3;4]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns


let korean_ne_lex_fns lex =
  let main_fns = 
    [wd_f;
     edge_f;
     n_gram_fn wd_f [-1;0];
     n_gram_fn wd_f [-2;-1;0];
     n_gram_fn wd_f [-1;0;1];
     node_f;
     window_fns [wd_f] [-2;-1;1;2]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns


let get_feature_fns feature_set data =
  match feature_set with
    | "KOREAN-NE" -> korean_ne_fns data
    | "KOREAN-NE-2" -> korean_ne_fns_2 data
    | "KOREAN-NE-3" -> korean_ne_fns_3 data
    | "KOREAN-NE-4" -> korean_ne_fns_4 data
    | "KOREAN-NE-LEX-1" -> korean_ne_lex_fns data
    | _ -> raise Not_found

