(**
  Copyright The MITRE Corporation 2005.  All rights reserved.
*)

(** The Feature Manager module simply associates a feature set ID with a set of feature functions *)

open TextFeatures
open FeatureGen
open Camomile

type t = (string * string) list


let of_string s = [(s,s)]
let of_assoc_list s = s

let r1 = Str.regexp "[ \n\t]+"

let rec prefixes_of c l =
  match l with
      [] -> c
    | a :: b -> prefixes_of ((a :: (List.hd c)) :: c) b

(* gets characters as well as affixes *)
let utf8_chars_contained dseq p =
  let char_ids = ref [] in
    begin
    match FeatureGen.Source.get_obs dseq.(p) with
	Some t ->
	  UTF8.iter 
	  (fun c -> 
	     char_ids := ("C:"^(string_of_int (UChar.code c))) :: !char_ids) t
      | None -> () end;
    let ordered: string list = List.rev !char_ids in
    let prefs: string list list = prefixes_of [[(List.hd ordered)]] (List.tl ordered) in
    let sufs: string list list = prefixes_of [[(List.hd !char_ids)]] (List.tl !char_ids) in
    let prefixes: string list = List.map (fun e -> String.concat "&&" e) prefs in
    let suffixes: string list = List.map (fun e -> String.concat "&&" e) sufs in
    let all_features = List.map (fun e -> (e,1.0))
			 ((("numChars:"^(string_of_int (List.length ordered))) :: ordered) @ prefixes @ suffixes) in
      FeatureGen.set_fs (all_features, false)

let affix_of dseq p fn inc =
  let np = ref p in
  let len = Array.length dseq in
  let affix = ref [] in
  let str_of pos = match FeatureGen.Source.get_obs dseq.(pos) with Some s -> s | None -> " " in
  let cond_fn pos = 
    let s = str_of pos in
      not (Str.string_match r1 s 0) in
    while !np >= 0 && !np < len && (cond_fn !np) do
      affix := (str_of !np) :: !affix;
      np := fn !np;
    done;
    if inc then List.rev !affix,(List.hd !affix) else !affix,(List.hd !affix)

(* crafty feature: find previous remaining characters within the current "word" *)
(* also get the first char and the last char from the current "word" *)
let char_level_affixes ?(edge=false) dseq p =
  let prev_l,hprev = try (affix_of dseq p (fun x -> x - 1) false) with Failure "hd" -> [],""
  and post_l,hpost = try (affix_of dseq p (fun x -> x + 1) true) with Failure "hd" -> [],"" in
  let prev = String.concat "" prev_l in
  let post = String.concat "" post_l in
    if (not (prev = "")) && (not (post = "")) then
      FeatureGen.set_fs ([(prev,1.0);(("First:"^hprev),1.0);(post,1.0);(("Last:"^hpost),1.0)], edge)
    else if (not (prev = "")) then
      FeatureGen.set_fs ([(prev,1.0);(("First:"^hprev),1.0)], edge)
    else if (not (post = "")) then
      FeatureGen.set_fs ([(post,1.0);(("Last:"^hpost),1.0)], edge)
    else [],edge

(* assume tokens are single chars for this feature function *)
let utf8_char_type dseq p = 
  let ctype = 
    match FeatureGen.Source.get_obs dseq.(p) with
	Some t ->
	  begin
	    match (UCharInfo.script (UTF8.get t 0)) with
		`Hangul -> "Script:Hangul"
	      | `Han -> "Script:Han"
	      | `Common -> "Script:Common"
	      | _ -> "Script:unk" end;
      | None -> "Script:unk" in
  let c_class = 
    match FeatureGen.Source.get_obs dseq.(p) with
	Some t ->
	  begin
	  match (UCharInfo.general_category (UTF8.get t 0)) with
	      `Nd | `Nl | `No -> "T:Num"
	    | `Pc | `Pd | `Ps | `Pe | `Pi | `Pf | `Po -> "T:Punct"
	    | _ -> "" end
      | None -> "" in
    if c_class = "" then
      FeatureGen.set_fs ([(ctype,1.0)], false)
    else FeatureGen.set_fs ([(c_class,1.0);(ctype,1.0)], false)

(* all character-based feature extraction sets *)
let korean_ne_fns lex =
  let main_fns = 
    [wd_f;
     edge_f;
     node_f;
     window_fns [wd_f] [-4;-3;-2;-1;1;2;3;4]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns

let korean_ne_fns_2 lex =
  let main_fns =
  [wd_f;
   edge_f;
   node_f;
   regexp_f_gen ":WS:" (Str.regexp "[ \n\t]+");
   regexp_f_gen "ALLCAP" (Str.regexp "^[A-Z]+$");
   regexp_f_gen "INITCAP" (Str.regexp "^[A-Z].*$");
   regexp_f_gen "NAT_NUM" (Str.regexp "^[0-9]+$");
   utf8_char_type;
   n_gram_fn (regexp_f_gen "NAT_NUM" (Str.regexp "[0-9]+$")) [-1;0];
   window_fns [(regexp_f_gen ":WS:" (Str.regexp "[ \n\t]+"))] [-1;1];
   n_gram_fn wd_f [-1;0];
   n_gram_fn wd_f [-2;-1;0];
   n_gram_fn wd_f [-3;-2;-1;0];
   window_fns [wd_f] [-4;-3;-2;-1;1;2;3;4]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns

let korean_ne_fns_3 lex =
  [wd_f;
   edge_f;
   node_f;
   utf8_char_type;
   regexp_f_gen "INITCAP" (Str.regexp "[A-Z].*$");
   regexp_f_gen "NAT_NUM" (Str.regexp "[0-9]+$");
   regexp_f_gen "INITCAP_ALPHA" (Str.regexp "[A-Z][a-z]*$");
   regexp_f_gen "ALPHA_NUM" (Str.regexp "[0-9A-Za-z]+$");
   n_gram_fn (regexp_f_gen "NAT_NUM" (Str.regexp "[0-9]+$")) [-1;0];
   n_gram_fn (regexp_f_gen "ALPHA_NUM" (Str.regexp "[0-9A-Za-z]+$")) [-1;0];
   n_gram_fn utf8_char_type [-1;0];
   n_gram_fn wd_f [-1;0];
   n_gram_fn wd_f ~edge:true [-1;0];
   n_gram_fn wd_f [-2;-1;0];
   n_gram_fn wd_f ~edge:true [-2;-1;0];
   n_gram_fn wd_f [-3;-2;-1;0];
   n_gram_fn wd_f [-1;0;1];
   n_gram_fn wd_f ~edge:true [-1;0;1];
   n_gram_fn wd_f [-2;-1;0;1];
   n_gram_fn wd_f [-2;-1;0;1;2];
   n_gram_fn wd_f [-3;-2;-1;0;1;2];
   n_gram_fn wd_f [-4;-3;-2;-1;0];
   window_fns [wd_f] [-4;-3;-2;-1;1;2;3;4]]

let korean_ne_fns_4 lex =
  let main_fns = 
    [
     edge_f;
     node_f;
     (* char_level_affixes; *)
     n_gram_fn wd_f [-1;0];
     n_gram_fn wd_f [-2;-1;0];
     n_gram_fn wd_f ~edge:true [-1;0];
     n_gram_fn wd_f ~edge:true [-2;-1;0];
     window_fns [wd_f] [-4;-3;-2;-1;0;1;2;3;4]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns


let korean_ne_lex_fns lex =
  let main_fns = 
    [wd_f;
     utf8_chars_contained;
     utf8_char_type;
     edge_f;
     n_gram_fn wd_f [-1;0];
     n_gram_fn wd_f [-2;-1;0];
     n_gram_fn wd_f [-1;0;1];
     node_f;
     window_fns [wd_f] [-2;-1;1;2]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns


let get_feature_fns feature_set data =
  match feature_set with
    | "KOREAN-NE" -> korean_ne_fns data
    | "KOREAN-NE-2" -> korean_ne_fns_2 data
    | "KOREAN-NE-3" -> korean_ne_fns_3 data
    | "KOREAN-NE-4" -> korean_ne_fns_4 data
    | "KOREAN-NE-LEX-1" -> korean_ne_lex_fns data
    | _ -> raise Not_found




(* OLD Guys *)

(*
let korean_ne_fns lex =
  let main_fns = 
    [wd_f;
     edge_f;
     utf8_char_type;
     window_fns [wd_f] [-3;-2;-1;1;2;3];
     window_fns [utf8_char_type] [-3;-2;-1;1;2;3];
     n_gram_fn ~edge:false wd_f [-1;0];
     n_gram_fn ~edge:false wd_f [0;1];
     n_gram_fn ~edge:true wd_f [-1;0];
     n_gram_fn ~edge:true wd_f [0;1];
     n_gram_fn ~edge:false wd_f [-2;-1;0];
     n_gram_fn ~edge:true wd_f [-2;-1;0];
     node_f] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns

let korean_ne_fns_2 lex =
  let main_fns = 
    [wd_f;
     utf8_chars_contained;
     utf8_char_type;
     edge_f;
     n_gram_fn wd_f [-1;0];
     n_gram_fn wd_f [-2;-1;0];
     n_gram_fn wd_f [-1;0;1];
     node_f;
     window_fns [wd_f] [-2;-1;1;2]] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns

let korean_ne_fns_3 lex =
  let main_fns = 
    [wd_f;
     utf8_chars_contained;
     utf8_char_type;
     edge_f;
     n_gram_fn ~edge:true wd_f [-2;-1;0];
     n_gram_fn ~edge:false wd_f [-2;-1;0];
     n_gram_fn ~edge:true wd_f [-1;0;1];
     n_gram_fn ~edge:false wd_f [-1;0;1];
     n_gram_fn ~edge:true wd_f [0;1;2];
     n_gram_fn ~edge:false wd_f [0;1;2];
     n_gram_fn ~edge:true wd_f [0;1;2];
     node_f] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns

let korean_ne_fns_4 lex =
  let main_fns = 
    [wd_f;
     edge_f;
     n_gram_fn wd_f [-1;0];
     n_gram_fn wd_f [-2;-1;0];
     node_f] in
    match lex with 
	None -> main_fns
      | Some lex -> (lexicon_membership_f lex) :: main_fns
*)
