open Bigarray

type transition_label = string option
type barr = (float, float64_elt, c_layout) Array1.t
type 'a t = {u_crf: 'a Crf.user_crf_model;
	     crf: Crf.model;
	     tr: Crf.t;
	     grad: barr;
	     feature_fns: (transition_label FeatureGen.Source.t array -> int -> 
			     (((FeatureGen.feature_name * float) list) * bool)) list}

let init u_model fset = 
  let crf = u_model.Crf.crf in
    {u_crf=u_model; crf=crf; tr=(Crf.init crf.Crf.nfs crf.Crf.nls crf.Crf.parameters);
     grad=(Array1.create float64 c_layout crf.Crf.nfs);
     feature_fns = DistMgr.get_feature_fns fset None}

let collect_raw_seq ?(fixed=false) d ops la =
  let ff = if fixed then (fun x -> 0) else FeatureGen.Alphabet.update la fixed in
  Array.of_list
    ((FeatureGen.Source.create (ff (Some "-START-")) (Some "-START-") None) ::
       (List.rev_map
	  (fun (op, hid, obs) ->
	     let h = match hid with None -> None | Some c -> Some (Char.escaped c) in
	     let o = match obs with None -> None | Some c -> Some (Char.escaped c) in
	       FeatureGen.Source.create (ff h) o None) 
	  (("sub",Some '_', Some '_') :: (List.rev ops))))

let get_lines regexp file =
  let chan = open_in file in
  let ls = ref [] in
  let rec loop () =
    let line = Str.split regexp (input_line chan) in
      ls := line :: !ls;
      loop () in
  try
    loop ()
  with
    End_of_file -> close_in chan; !ls

let get_info file = 
  let lines = get_lines (Str.regexp "\t") file in
  let all_pairs = List.rev_map (fun li -> match li with a :: b :: _ -> a,b | _ -> failwith "Invalid line") lines in
  let seqs_info = List.rev_map (fun (a,b) -> Levenstein.string_dist_ops a b) all_pairs in
    seqs_info

let get_seq_pairs file la =
  let seqs_info = get_info file in
    List.rev_map (fun (d,ops) -> collect_raw_seq d ops la) seqs_info

let create_training_data_seqs (tl_pairs: 'a FeatureGen.Source.t array list) fset fa la fns =
  let s = Sys.time() in
  let yprev = ref (-1) in
  let insts = 
    List.rev_map
      (fun x ->
	 yprev := -1;
	 let size = Array.length x in
	 let inst_seq = Array.init size (fun _ -> FeatureGen.create_instance (-1)) in
	 let s1 = Sys.time() in
	   for i=0 to size - 1 do
	     let t1 = FeatureGen.Source.get_obs x.(i)
	     and yp = FeatureGen.Source.get_lab x.(i) in
		   FeatureGen.apply_feature_fns fset fa la fns !yprev yp inst_seq x i;
		   FeatureGen.set_inst_label inst_seq.(i) yp;
		   yprev := yp
	   done;
	   inst_seq ) tl_pairs  in
    insts

let create_training_data_seqs_fixed (tl_pairs: 'a FeatureGen.Source.t array list) fset la fns =
  let s = Sys.time() in
  let yprev = ref (-1) in
  let insts = 
    List.rev_map
      (fun x ->
	 yprev := -1;
	 let size = Array.length x in
	 let inst_seq = Array.init size (fun _ -> FeatureGen.create_instance (-1)) in
	 let s1 = Sys.time() in
	   for i=0 to size - 1 do
	     let t1 = FeatureGen.Source.get_obs x.(i)
	     and yp = FeatureGen.Source.get_lab x.(i) in
	       FeatureGen.apply_feature_fns_runtime fset fns inst_seq x i;
	       FeatureGen.set_inst_label inst_seq.(i) yp;
	       yprev := yp
	   done;
	   inst_seq ) tl_pairs  in
    insts

let train_on_seqs ?(i_file_root=None) ?(compact=false) ?(max_its=1000) ?(vp_train=false) feature_set file =
  let (fa: (int * int * FeatureGen.feature_name) FeatureGen.Alphabet.t) = FeatureGen.Alphabet.create () in
  let (la: transition_label FeatureGen.Alphabet.t) = FeatureGen.Alphabet.create () in
  let (fset: (FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t) = 
    Hashtbl.create 10 in (* hash of features *)
  let pairs = get_seq_pairs file la in
  let edit_dist_fns = DistMgr.get_feature_fns feature_set None in
  let tr_seqs = Array.of_list (create_training_data_seqs pairs fset fa la edit_dist_fns) in
    
  let la_fset_d = match i_file_root with Some f -> Some(f, fset, la) | None -> None in
    Printf.printf "\n training sequences created...\n"; flush stdout;
  let crf = 
    if vp_train then
      VotedPerceptron.vp_train ~la_fset_d:la_fset_d ~compact:compact ~max_its:max_its (FeatureGen.Alphabet.get_size fa)
	(FeatureGen.Alphabet.get_size la) tr_seqs
    else
      Crf.train_crf ~compact:compact ~max_its:max_its (FeatureGen.Alphabet.get_size fa) 
	(FeatureGen.Alphabet.get_size la) tr_seqs in
  let (umodel:transition_label Crf.user_crf_model) = (Crf.create_user_model crf fset la None) in
    umodel, fa

let eval_pair seq_crf ss1 ss2 =
  let l1,l2 = String.length ss1, String.length ss2 in
  let s1,s2 = if (l1 > l2) || (l1 = l2 && (String.compare ss1 ss2) < 0) then ss2,ss1 else ss1,ss2 in
  let d,ops = Levenstein.string_dist_ops s1 s2 in
  let raw_seq = collect_raw_seq d ops seq_crf.u_crf.Crf.crf_la in
  let tst_seqs = Array.of_list (create_training_data_seqs_fixed [raw_seq] seq_crf.u_crf.Crf.crf_fset 
				  seq_crf.u_crf.Crf.crf_la seq_crf.feature_fns) in
    Array1.fill seq_crf.grad 0.0;
    Crf.get_gradient_dense seq_crf.crf.Crf.parameters seq_crf.grad tst_seqs seq_crf.tr seq_crf.crf.Crf.nfs true

(* this uses a weak hashtbl to store values *)
let crf_cache_evalpair seq_crf ht p p' =
  try
    Hashtbl.find ht (p,p')
  with Not_found ->
    let sc = eval_pair seq_crf p p' in
      Hashtbl.add ht (p,p') sc;
      sc

let neval_lev x y =
  let d1, d2 = String.lowercase x, String.lowercase y in
  let s = Levenstein.string_dist d1 d2 in
    s /. ((float_of_int (String.length d1)) +. (float_of_int (String.length d2)))

(* this evaluation function is expensive but seems to give the best
   intuitive measure of similarity, lower values (e.g. < 0.96) indicate
   higher-likelihood of the strings matching *)
let neval crf_m ht x y =
  let s = crf_cache_evalpair crf_m ht x y in
  let n1 = crf_cache_evalpair crf_m ht x "" in
  let n2 = crf_cache_evalpair crf_m ht y "" in
    s /. ((n1 +. n2) /. 2.)

let remap_sequence inv_la seq =
  let li = Array.to_list 
	     (Array.init ((Array.length seq) - 2)
		(fun i -> 
		   match (FeatureGen.InvAlphabet.get inv_la seq.(i+1)) with
		       None -> ""
		     | Some s -> s)) in
    String.concat "" li

let score_pairs crf pairs = 
  List.rev_map
    (fun (s1,s2) ->
       eval_pair crf s1 s2) pairs

let top_n_forms crf feature_set n strs =
  let invla = FeatureGen.Alphabet.get_inverse crf.Crf.crf_la in
  let edit_dist_fns = DistMgr.get_feature_fns feature_set None in
    List.rev_map
      (fun str ->
	 let d,ops = Levenstein.string_dist_ops "" str in
	 let rs = collect_raw_seq ~fixed:true d ops crf.Crf.crf_la in
	 let rcrf = crf.Crf.crf in
	 let tst_seqs = Array.of_list (create_training_data_seqs_fixed [rs] crf.Crf.crf_fset 
					 crf.Crf.crf_la edit_dist_fns) in
	 let topn_seqs,scores = Viterbi.get_top_n_label_seqs n tst_seqs.(0) rcrf in
	   (str, (Array.map (remap_sequence invla) topn_seqs))) strs
    
