(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** An implementation of Conditional Random Fields *)

open Bigarray
open Vectors

module MLO = Lbfgs.Opt

exception Return 


type barr = (float, float64_elt, c_layout) Array1.t
type arr = float array
type mat2 = float array array 

type model = { parameters: barr;
	       nfs: int;
	       nls: int }

(* parameter is over the type for each label (maybe string or other structure) *)
type 'a user_crf_model = {crf: model;
			  crf_fset: (FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t;
			  crf_la: 'a FeatureGen.Alphabet.t;
			  fset_mask: ((FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t * 
				      (FeatureGen.feature_name, FeatureGen.feature_name) Hashtbl.t) option}

type t = {alpha: arr;
	  new_alpha: arr;
	  expF:    barr;
	  mi:   mat2;
	  ri:    arr;
	  tmp:   arr;
	  lambda:  barr;
	  num_F:    int;
	  num:    int }

let debug = 1
let sigma = 5.0
let invSigSquare = ref 0.01

let create_model params nfs nls = {parameters=params;nfs=nfs;nls=nls}

let create_user_model model fset la fmask =
  {crf=model; crf_fset=fset; crf_la=la; fset_mask=fmask}

let marshal_crf file user_crf = 
  let oc = open_out_bin file in
  Marshal.to_channel oc user_crf [];
  close_out oc


let print_a1 a =
  print_string "\n";
  for i=0 to (Array1.dim a) - 1 do
    Printf.printf " %f" a.{i} done

let make_a1 s : barr = 
  let a1 = Array1.create float64 c_layout s in
  Array1.fill a1 0.;
  a1

let set_matrix mat d1 d2 v =
  for i = 0 to d1 - 1 do
    Array.fill mat.(i) 0 d2 v
  done


let computeLogMi (lambdas:barr) nfs nls inst_features size (mi:mat2) (ri:arr) takeExp =
  Array.fill ri 0 nls 0.;
  set_matrix mi nls nls 0.;
  for k=0 to size - 1 do
    let (yprev,yp,f,vl) = inst_features.(k) in
      if yprev < 0 then 
	ri.(yp) <- ri.(yp) +. lambdas.{f} *. vl
      else
	mi.(yprev).(yp) <- mi.(yprev).(yp) +. lambdas.{f} *. vl
  done;
  if takeExp then
    for r = 0 to nls - 1 do
      ri.(r) <- exp ri.(r);
      for c = 0 to nls - 1 do
	mi.(r).(c) <- exp mi.(r).(c)
      done
    done

(* global reusable references *)
let (beta:mat2 ref) = ref (Array.make_matrix 1 1 0.)
let (scale:arr ref) = ref (Array.make 1 0.)


let init nfs nls l =
  beta := Array.make_matrix 1 1 0.;
  scale := Array.make 1 0.;
  {lambda=l; num=nls; num_F=nfs; alpha = Array.make nls 0.; new_alpha = Array.make nls 0.;
   expF = make_a1 (Array1.dim l); 
   mi = Array.make_matrix nls nls 0.;
   ri = Array.make nls 0.; tmp = Array.make nls 0.}

let assign1 (v1:arr) (v2:arr) len f =
  for i = 0 to len - 1 do
    v1.(i) <- f v1.(i) v2.(i) done
    
let assign (v1:arr) len f =
  for i = 0 to len - 1 do
    v1.(i) <- f v1.(i) done
    
(* simply A * x - mat * vec = r_vec *)
let matrix_mult (mat:mat2) (vec:arr) (r_vec:arr) alpha beta trans_p =
  let r = ref 0. in
  let l1 = Array.length r_vec in
  let l2 = Array.length vec in
    if trans_p then
      for i=0 to l1 - 1 do
	r := 0.;
	for j=0 to l2 - 1 do
	  r := !r +. (vec.(j) *. mat.(j).(i)) done;
	r_vec.(i) <- (r_vec.(i) *. beta) +. (!r *. alpha) done
    else
      for i=0 to l1 - 1 do
	r := 0.;
	for j=0 to l2 - 1 do
	  r := !r +. (vec.(j) *. mat.(i).(j)) done;
	r_vec.(i) <- (r_vec.(i) *. beta) +. (!r *. alpha) done
	
let vec_sum (vec:arr) len =
  let sum = ref 0. in
    for i = 0 to len - 1 do
      sum := !sum +. vec.(i) done;
    !sum
      
let gg_1 len (alpha:arr) (beta: mat2 ref) (scale: arr ref) num =
  Array.fill alpha 0 num 1.;
  (* Array1.fill expF 0.; *)
  if (Array.length !beta) < len || (Array.length !scale) < len then (
    beta := Array.make_matrix (2 * len) num 0.;
    scale := Array.make (2 * len) 0. )
    
let backward_pass doScaling len data_seq (lambdas:barr) (mi:mat2) (ri:arr) (tmp:arr) (beta:mat2 ref) 
  (scale:arr ref) (num_F:int) (num: int) =

  !scale.(len - 1) <- if doScaling then (float_of_int num) else 1.;
  assign (!beta).(len-1) num (fun x -> (1. /. !scale.(len - 1)));
  for i=(len - 1) downto 1 do
    let (inst_features, size) = FeatureGen.get_compiled_vec data_seq.(i) in
      computeLogMi lambdas num_F num inst_features size mi ri true;
      (* we'll do a copy here now, but may not be necessary *)
      Array.blit (!beta).(i) 0 tmp 0 num;
      assign1 tmp ri num (fun x y -> x *. y);
      let sl_l = (!beta).(i-1) in
	matrix_mult mi tmp sl_l 1. 0. false;
	!scale.(i-1) <- if doScaling then vec_sum sl_l num else 1.;
	assign sl_l num (fun x -> x *. (1.0 /. !scale.(i-1)));
  done

let forward_pass_dense len (lambdas:barr) (gradient:barr) (expF:barr) num_F num data_seq (mi:mat2) (ri:arr) 
  (alpha:arr) (tmp:arr) (beta:mat2 ref) (new_alpha:arr) (scale:arr ref) =

  let thisSeqLogLi = ref 0. in
    for i=0 to len - 1 do
      let (inst_features,size) = FeatureGen.get_compiled_vec data_seq.(i) in
	computeLogMi lambdas num_F num inst_features size mi ri true;
	Array.blit alpha 0 tmp 0 num;
	matrix_mult mi tmp new_alpha 1. 0. true;
	assign1 new_alpha ri num (fun x y -> x *. y);
	let label = FeatureGen.get_inst_label data_seq.(i) in
	let sl_l = (!beta).(i) in
	  for k=0 to size - 1 do
	    let (yprev,yp,f,vl) = inst_features.(k) in
	      if (label = yp && (((i > 0) && ((FeatureGen.get_inst_label data_seq.(i-1)) = yprev)) || yprev < 0)) then (
		gradient.{f} <- gradient.{f} +. vl;
		thisSeqLogLi := !thisSeqLogLi +. vl *. lambdas.{f});
	      if yprev < 0 then
		expF.{f} <- expF.{f} +. new_alpha.(yp) *. vl *. sl_l.(yp)
	      else 
		expF.{f} <- expF.{f} +. alpha.(yprev) *. vl *. ri.(yp) *.
		  mi.(yprev).(yp) *. sl_l.(yp)
	  done;
	  Array.blit new_alpha 0 alpha 0 num;
	  assign alpha num (fun x -> x *. (1.0 /. !scale.(i)));
    done;
    !thisSeqLogLi


let forward_pass_sparse len (lambdas:barr) (gradient:barr) (exp_table:(int,float) Hashtbl.t) num_F num data_seq (mi:mat2) (ri:arr) 
  (alpha:arr) (tmp:arr) (beta:mat2 ref) (new_alpha:arr) (scale:arr ref) =

  let thisSeqLogLi = ref 0. in
    for i=0 to len - 1 do
      let (inst_features,size) = FeatureGen.get_compiled_vec data_seq.(i) in
	computeLogMi lambdas num_F num inst_features size mi ri true;
	Array.blit alpha 0 tmp 0 num;
	matrix_mult mi tmp new_alpha 1. 0. true;
	assign1 new_alpha ri num (fun x y -> x *. y);
	let label = FeatureGen.get_inst_label data_seq.(i) in
	let sl_l = (!beta).(i) in
	  for k=0 to size - 1 do
	    let (yprev,yp,f,vl) = inst_features.(k) in
	      if (label = yp && (((i > 0) && 
				  ((FeatureGen.get_inst_label data_seq.(i-1)) = yprev)) || yprev < 0)) then (
		gradient.{f} <- gradient.{f} +. vl;
		thisSeqLogLi := !thisSeqLogLi +. vl *. lambdas.{f});
	      if yprev < 0 then
		let d1 = new_alpha.(yp) *. vl *. sl_l.(yp) in
		  try
		    let cur = Hashtbl.find exp_table f in
		      (* might be fast to just add *)
		      Hashtbl.replace exp_table f (cur +. d1)
		  with Not_found -> Hashtbl.add exp_table f d1
	      else 
		let d1 = alpha.(yprev) *. ri.(yp) *. 
			   mi.(yprev).(yp) *. vl *. sl_l.(yp) in
		  try
		    let cur = Hashtbl.find exp_table f in
		      Hashtbl.replace exp_table f (cur +. d1)
		  with Not_found -> Hashtbl.add exp_table f d1
	  done;
	  Array.blit new_alpha 0 alpha 0 num;
	  assign alpha num (fun x -> x *. (1.0 /. !scale.(i)));
    done;
    !thisSeqLogLi

let get_gradient_sparse (lambdas:barr) (gradient:barr) (data_seqs:FeatureGen.train_data) tr nfs doScaling =
  let logli = ref 0. in
  let exp_table = Hashtbl.create 10 in
    for i=0 to (Array1.dim lambdas) - 1 do
      let li = lambdas.{i} in
	gradient.{i} <- (-1.0) *. li *. !invSigSquare;
	logli := !logli -. (li *. li *. !invSigSquare) /. 2.
    done;
    for j=0 to (Array.length data_seqs) - 1 do
      Hashtbl.clear exp_table;
      let data_seq = data_seqs.(j) in
      let len = Array.length data_seq in
	gg_1 len tr.alpha beta scale tr.num;
	backward_pass doScaling len data_seq lambdas tr.mi tr.ri tr.tmp beta scale tr.num_F tr.num;
	let thisSeqLogLi = ref (forward_pass_sparse len lambdas gradient exp_table 
				  tr.num tr.num data_seq tr.mi tr.ri 
				  tr.alpha tr.tmp beta tr.new_alpha scale) in
	let zx = vec_sum tr.alpha tr.num in
	  thisSeqLogLi := !thisSeqLogLi -. log zx;
	  for i=0 to len - 1 do
	    thisSeqLogLi := !thisSeqLogLi -. log !scale.(i)
	  done;
	  logli := !logli +. !thisSeqLogLi;
	  Hashtbl.iter
	    (fun f exp_f ->
	       gradient.{f} <- gradient.{f} -. (exp_f /. zx)) exp_table;
    done;
    for f = 0 to nfs - 1 do
      gradient.{f} <- gradient.{f} *. (-1.0);
    done;  (* flip sign here *)
    -.(!logli)  (* return negative log-likelihood since we want to minimize this *)

let get_gradient_dense (lambdas:barr) (gradient:barr) (data_seqs:FeatureGen.train_data) tr nfs doScaling =
  let logli = ref 0. in
    for i=0 to (Array1.dim lambdas) - 1 do
      let li = lambdas.{i} in
	gradient.{i} <- (-1.0) *. li *. !invSigSquare;
	logli := !logli -. (li *. li *. !invSigSquare) /. 2.
    done;
    for j=0 to (Array.length data_seqs) - 1 do
      let data_seq = data_seqs.(j) in
      let len = Array.length data_seq in
	Array1.fill tr.expF 0.;
	gg_1 len tr.alpha beta scale tr.num;
	backward_pass doScaling len data_seq lambdas tr.mi tr.ri tr.tmp beta scale tr.num_F tr.num;
	let thisSeqLogLi = ref (forward_pass_dense len lambdas gradient tr.expF
				  tr.num tr.num data_seq tr.mi tr.ri 
				  tr.alpha tr.tmp beta tr.new_alpha scale) in
	let zx = vec_sum tr.alpha tr.num in
	  thisSeqLogLi := !thisSeqLogLi -. log zx;
	  for i=0 to len - 1 do
	    thisSeqLogLi := !thisSeqLogLi -. log !scale.(i)
	  done;
	  logli := !logli +. !thisSeqLogLi;
	  for f = 0 to nfs - 1 do
	    gradient.{f} <- gradient.{f} -. (tr.expF.{f} /. zx);
	  done;
    done;
    for f = 0 to nfs - 1 do
      gradient.{f} <- gradient.{f} *. (-1.0);
    done;  (* flip sign here *)
    -.(!logli)  (* return negative log-likelihood since we want to minimize this *)

let print_sequence tr_data =
  Array.iter
    (fun x ->
       Printf.printf "\n--------";
       Array.iter
	 (fun y ->
	    Printf.printf " %d" (FeatureGen.get_inst_label y)) x ) tr_data

let train_crf ?(verbose=true) ?(intermediate=50) ?(la_fset_d=None) ?(i_params=None) ?(compact=true) ?(max_its=1000) num_features num_labels tr_data =
  let iprint = Array.make 2 0
  and params = match i_params with None -> Array1.create float64 c_layout num_features | Some p -> p
  and gradient = Array1.create float64 c_layout num_features
  and diagco = ref false
  and diag = Array1.create float64 c_layout num_features
  and eps = 1.0e-2
  and xtol = 4.9e-300
  and icall = ref 0
  and m = 7
  and iflag = ref 0 in
  let max_iters = max_its in
  let iter = ref 0 in
  let n = num_features in
  let float_n = float_of_int n in
  let num_seqs = float_of_int (Array.length tr_data) in
  let f = ref 0. in
  let grad_fn = if (float_n *. num_seqs) > 1e+9 then get_gradient_sparse else get_gradient_dense in
    if not compact then (
      Gc.compact(); (* do a major GC before compiling data *)
      FeatureGen.compile_data tr_data);
    begin match i_params with None -> Array1.fill params 0. | _ -> () end;
    Array1.fill gradient 0.;
    Array1.fill diag 0.;
    iprint.(0) <- (if verbose then 1 else 0);
    iprint.(1) <- 0;
    Printf.printf "\n initializing CRF\n";
    let tr = init num_features num_labels params in
      Printf.printf "\n Number of features: %d" num_features;
      f := grad_fn params gradient tr_data tr num_features true;
      Printf.printf "\n Initial f-val: %f" !f; flush stdout;
      MLO.Lbfgs.init (); (* initialize lbfgs routine *)
      if !f > 0.0 then (
	try
	  while (!iter < max_iters) do
	    if !iter > 0 && (!iter mod intermediate) = 0 then (
	      match la_fset_d with
		  Some (f, fset, la) ->
		    Printf.printf "Writing out intermediate model..%d\n" !iter;
		    marshal_crf (f^"_at_"^(string_of_int !iter)) 
		    (create_user_model (create_model params num_features num_labels) fset la None)
		| None -> ());
            MLO.Lbfgs.lbfgs n m params f gradient diagco diag iprint eps xtol iflag;
	    if (!iflag <= 0) then
	      raise Return;
	    f := grad_fn params gradient tr_data tr num_features true;
	    iter := !iter + 1
	  done;
	  Printf.printf "\n Final f: %f\n\n" !f;
	  create_model params num_features num_labels
	with Return -> 
	  Printf.printf "\n Final f -> %f\n\n" !f;
	  create_model params num_features num_labels)
      else create_model params num_features num_labels

		     
	   
