(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(**
  Implementation of CRF {e training}.  Decoding is implemented in the module {! Viterbi}.
  This implementation follows the algorithm outline and notation Sha and Pereira 2003.
  Additionally, parts of this code draw inspiration (particularly the inner loops within the training 
  procedure) from the {e CRF Project} implementation of CRFs by Sunita Sarawagi (this is the CRF engine
  within MinorThird).

  @author Ben Wellner

  @see "Sha and Pereira" Fei Sha, Fernando C.N. Pereira. Shallow parsing with conditional random fields. HLT-NAACL 2003.
  @see <http://crf.sourceforge.net/> CRF project.
  @see <http://mallet.cs.umass.edu/index.php/Main_Page> MALLET
  @see <http://minorthird.sourceforge.net> MinorThird 

*)


open Bigarray
open Vectors

(** {6 Types} *)

(** 1-D Bigarray type with double floats and C layout *)
type barr = (float, float64_elt, c_layout) Array1.t
    
(** 1-D Array of floats *)
type arr = float array

(** Matrix of floats using Arrays *)
type mat2 = float array array 

(** Raw CRF model including parameters, number of features and number of labels *)
type model = { parameters: barr;
	       nfs: int;
	       nls: int }

(** User CRF model. *)
type 'a user_crf_model = 
    {
      crf: model; (** Raw model *)
      crf_fset: (FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t; (** Feature set.  A hash from feature 
										   names to feature types *)
      crf_la: 'a FeatureGen.Alphabet.t; (** Label alphabet *)
      fset_mask: ((FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t * 
		  (FeatureGen.feature_name, FeatureGen.feature_name) Hashtbl.t) option (** Optional feature mask
											   for feature induction *)
    }

(** Internal data record containing structures used during training. *)
type t = {alpha: arr; (** Alpha values for forward computation *)
	  new_alpha:arr; (** Updated alpha values *)
	  expF:    barr;   (** Empirical expected value of the features *)
	  mi:   mat2;   (** Transition matrix *)
	  ri:    arr;    (** Initial state probabilities *)
	  tmp:   arr;    (** Temp vector *)
	  lambda:  barr;   (** Current paramters *)
	  num_F:    int;   (** Number of features *)
	  num:    int   (** Number of labels *)
	 }

(** {6 Constants} *)

(** Debug flag *)
val debug : int

(** Guassian prior variance. *)
val sigma : float
  
(** Inverse sigma squared reference *)
val invSigSquare : float ref

(** {6 Constructors} *)

(** Create a core CRF model *)
val create_model : barr -> int -> int -> model

(** Create user CRF model *)
val create_user_model : model -> (FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t ->
  'a FeatureGen.Alphabet.t ->
  ((FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t * 
   (FeatureGen.feature_name, FeatureGen.feature_name) Hashtbl.t) option -> 'a user_crf_model

(** {6 Utility Functions} *)
  
(** Make a Bigarray. [Crf.make_a1 int] creates a [barr], of size [int] *)
val make_a1 : int -> barr

(** Initialize the CRF data structures. This must be called prior to calling [Crf.get_gradient]. *)   
val init : int -> int -> barr -> t

(** Save CRF in machine readable form to file. [Crf.marshal_crf fname crf_model] generates
    an internal, machine readable representation of [crf_model] in the file [fname]. *)
val marshal_crf : string -> 'a user_crf_model -> unit

(** {6 Main Functions} *)

(** Main training function. [Crf.train_crf ~la_fset_d:lfd ~compact:compact ~max_its:max_its fsize lsize trseqs]
  initiates log-liklihood driven training on [trseqs] where [fsize] is the number of features [lsize] is the
  number of labels, [max_its] is the maximum number of iterations (default 1000), [compact], when true specifies
  a compact (but slighty slower) representation of features.  [lfd] is an option triple where the first
  argument should be a path root for generating intermediate model files, the second argument is the feature
  set (of type [FeatureGen.fset]) and the third argument is the label alphabet.
 *)
val train_crf : 
  ?verbose:bool ->
  ?intermediate:int ->
  ?la_fset_d:(string * (FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t *
	      'a FeatureGen.Alphabet.t) option -> 
  ?i_params: barr option -> 
  ?compact:bool -> 
  ?max_its:int -> int -> int -> FeatureGen.train_data -> model

(** Function to compute the log-probabilities of a transition matrix. Generally an internal function, but
    decoding module makes use of this function as well. *)
val computeLogMi : barr -> int -> int -> FeatureGen.FeatureSet.elt array -> int -> mat2 -> arr -> bool -> unit

(** Get the current gradient and log-likelihood given the current parameters and data. *)
val get_gradient_dense : barr -> barr -> FeatureGen.train_data -> t -> int -> bool -> float

(** Get the current gradient and log-likelihood given the current parameters and data. *)
val get_gradient_sparse : barr -> barr -> FeatureGen.train_data -> t -> int -> bool -> float

val print_sequence : FeatureGen.train_data -> unit
