(**
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** Module takes care of features for CRFs.  Not really able to re-use MaxEnt approach here directly
    because features are encoded differently in sequence models *)


module Feature =
  struct
    type t = int * int * int * float
    let compare (ypp1,yp1,fi1,_) (ypp2,yp2,fi2,_) =
      let x1 = fi1 - fi2 in
      if x1 <> 0 then x1
      else 
	let x2 = yp1 - yp2 in
	if x2 <> 0 then x2
	else ypp1 - ypp2
  end

module FeatureSet = Set.Make(Feature)

type feature_name = F of string | Conj of feature_name * feature_name

module AgglFeature =
  struct
    type t = feature_name

    let compare fn1 fn2 =
      match fn1, fn2 with
	  F(f1), F(f2) ->
	    String.compare f1 f2
	| F(_), Conj(_,_) -> -1
	| Conj(_,_), F(_) -> 1
	| Conj(f1,f2), Conj(g1,g2) -> 
	    let c1 = compare f1 g1 in 
	      if c1 <> 0 then c1 else compare f2 g2

    let compose fn1 fn2 =
      if compare fn1 fn2 = 1 then Conj(fn2,fn1)
      else Conj(fn1,fn2)
  end

module RankedAgglFeature = 
  struct
    type t = feature_name * float
    let compare (f1,s1) (f2,s2) = if s1 > s2 then 1 else if s2 > s1 then -1 else AgglFeature.compare f1 f2
  end

module RankedAgglFeatureSet = Set.Make(RankedAgglFeature)
module AgglFeatureSet = Set.Make(AgglFeature)


(* Note that comparison is _only_ done on fname *)
module FeatureType =
  struct
    type t = {fname: feature_name;
	      mutable fdetail: FeatureSet.t;
	      edgep: bool}
    let compare {fname=fname1} {fname=fname2} =
      let compare_x fname1 fname2 =
	match fname1, fname2 with
	    F(fn1), F(fn2) ->
	      String.compare fn1 fn2
	  | F(_), Conj(_,_) -> -1
	  | Conj(_,_), F(_) -> 1
	  | Conj(f1,f2), Conj(g1,g2) -> 
	      let c1 = compare f1 g1 in if c1 <> 0 then c1 else compare f2 g2 in
	compare_x fname1 fname2
    let edgep ft = ft.edgep
  end

module FeatureTypeSet = Set.Make(FeatureType)

(** Module encapsulating information at each sequence position *)
module Source =
  struct
    (** Main data structure for a single position in the sequence. *)
    type 'a t = {lab: int;  (** Label identifier [int] for position *)
		  obs: string option; (** (Optional) Observed string at position *)
		  info: 'a option  (** Optional data of arbitrary type.  For extracting features. *)
		 }
    let create l o i = {lab=l;obs=o;info=i}
    let get_lab s = s.lab
    let get_obs s = s.obs
    let get_info s = s.info
  end

exception IncorrectVectorType
exception Feature_failure

type f_crf_vec = {fids: int array;
		  mapping: (int,(int * int)) Hashtbl.t }

type crf_vec =
    UserVec of FeatureTypeSet.t
  | CompVec of FeatureSet.elt array * int (* The int is just the length of the array *)

type user_feature = int * int * string

type crf_instance = {mutable crf_label: int;
		     mutable crf_orig_label: int;
		     mutable crf_f_vec: crf_vec}

type data_seq = crf_instance array
type train_data = crf_instance array array


module InvAlphabet =
  struct
    type 'a t = {inv_map: (int, 'a) Hashtbl.t;
		 mutable size: int}
    let create () = {inv_map=Hashtbl.create 10; size=0}
    let get inv_alphabet id = Hashtbl.find inv_alphabet.inv_map id

    let get_inv_map inv_alphabet = inv_alphabet.inv_map
    let get_size inv_alphabet = inv_alphabet.size

    let add inv_alphabet id vl =
      Hashtbl.add inv_alphabet.inv_map id vl;
      inv_alphabet.size <- inv_alphabet.size + 1
      
  end

module Alphabet =
  struct
    type 'a t = {map: ('a,int) Hashtbl.t;
		 mutable size: int}

    let create () = {map=Hashtbl.create 10; size=0}
    let get alphabet item = Hashtbl.find alphabet.map item
    let get_map alphabet = alphabet.map
    let get_size alphabet = alphabet.size

    let update (alphabet: 'a t) (fix: bool) item =
      try 
	Hashtbl.find alphabet.map item
      with Not_found ->
	if not fix then (
	  let s = alphabet.size in
	    Hashtbl.add alphabet.map item s;
	    alphabet.size <- s + 1;
	    s )
	else (* with fixed alphabet, if not found, return -1 *)
	  -1

    let get_inverse alphabet =
      let inv = InvAlphabet.create () in
	Hashtbl.iter (fun k v -> InvAlphabet.add inv v k) (get_map alphabet);
	inv
  end

let rec unset f =
  match f with
      F(x) -> x
    | Conj(x,y) -> (unset x)^(unset y)

let set_f f = F(f)

let set_fs (fs,n) = (List.map (fun (f,v) -> (set_f f,v)) fs),n

let unset_fs (fs,n) = (List.map (fun (f,v) -> (unset f),v) fs), n

let flatten_feature_name fname = 
  let rec flatten fname acc = 
    match fname with 
	F s -> if acc = "" then s else acc^"++"^s
      | Conj(f1,f2) -> let s1 = (flatten f1 acc) in flatten f2 s1 in
    flatten fname ""

let set_inst_label crf_inst v = crf_inst.crf_label <- v
let set_inst_orig_label crf_inst v = crf_inst.crf_orig_label <- v
(* simple accessor for crf instances *)
let get_inst_label crf_inst = crf_inst.crf_label

let set_as_uvec crf_inst uv = crf_inst.crf_f_vec <- UserVec uv

let expand_vector v =
  (FeatureTypeSet.fold
     (fun {FeatureType.fdetail=d} cur_vs -> 
	List.rev_append (FeatureSet.elements d) cur_vs) v [])

let get_compiled_vec v =
  match v.crf_f_vec with
      CompVec(cv,l) -> (cv,l) 
    | UserVec uv ->
	let li = expand_vector uv in
	  (Array.of_list li, List.length li)

let get_uncompiled_vec v =
  match v.crf_f_vec with
      UserVec uv -> uv
    | CompVec(_,_) -> raise IncorrectVectorType

let create_instance v =
  {crf_label = v; crf_orig_label = v; crf_f_vec = UserVec(FeatureTypeSet.empty)}


(* add a new training feature - modify fset and feature fdetail appropriately *)
let add_feature ?(unsupported=None) instseq fset pos yprev yp name fa value =
  let feature = 
    try Hashtbl.find fset name 
    with Not_found -> 
      let n = {FeatureType.fname=name; FeatureType.fdetail=FeatureSet.empty; FeatureType.edgep=(yprev >= 0)} in
      Hashtbl.add fset name n;
      n in
    (* add un unsupported node features *)
(*
    begin
      match unsupported with 
	  None -> feature.FeatureType.fdetail <- FeatureSet.add (yprev,yp,(Alphabet.update fa false (yprev,yp,name)),value) feature.FeatureType.fdetail;
	| Some nls ->
	    for i = 0 to nls - 1 do (
	      let v = if i = yp then 1.0 else 0.0 in
		feature.FeatureType.fdetail <- FeatureSet.add (yprev,i,(Alphabet.update fa false (yprev,i,name)),v) feature.FeatureType.fdetail ) done
    end; *)
    feature.FeatureType.fdetail <- FeatureSet.add (yprev,yp,(Alphabet.update fa false (yprev,yp,name)),value) feature.FeatureType.fdetail;
    let iv = get_uncompiled_vec instseq.(pos) in
      instseq.(pos).crf_f_vec <- UserVec(FeatureTypeSet.add feature iv)

(* add a runtime feature with a given, fixed feature set - fset *)
let add_runtime_feature ?(unsupported=None) instseq fset pos name value =
  try
    let feature = Hashtbl.find fset name in 
    let iv = get_uncompiled_vec instseq.(pos) in
      instseq.(pos).crf_f_vec <- UserVec(FeatureTypeSet.add feature iv)
  with Not_found -> ()

let inst_features inst =
  let vec = get_uncompiled_vec inst in
  let li = FeatureTypeSet.fold (fun {FeatureType.fdetail=d} cur_vs -> cur_vs @ (FeatureSet.elements d)) vec [] in
    (Array.of_list li, List.length li)

(* OLD version
let compile_seq seq =
    Array.iter
	 (fun x ->
	    try
	      x.crf_f_vec <-
	      let li = expand_vector (get_uncompiled_vec x) in
		CompVec(Array.of_list li, List.length li)
	    with IncorrectVectorType -> ()) (* in this case, vector already "compiled" *)
      seq *)

let compile_seq seq =
    Array.iter
	 (fun x ->
	    try
	      x.crf_f_vec <-
	      let li = expand_vector (get_uncompiled_vec x) in
		CompVec(Array.of_list li, List.length li)
	    with IncorrectVectorType -> ()) (* in this case, vector already "compiled" *)
      seq

let compile_data inst_seqs = 
    Array.iter compile_seq inst_seqs

let apply_feature_fns fset fa la fn_list yprev yp inst_seq data_seq pos =
    List.iter
      (fun fn ->
	 try
	   let (fv_pairs,edgep) = fn data_seq pos in
	     if (not edgep) || pos > 0 then
	       List.iter
		 (fun (f,v) ->
		    add_feature inst_seq fset pos (if edgep then yprev else (-1)) yp f fa v) fv_pairs
	 with Feature_failure -> ()) fn_list

let apply_feature_fns_runtime fset fn_list inst_seq data_seq pos =
  List.iter
    (fun fn ->
      try
	let (fv_pairs,_) = fn data_seq pos in
	List.iter (fun (f,v) -> add_runtime_feature inst_seq fset pos f v) fv_pairs
      with Feature_failure -> ()) fn_list

let print_train_data tr_data = 
  Array.iter
    (fun x ->
       Printf.printf "\n -----------";
       Array.iter
	 (fun y ->
	    Printf.printf " %d" (get_inst_label y);
	    match y.crf_f_vec with
		UserVec(u) ->
		  FeatureTypeSet.iter 
		  (fun {FeatureType.fname=n;FeatureType.fdetail=d} ->
		     begin
		     match n with
			 F(nn) -> Printf.printf "  %s: " nn
		       | _ -> () end;
		     FeatureSet.iter (fun (ypr,yp,f,v) ->
					Printf.printf " (%d,%d,%d,%f)" ypr yp f v) d;
		     Printf.printf "\n") u
	      | _ -> ()) x) tr_data
			   
