(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** 
  Implementation of features.  This module contains the encodings for
  features, higher-order functions for applying user-defined feature extraction
  functions, functions for compiling features into a more efficient representation,
  and more.  It also contains fundamental data structures for each position
  within a sequence.
   
  @author Ben Wellner

*)

(** {6 Modules} *)

(** Basic feature type module.  Features are 4-tuples.  Comparator function
    is provided for creating feature sets. *)
module Feature :
    sig
      type t = int * int * int * float
      val compare : int * int * int * 'a -> int * int * int * 'b -> int
    end

(** Sets of [Feature]. *)
module FeatureSet :
    sig
      type elt = Feature.t
      and t = Set.Make(Feature).t
      val empty : t
      val mem : elt -> t -> bool
      val add : elt -> t -> t
      val iter : (elt -> unit) -> t -> unit
      val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
      val for_all : (elt -> bool) -> t -> bool
    end

type feature_name = F of string | Conj of feature_name * feature_name

(** Basic {e feature type}.  This is a class of features (e.g. word to the 
    left is "foo".  The specific features include the labels associated with
    an observation of the feature. The FeatureType structure also includes a 
    set of the basic instantiated features associated with it. *)
module FeatureType :
    sig
      type t = { fname : feature_name; mutable fdetail : FeatureSet.t; edgep: bool}
      val compare : t -> t -> int
      val edgep : t -> bool
    end


(** Sets of [FeatureType] *)
module FeatureTypeSet :
    sig
      type elt = FeatureType.t
      and t = Set.Make(FeatureType).t
      val empty : t
      val mem : elt -> t -> bool
      val add : elt -> t -> t
      val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
      val iter : (elt -> unit) -> t -> unit
      val for_all : (elt -> bool) -> t -> bool
    end      

(** Agglomerated features. *)
module AgglFeature :
  sig
    type t = feature_name
    val compare : t -> t -> int
  end

(** Set of [AgglFeature] *)
module AgglFeatureSet :
    sig
      type elt = AgglFeature.t
      and t = Set.Make(AgglFeature).t
      val empty : t
      val elements : t -> elt list
      val mem : elt -> t -> bool
      val add : elt -> t -> t
      val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
      val iter : (elt -> unit) -> t -> unit
      val union : t -> t -> t
    end      

module RankedAgglFeature :
  sig
    type t = feature_name * float
    val compare : t -> t -> int
  end

module RankedAgglFeatureSet :
    sig
      type elt = RankedAgglFeature.t
      and t = Set.Make(RankedAgglFeature).t
      val empty : t
      val elements : t -> elt list
      val mem : elt -> t -> bool
      val add : elt -> t -> t
      val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
      val iter : (elt -> unit) -> t -> unit
    end      

(** Inverse alphabet.   Maps integers back to alphabet type *)
module InvAlphabet :
  sig
    type 'a t
    val create : unit -> 'a t
    val get_inv_map : 'a t -> (int,'a) Hashtbl.t
    val get : 'a t -> int -> 'a
    val get_size : 'a t -> int
    val add : 'a t -> int -> 'a -> unit
  end

(** Alphabet.  Maps arbitrary types to integers. *)
module Alphabet :
  sig
    type 'a t
    val create : unit -> 'a t
    val get_map : 'a t -> ('a,int) Hashtbl.t
    val get_size : 'a t -> int
    val get : 'a t -> 'a -> int
    val update : 'a t -> bool -> 'a -> int
    val get_inverse : 'a t -> 'a InvAlphabet.t
  end

(** Module encapsulating information at each sequence position *)
module Source :
  sig
    type 'a t

    (** [create lab obs info] creates a new [Source.t] record with
      label [lab], observation [obs] and arbitrary typed infor in [info] *)
    val create : int -> string option -> 'a option -> 'a t

    (** Get the label, an [int] *)
    val get_lab : 'a t -> int

    (** Get the observation, a [string option] *)
    val get_obs : 'a t -> string option

    (** Get the arbitrary info, ['a option] *)
    val get_info : 'a t -> 'a option
  end



(** {6 Exceptions} *)

(** Thrown when a feature function fails due to unexpected observations
    or labels. *)
exception Feature_failure

(** {6 Types} *)

(** Main feature vector representation for CRFs.  Feature vectors are
    either compact [UserVec] or expanded (i.e. compiled) [CompVec]. *)
type crf_vec =
    UserVec of FeatureTypeSet.t
  | CompVec of FeatureSet.elt array * int
(*  | CompVec of FeatureSet.elt array * int *)


type user_feature = int * int * string

type crf_instance = {mutable crf_label: int;
		     mutable crf_orig_label: int;
		     mutable crf_f_vec: crf_vec}

type data_seq = crf_instance array

type train_data = crf_instance array array

(** {6 Feature manipulation functions} *)

val flatten_feature_name : feature_name -> string
val unset : feature_name -> string
val set_f : string -> feature_name
val set_fs : (string * 'a) list * 'b -> (feature_name * 'a) list * 'b
val unset_fs : (feature_name * 'a) list * 'b -> (string * 'a) list * 'b

(** {6 Feature vector accessors} *)

val get_compiled_vec : crf_instance -> FeatureSet.elt array * int
val get_uncompiled_vec : crf_instance -> FeatureTypeSet.t

(** {6 Instance accessors} *)

val get_inst_label : crf_instance -> int
val set_inst_label : crf_instance -> int -> unit
val set_inst_orig_label : crf_instance -> int -> unit
val set_as_uvec : crf_instance -> FeatureTypeSet.t -> unit

val create_instance : int -> crf_instance

val add_feature : ?unsupported:int option -> data_seq -> (feature_name, FeatureTypeSet.elt) Hashtbl.t -> int -> int -> int -> feature_name ->
  (int * int * feature_name) Alphabet.t -> float -> unit

val inst_features : crf_instance -> FeatureSet.elt array * int

val add_runtime_feature :
  ?unsupported: int option ->
  data_seq ->
  ('a, FeatureTypeSet.elt) Hashtbl.t -> int -> 'a -> 'b -> unit

(** Higher order function that applies a list of feature functions to the current position in
    the sequence. [FeatureGen.apply_feature_fns fset fa la fns yprev yp seq x i] applies
  the feature functions [fns] to the current position [i] in the data sequence [x].  [yp]
  is the label at position [i] and [yprev] is the label at [i] - 1.  [fset] is the feature set
  hashtable and [fa] and [la] are the feature and label alphabets, respectively. [seq] is the 
  sequence of instances.
*)
val apply_feature_fns :
  (feature_name, FeatureTypeSet.elt) Hashtbl.t ->
  (int * int * feature_name) Alphabet.t ->
  'a Alphabet.t ->
  ('b Source.t array -> int -> (feature_name * float) list * bool) list ->
  int -> int -> crf_instance array -> 'b Source.t array -> int -> unit

(** Higher level function that applies a list of feature functions to the current position
    at {e runtime}. [FeatureGen.apply_feature_fns_runtime fset fns seq x i] applies the functions
  [fns] to the current position [i] in the data sequence [x].
*)
val apply_feature_fns_runtime :
  (feature_name, FeatureTypeSet.elt) Hashtbl.t ->
  ('b Source.t array -> int -> (feature_name * float) list * bool) list ->
  crf_instance array -> 'b Source.t array -> int -> unit

val compile_data : train_data -> unit

val compile_seq : data_seq -> unit

val print_train_data : train_data -> unit
