(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** Module implementing Feature Induction.

Some thoughts:

What is the result?  We derive a "feature mask".  At runtime, for example, the same "atomic"
feature functions are applied, when an atomic feature is in the "feature mask" it is added as a
feature.  Then a second pass adds 1st order conjunctions, a third, 2nd order, etc.

During training the same idea applies.  We incrementally build up a "feature mask" - which is just
a new "fset".  We iterate over the "original training instances" and create new training instances
that only include features in the mask.  We make 2nd and 3rd passes as above to derive conjunctions.

*)

open FeatureGen

type lookup_t = KSet of AgglFeatureSet.t | KHash of (FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t

let num_training_iters = ref 20
let beam_size = ref 500
let num_rounds = ref 30

let set_beam_size s = beam_size := s
let set_num_rounds n = num_rounds := n

(* wasteful, but makes code a bit simpler - "coerce" AgglFeature to FeatureType
   and use set membership
   XXX - this is way down in an inner loop - may hurt performance *)
let feature_type_set_member aggl_feature tset =
  FeatureTypeSet.mem {FeatureType.fname=aggl_feature; FeatureType.fdetail = FeatureSet.empty; FeatureType.edgep=false} tset

(* create new version of training data with no features *)
let re_initialize tr_data =
  Array.map
    (Array.map
       (fun inst -> FeatureGen.create_instance (FeatureGen.get_inst_label inst))) tr_data

(*       
let rec compute_conjunctions features acc =
  match features with 
      f :: r ->
	compute_conjunctions r (List.rev_append (List.fold_left (fun c1 f1 -> Conj(f,f1) :: c1) [] r) acc)
    | [] -> acc *)

let compute_conjunctions fset =
  let felements = AgglFeatureSet.elements fset in
  let conjset = AgglFeatureSet.empty in
  let rec compute_conjunctions_x features cset =
    match features with
	f :: r -> 
	  compute_conjunctions_x r (List.fold_left (fun c1 f1 -> AgglFeatureSet.add (Conj(f,f1)) c1) cset r)
      | [] -> cset in
    compute_conjunctions_x felements conjset

let rec conj_present conj fvec =
  match conj with
      F f -> feature_type_set_member (F(f)) fvec
    | Conj(f1,f2) -> (conj_present f1 fvec) && (conj_present f2 fvec)

(* this deals with the runtime use of conjuntive features *)	
let augment_tst_data tr_data la aug_fset conj_tbl =
    Array.map
      (fun seq ->
	 let nseq = Array.init (Array.length seq) (fun _ -> create_instance (0)) in
	   Array.iteri
	     (fun i {crf_label=yp; crf_f_vec=fvec} -> 
		match fvec with 
		    UserVec(fvec) ->
		      FeatureTypeSet.iter 
			(fun f ->
			   if Hashtbl.mem aug_fset f.FeatureType.fname then (
			     add_runtime_feature nseq aug_fset i f.FeatureType.fname 1.0;
			     let conjs = Hashtbl.find_all conj_tbl f.FeatureType.fname in
			       List.iter
				 (fun conj ->
				    if conj_present conj fvec then (
				      add_runtime_feature nseq aug_fset i conj 1.0)) conjs)) fvec
		  | _ -> failwith "Uncompiled Vector required during sequence augmentation") seq;
	   nseq) tr_data
  

(* this creates a new "temporary" training set with only those features in k_set
   included in the feature vectors *)
let get_new_tr_data_x tr_data la k_map conj_tbl =
  let nfset = Hashtbl.create 10 in
  let nfa = Alphabet.create () in
  let nlabels = Alphabet.get_size la in
  let yprev = ref (-1) in
  let new_tr_data =
    Array.map
      (fun seq ->
	 yprev := -1;
	 let nseq = Array.init (Array.length seq) (fun _ -> create_instance (-1)) in
	   Array.iteri
	     (fun i {crf_label=yp; crf_f_vec=fvec} -> 
		begin
		match fvec with 
		    UserVec(fvec) ->
		      set_inst_label nseq.(i) yp;
		      FeatureTypeSet.iter 
			(fun f -> 
			   if (match k_map with
				   KSet(k_set) -> AgglFeatureSet.mem f.FeatureType.fname k_set
				 | KHash(k_hash) -> Hashtbl.mem k_hash f.FeatureType.fname) then (
			     let act_yprev = if (FeatureType.edgep f) then !yprev else -1 in
			       add_feature nseq nfset i act_yprev yp f.FeatureType.fname nfa 1.0;
			       let conjs = Hashtbl.find_all conj_tbl f.FeatureType.fname in
				 List.iter
				   (fun conj ->
				      if conj_present conj fvec then (
					add_feature nseq nfset i act_yprev yp conj nfa 1.0)) conjs)) fvec
		  | _ -> failwith "Uncompiled Vector required\n"
		end;
		yprev := yp) seq;
	   nseq) tr_data in
    (new_tr_data, nfa, nfset)

let get_new_tr_data tr_data la kset conj_tbl = get_new_tr_data_x tr_data la (KSet kset) conj_tbl

let augment_tr_data tr_data la kset conj_tbl = get_new_tr_data_x tr_data la (KHash kset) conj_tbl

(* create new insts (now an array of lists), including _only_ the mislabeled instances *)
let get_mislabeled_insts true_labels insts =
  assert ((Array.length true_labels) = (Array.length  insts));
  Array.mapi
    (fun i seq ->
       let j = ref 0 in
	 List.rev 
	   (Array.fold_left
	      (fun c inst -> 
		 incr j;
		 if not (inst.crf_label = true_labels.(i).(!j-1)) then
		   (inst :: c)
		 else c) [] seq)) insts

let rec get_atomics conj =
  match conj with
      F(x) -> [F(x)]
    | Conj(x,y) -> List.rev_append (get_atomics x) (get_atomics y)

let update_conj_table ctable g_set =
  AgglFeatureSet.iter
    (fun x ->
       match x with
	   F(_) -> ()
	 | c -> 
	     let atomics = get_atomics c in
	       List.iter (fun a -> Hashtbl.add ctable a c) atomics) g_set

let tmp_conj_table orig_ctable g_set = 
  let copy = Hashtbl.copy orig_ctable in
    update_conj_table copy g_set;
    copy


let rank_cand_feature_types crf_i nls cand_nfa cand_fset cand_tr_data =
  let nfs = Alphabet.get_size cand_nfa in
  let nparams = Crf.make_a1 nfs in
  let ngradient = Crf.make_a1 nfs in
  let last_gain = ref 0.0 in
  let f_nfs = float_of_int nfs in
  let f_tr_len = float_of_int (Array.length cand_tr_data) in
  let grad_fn = if (f_tr_len *. f_nfs) > 1e+9 then Crf.get_gradient_sparse else Crf.get_gradient_dense in
    assert (crf_i.Crf.nfs <= nfs);
    for i = 0 to crf_i.Crf.nfs - 1 do
      nparams.{i} <- crf_i.Crf.parameters.{i} done;
    let tr = Crf.init nfs nls nparams in
      ignore (grad_fn nparams ngradient cand_tr_data tr nfs true);
      let ranked_features = 
	Hashtbl.fold
	  (fun k {FeatureType.fname=fname; FeatureType.fdetail=fdetail} c ->
	     let gain = ref 0. in
	       FeatureSet.iter (fun (_,_,f,_) -> gain := !gain +. ngradient.{f}) fdetail;
	       RankedAgglFeatureSet.add (fname,!gain) c) cand_fset RankedAgglFeatureSet.empty in
	ranked_features

let top_n_of_set ranked_aggl_set block_set n =
  let ra_list = RankedAgglFeatureSet.elements ranked_aggl_set in
  let rec get_top_n n li last_gain acc =
    if n <= 0 then acc
    else
      match li with
	  [] -> acc
	| (a,v) :: b -> 
	    if not (v = last_gain) && not (AgglFeatureSet.mem a block_set) then 
	      get_top_n (n-1) b v (AgglFeatureSet.add a acc) 
	    else get_top_n n b v acc in
    get_top_n n ra_list 0.0 AgglFeatureSet.empty


(* assumes tr_data contains the initial "atomic" features *)
let induce_crf_features ?(intermediate=50) ?(la_fset_d=None) ?(compact=false) ?(max_its=500) fset la tr_data =
  let nls = Alphabet.get_size la in
  let tr_labels = Array.map (Array.map (fun {crf_label=l} -> l)) tr_data in
  let c_set = ref (Hashtbl.fold (fun k v cset -> AgglFeatureSet.add k cset) fset AgglFeatureSet.empty) in
  let k_set = ref AgglFeatureSet.empty in
  let conj_table = Hashtbl.create 10 in (* hash from atomic features to conjunctions they're involved in *)
    for i = 1 to !num_rounds do
      let new_tr_data, nfa, nfset = get_new_tr_data tr_data la !k_set conj_table in
      let crf_i = Crf.train_crf ~max_its:(!num_training_iters) (Alphabet.get_size nfa) nls new_tr_data in
      let cand_tr_data, cand_nfa, cand_fset = get_new_tr_data tr_data la !c_set conj_table in
      let ranked_features = rank_cand_feature_types crf_i nls cand_nfa cand_fset cand_tr_data in
      let a_cand_set = top_n_of_set ranked_features !k_set !beam_size in
      let g_cand_set = compute_conjunctions a_cand_set in
      let new_cand_set = AgglFeatureSet.union (AgglFeatureSet.union a_cand_set g_cand_set) !k_set in
	(*
	Printf.printf "\n ---- New Cand Set ----\n";
	AgglFeatureSet.iter (fun f -> Printf.printf "%s\n" (flatten_feature_name f)) new_cand_set; flush stdout; *)
      let tmp_conj_tbl = tmp_conj_table conj_table new_cand_set in
      let conj_tr_data, conj_nfa, conj_fset = get_new_tr_data tr_data la new_cand_set tmp_conj_tbl in
      let ranked_new_features = rank_cand_feature_types crf_i nls conj_nfa conj_fset conj_tr_data in
      let g_set = top_n_of_set ranked_new_features !k_set !beam_size in
	Printf.printf "\n+++++++++++++++++++++++++++++++++++\n";
	Printf.printf "\n Top (%d) induced features: \n\n" !beam_size;
	AgglFeatureSet.iter (fun f -> Printf.printf "%s\n" (flatten_feature_name f)) g_set; flush stdout;
	update_conj_table conj_table g_set;
	k_set := AgglFeatureSet.union g_set !k_set;
	c_set := AgglFeatureSet.union !c_set g_set;
    done;
    let final_tr_data, final_nfa, final_fset = get_new_tr_data tr_data la !k_set conj_table in
    let new_la_f_set = match la_fset_d with None -> None | Some(f,fset,la) -> Some(f,final_fset,la) in
    let final_crf = 
      Crf.train_crf ~intermediate:intermediate ~la_fset_d:new_la_f_set ~max_its:max_its 
	(Alphabet.get_size final_nfa) nls final_tr_data in
      (final_crf, Some(final_fset,conj_table))
