(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** An implementation of Maximum Entropy Markov Models *)
open Bigarray
module MLO = Lbfgs.Opt

exception Return 
type barr = Crf.barr

let check s = print_endline s; flush stdout

let get_gradient (lambdas:barr) (gradient:barr) (data_seqs:FeatureGen.train_data) nfs nls =
  let logli = ref 0. in
  let scores = Array.init nls (fun _ -> 0.0) in
    for i=0 to (Array1.dim lambdas) - 1 do
      let li = lambdas.{i} in
	gradient.{i} <- li *. !Crf.invSigSquare;
	logli := !logli +. (li *. li *. !Crf.invSigSquare) /. 2.
    done;
    for j=0 to (Array.length data_seqs) - 1 do
      let data_seq = data_seqs.(j) in    
      let len = Array.length data_seq in
	for k = 0 to len - 1 do
	  Array.fill scores 0 nls 0.0; (* reset scores *)
	  let label = FeatureGen.get_inst_label data_seq.(k) in
	  let previous = if k > 0 then FeatureGen.get_inst_label data_seq.(k-1) else (-1) in
	  let (inst_features,size) = FeatureGen.get_compiled_vec data_seq.(k) in
	    for m = 0 to size - 1 do
	      let (yprev,yp,f,vl) = inst_features.(m) in
		if (((k > 0) && (previous = yprev)) || yprev < 0) then
		  if yp = label then gradient.{f} <- gradient.{f} -. vl; (* update constraints *)
		scores.(yp) <- scores.(yp) +. vl *. lambdas.{f};
	    done;
	    let max = ref 0.0 and sum = ref 0.0 in
	      for i = 0 to nls - 1 do if scores.(i) > !max then max := scores.(i) done;
	      for i = 0 to nls - 1 do 
		scores.(i) <- exp (scores.(i) -. !max);
		sum := !sum +. scores.(i)
	      done;
	      for i = 0 to nls - 1 do scores.(i) <- scores.(i) /. !sum done;
	      logli := !logli -. (log scores.(label));
	      for m = 0 to size - 1 do (* loop back through features and update gradient *)
		let (yprev,yp,f,vl) = inst_features.(m) in
		  if (((k > 0) && (previous = yprev)) || yprev < 0) then
		    gradient.{f} <- gradient.{f} +. scores.(yp) *. vl
	      done;
	done;
    done;
    !logli
		  

let train_memm ?(verbose=true) ?(intermediate=50) ?(la_fset_d=None) ?(i_params=None) ?(compact=true) ?(max_its=1000) num_features num_labels tr_data =
  let iprint = Array.make 2 0
  and params = match i_params with None -> Array1.create float64 c_layout num_features | Some p -> p
  and gradient = Array1.create float64 c_layout num_features
  and diagco = ref false
  and diag = Array1.create float64 c_layout num_features
  and eps = 1.0e-2
  and xtol = 4.9e-300
  and icall = ref 0
  and m = 7
  and iflag = ref 0 in
  let max_iters = max_its in
  let iter = ref 0 in
  let n = num_features in
  let float_n = float_of_int n in
  let num_seqs = float_of_int (Array.length tr_data) in
  let f = ref 0. in
    if not compact then (
      Gc.compact(); (* do a major GC before compiling data *)
      FeatureGen.compile_data tr_data);
    begin match i_params with None -> Array1.fill params 0. | _ -> () end;
    Array1.fill gradient 0.;
    Array1.fill diag 0.;
    iprint.(0) <- (if verbose then 1 else 0);
    iprint.(1) <- 0;
    Printf.printf "\n initializing MEMM\n";
      Printf.printf "\n Number of features: %d" num_features;
      f := get_gradient params gradient tr_data num_features num_labels;
      Printf.printf "\n Initial f-val: %f" !f; flush stdout;
      MLO.Lbfgs.init (); (* initialize lbfgs routine *)
      if !f > 0.0 then (
	try
	  while (!iter < max_iters) do
	    if !iter > 0 && (!iter mod intermediate) = 0 then (
	      match la_fset_d with
		  Some (f, fset, la) ->
		    Printf.printf "Writing out intermediate model..%d\n" !iter;
		    Crf.marshal_crf (f^"_at_"^(string_of_int !iter)) 
		      (Crf.create_user_model (Crf.create_model params num_features num_labels) fset la None)
		| None -> ());
            MLO.Lbfgs.lbfgs n m params f gradient diagco diag iprint eps xtol iflag;
	    if (!iflag <= 0) then
	      raise Return;
	    f := get_gradient params gradient tr_data num_features num_labels;
	    iter := !iter + 1
	  done;
	  Printf.printf "\n Final f: %f\n\n" !f;
	  Crf.create_model params num_features num_labels
	with Return -> 
	  Printf.printf "\n Final f -> %f\n\n" !f;
	  Crf.create_model params num_features num_labels)
      else Crf.create_model params num_features num_labels
