(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)


module type FEATURE_MANAGER = 
  sig
    type t
    val of_string : string -> t
    val of_assoc_list : (string * string) list -> t
    val get_feature_fns : string -> TextFeatures.Lexicon.t option -> 
      (t FeatureGen.Source.t array ->
	 int -> (FeatureGen.feature_name * float) list * bool)
      list
  end

(** Module type for a sequence which encapsulates a feature manager *)
module type SEQ =
  sig
    (** Type "inherited" from the user-defined {! Sequences.FEATURE_MANAGER} *)
    type t

    (** Data structure for a sequence - an array of [t FeatureGen.Source.t] records *)
    type seq = t FeatureGen.Source.t array
    
    (** Accessor to get back auxillary data from {! Sequences.FEATURE_MANAGER} *)
    val of_string : string -> t

    (** Get back/set aux data in assoc list form *)
    val of_assoc_list : (string * string) list -> t

    (** Get the user defined feature functions. *)      
    val get_feature_fns : string -> TextFeatures.Lexicon.t option -> 
      (seq ->
	 int -> (FeatureGen.feature_name * float) list * bool)
      list

    (** Get the training data sequences *)
    val create_training_data_seqs :
      seq list ->
      (FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t ->
      (int * int * FeatureGen.feature_name) FeatureGen.Alphabet.t ->
      'b FeatureGen.Alphabet.t ->
      (seq -> int -> (FeatureGen.feature_name * float) list * bool)
      list -> FeatureGen.data_seq list

    (** Create the runtime sequences *)
    val create_runtime_seqs :
      seq list ->
      (FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t ->
      (seq -> int -> (FeatureGen.feature_name * float) list * bool) list ->
      FeatureGen.data_seq list
  end


module Make (Fm: FEATURE_MANAGER): SEQ with type t = Fm.t =
  struct
    type t = Fm.t
    type seq = t FeatureGen.Source.t array
    let get_feature_fns = Fm.get_feature_fns

    let of_string s = Fm.of_string s
    let of_assoc_list = Fm.of_assoc_list

    let create_training_data_seqs (tl_pairs: 'a FeatureGen.Source.t array list) fset fa la fns =
      let yprev = ref (-1) in
      let insts = 
	List.map
	  (fun x ->
	     yprev := -1;
	     let size = Array.length x in
	     let inst_seq = Array.init size (fun _ -> FeatureGen.create_instance (-1)) in
	       for i=0 to size - 1 do
		 let t1 = FeatureGen.Source.get_obs x.(i)
		 and yp = FeatureGen.Source.get_lab x.(i) in
		   FeatureGen.apply_feature_fns fset fa la fns !yprev yp inst_seq x i;
		   FeatureGen.set_inst_label inst_seq.(i) yp;
		   FeatureGen.set_inst_orig_label inst_seq.(i) yp;
		   yprev := yp
	       done;
	       inst_seq ) tl_pairs  in
	insts

    let create_runtime_seqs tokens fset fns =
      List.map
	(fun x ->
	   let size = Array.length x in
	   let (inst_seq: FeatureGen.data_seq) = Array.init size (fun x -> FeatureGen.create_instance 0) in	 
	     for i=0 to size - 1 do
	       let yp = FeatureGen.Source.get_lab x.(i) in
	       FeatureGen.apply_feature_fns_runtime fset fns inst_seq x i;
	       FeatureGen.set_inst_orig_label inst_seq.(i) yp;
	     done;
	     inst_seq)
	tokens
  end

