(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** Functor for customizing sequence data types and extraction functions.
 
  @author Ben Wellner 
*)

(** Module type that defines feature managers.  These will be specified by the
    user developer for each application or task. *)
module type FEATURE_MANAGER = 
  sig
    type t
    val of_string : string -> t
    val of_assoc_list : (string * string) list -> t
    val get_feature_fns : string -> TextFeatures.Lexicon.t option -> 
      (t FeatureGen.Source.t array ->
	 int -> (FeatureGen.feature_name * float) list * bool)
      list
  end

(** Module type for a sequence which encapsulates a feature manager *)
module type SEQ =
  sig
    (** Type "inherited" from the user-defined {! Sequences.FEATURE_MANAGER} *)
    type t

    (** Data structure for a sequence - an array of [t FeatureGen.Source.t] records *)
    type seq = t FeatureGen.Source.t array
    
    (** Accessor to get back auxillary data from {! Sequences.FEATURE_MANAGER} *)
    val of_string : string -> t

    (** Get back/set aux data in assoc list form *)
    val of_assoc_list : (string * string) list -> t

    (** Get the user defined feature functions. *)      
    val get_feature_fns : string -> TextFeatures.Lexicon.t option -> 
      (t FeatureGen.Source.t array ->
	 int -> (FeatureGen.feature_name * float) list * bool)
      list

    (** Get the training data sequences *)
    val create_training_data_seqs :
        seq list ->
      (FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t ->
        (int * int * FeatureGen.feature_name) FeatureGen.Alphabet.t ->
        'b FeatureGen.Alphabet.t ->
        (seq -> int -> (FeatureGen.feature_name * float) list * bool)
        list -> FeatureGen.data_seq list

    (** Create the runtime sequences *)
    val create_runtime_seqs :
      seq list ->
      (FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t ->
      (seq -> int -> (FeatureGen.feature_name * float) list * bool) list ->
      FeatureGen.data_seq list
  end

(** Functor to create a sequence model over the user defined data types and
    feature extraction functions. *)
module Make (Fm: FEATURE_MANAGER): SEQ with type t = Fm.t
