(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** An SGML front-end for CRFs.  Allows a CRF to be trained and applied to token sequences
where labels are indicated with tags (Ramshaw & Marcus style) and tokens are separated by
white space (i.e. there is a very simple tokenizer used here for now) *)

exception Invalid_target_tag


module type SGML_FRONT_TYPE =
    sig
      type att = Att of string * string 
      type tag = STag of string | Tag of string * att | Con of tag * tag 
      type 'a train_t = Crf | Vp | Hybrid | CrfInduce | CrfMask of 'a Crf.user_crf_model | Memm
      val apply_model_to_dir :
	?print_lex: bool ->
	?tagset: tag list option ->
	?toker: (Sgml.xml -> Sgml.xml) ->
	?odir:string option ->
        string ->
        Str.regexp ->
        string ->
        (FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t ->
        tag FeatureGen.Alphabet.t ->
        Crf.model ->
        string ->
        ((FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t *
         (FeatureGen.feature_name, FeatureGen.feature_name) Hashtbl.t)
        option ->
        TextFeatures.Lexicon.t option -> string -> bool -> bool -> unit
      val train_from_sgml :
	?toker: (Sgml.xml -> Sgml.xml) ->
	?verbose:bool ->
	?intermediate:int ->
        ?tri_gram:bool ->
        ?i_file_root:string option ->
        ?compact:bool ->
        ?max_its:int ->
        ?train_type:'a train_t ->
        ?tr_filter:Str.regexp ->
        ?add_begin:bool ->
	?input_file:bool ->
        string ->
        string ->
        string ->
        tag list -> TextFeatures.Lexicon.t option -> tag Crf.user_crf_model
      val run_on_sgml :
	?print_lex: bool ->
	?tagset:tag list option ->
	?toker: (Sgml.xml -> Sgml.xml) ->
        ?tri_gram:bool ->
        ?tst_filter:Str.regexp ->
        ?add_begin:bool ->
	?input_file:bool ->
	?output_file:string ->
        string ->
        string ->
        string ->
        tag Crf.user_crf_model ->
        string -> TextFeatures.Lexicon.t option -> unit
      val gather_tagset : string -> tag list
  end


(** Functor to create an SgmlFront module for a particular sequence type *)
module Make(Seq: Sequences.SEQ) : SGML_FRONT_TYPE =
  struct
    
    (** Training method *)
    type 'a train_t = 
	Crf  (** Standard CRF training *)
      | Vp   (** Voted perceptron training *)
      | Hybrid  (** Hybrid - initialize using voted perceptron, the CRF training *)
      | CrfInduce (** CRF training with automatic feature induction *)
      | CrfMask of 'a Crf.user_crf_model (** CRF training using initial, fixed feature mask *)
      | Memm (** Maximum Entropy Markov Model Training *)
	
    (** Attribute type *)
    type att = Att of string * string

    (** Type for tags which will be mapped to labels *)      
    type tag = 
	STag of string 
      | Tag of string * att
      | Con of tag * tag

    (** Token label pairs which have restricted types at this stage (polymorphic in the future) *)	  
    type tl_pair = {tok: string; (** Observed token *)
		    lab: tag;    (** The label of type tag *)
		    aux: Seq.t option} (** Arbitrary auxiallary information *)
	
    let split_reg = Str.regexp " +"

    let default_toker (s: Sgml.xml) : Sgml.xml = s

    let tag_of_interest tagset l ats : tag list =
      let just_on_tag = 
	List.filter
	  (fun x ->
	     match x with
		 STag (s) ->
		   s = l
	       | _ -> false) tagset in
      let attribute_matched =
	List.fold_left
	  (fun c x ->
	     match x with
		 Tag(s,Att(a,v)) ->
		   begin
		     try
		       let (aa,vv) = List.find (fun (a1,v1) -> a1 = a) ats in
			 if s = l then
			   Tag(s,(Att(aa,vv))) :: c
			 else c
		     with Not_found -> c end
	       | _ -> c) [] tagset in
	just_on_tag @ attribute_matched
	  
    let rec make_begin_tag tag =
      match tag with
	  STag t ->
	    STag("B_"^t)
	| Tag(t,a) ->
	    Tag("B_"^t,a)
	| x -> x
	    
    let un_begin_tag tag =
      match tag with
	  STag t -> 
	    if (String.sub t 0 2) = "B_" then
	      STag (String.sub t 2 ((String.length t) - 2))
	    else (STag t)
	| Tag(t,a) ->
	    if (String.sub t 0 2) = "B_" then
	      Tag ((String.sub t 2 ((String.length t) - 2)), a)
	    else Tag(t,a)
	| x -> x

	    
    (** This converts an Sgml structure to a list of tl_pairs.  Assumes tokens are indicated with
        <lex> tags. *)
    let sgml_to_tl_pairs_with_toks sgml tagset add_begin boundary_tag tri_gram =
      let no_pos = List.exists (fun x -> match x with STag(t) -> t="lex" | Tag(t,_) -> t="lex" | _ -> false) tagset in
      let last_tag = ref (STag "") in (* use this to keep previous tag to encode state tri-grams *)
      let rec create_features_from_toks toks rlab cnt =
	  match toks with 
	      [] -> []
	    | Sgml.Element("lex",atts,[Sgml.PCData t1]) :: rest ->
		let xlab = if add_begin && cnt = 0 then make_begin_tag rlab else rlab in
		  let lab = if tri_gram then Con(!last_tag,xlab) else xlab in
		    last_tag := xlab; (* set the last label just when we get to a tok *)
		    if false (* no_pos *) then  (* fairly sure this can go, always safe to include attributes *)
		      {tok=t1;lab=lab;aux=None} :: create_features_from_toks rest rlab (cnt+1)
		    else (
		      (* XXX - This will add all attributes of "lex" tags as auxillary "info" for the
			 token. Can thus be used to encode arbitrary additional features of the data
			 fairly straightforwardly *)
		      {tok=t1;lab=lab;aux=(try Some (Seq.of_assoc_list atts) with Not_found -> None)})
		      :: create_features_from_toks rest rlab (cnt+1)
	    | Sgml.Element(_,_,content) :: rest ->
		create_features_from_toks (content@rest) rlab cnt
	    | _ :: rest ->
		create_features_from_toks rest rlab cnt in
      let rec sgml_to_features_x sgml_l =
	match sgml_l with
	    [] -> []
	  | Sgml.Element(lab,atts,toks) :: rest ->
	      let tag = tag_of_interest tagset lab atts in
		if tag = [] then
		  if lab = "lex" || lab = "LEX" then
		    let x1 = create_features_from_toks [Sgml.Element(lab,atts,toks)] (STag lab) 0 in
		      x1 @ sgml_to_features_x rest
		  else
		    sgml_to_features_x (toks@rest)
		else 
		  if lab = "lex" || lab = "LEX" then (*special case *)
		    let x1 = create_features_from_toks [Sgml.Element(lab,atts,toks)] (List.hd tag) 0 in
		      x1 @ sgml_to_features_x rest
		  else 
		    let x1 = create_features_from_toks toks (List.hd tag) 0 in
		      x1 @ sgml_to_features_x rest
	  | Sgml.PCData _ :: rest ->
	      sgml_to_features_x rest in
      let rec sgml_to_sentences sgml_l =
	last_tag := (STag("=BEGIN="));
	match sgml_l with
	    [] -> []
	  | Sgml.Element(l,_,content) :: rest when l = boundary_tag ->
	      (* IMPORTANT - add a begin state *)
	      (Array.of_list ({tok="-BOUNDARY-";lab=STag("=BEGIN=");aux=None} :: (sgml_to_features_x content))) 
	      :: sgml_to_sentences rest
	  | Sgml.Element(_,_,content) :: rest ->
	      sgml_to_sentences (content @ rest)
	  | Sgml.PCData _ :: rest ->
	      sgml_to_sentences rest in
	sgml_to_sentences [sgml]

    let remove_lex_tags sg_l =
      let rec remove sg_l acc =
	match sg_l with
	    [] -> List.rev acc
	  | Sgml.Element("lex",_,[Sgml.PCData c]) :: r ->
	      remove r ((Sgml.PCData c) :: acc)
	  | Sgml.Element(l,a,c) :: r ->
	      let nc = remove c [] in
		remove r (Sgml.Element(l,a,nc) :: acc)
	  | Sgml.PCData c :: r ->
	      remove r ((Sgml.PCData c) :: acc) in
	remove sg_l []
	  
    (** This function converts labeled token sequences back into Sgml. *)
    let sequences_to_sgml ?(print_lex=true) outfile sgml seqs la inv_la seq_boundary begin_labels =
      let lex_ind_cand = FeatureGen.Alphabet.update la true (STag "lex") in
      let lex_ind = 
	if lex_ind_cand > -1 then lex_ind_cand else 
	  (FeatureGen.Alphabet.update la true (make_begin_tag (STag "lex"))) in
      let pos = ref 0 in
      let seq_index = ref 0 in
      let at_position seq pos =
	try
	  seq.(pos)
	with (Invalid_argument a) -> 
	  Printf.printf "\nposition: %d array length: %d\n" pos (Array.length seq); flush stdout;
	  raise (Invalid_argument a) in
      let rec add_annotations sgml_l seq add_tag seqlen =
	match sgml_l with
	    [] -> []
	  | Sgml.Element("lex",atts,c) :: rest as all ->
	      let beg_ind = FeatureGen.get_inst_label (at_position seq !pos) in
	      let i_beg_ind = FeatureGen.InvAlphabet.get inv_la beg_ind in
	      let utag = 
		(match i_beg_ind with
		     Con(_,t) -> if begin_labels then un_begin_tag t else t
		   | t -> if begin_labels then un_begin_tag t else t) in
	      let cind = 
		if begin_labels then 
		  try FeatureGen.Alphabet.get la utag with Not_found -> beg_ind else beg_ind in
		if beg_ind = lex_ind then (
		  incr pos;
		  Sgml.Element("lex",atts,c) :: add_annotations rest seq add_tag seqlen)
		else (
		  match utag with
		      Tag(t,Att(at,vl)) when t = "lex" || t = "LEX" ->
			incr pos;
			Sgml.Element(t,[(at,vl)],c) :: add_annotations rest seq add_tag seqlen
		    | STag t -> 
			let (elements,new_rest,_) = get_remaining_tokens cind beg_ind all [] seq 0 seqlen in
			  if add_tag && (not (t = "lex" || t = "LEX")) then 
			    Sgml.Element(t,[],elements) :: add_annotations new_rest seq add_tag seqlen
			  else
			    elements @ add_annotations new_rest seq add_tag seqlen
		    | Tag(t,Att(at,vl)) -> 
			let (elements,new_rest,_) = get_remaining_tokens cind beg_ind all [] seq 0 seqlen in
			  if add_tag then
			    Sgml.Element(t,[(at,vl)],elements) :: add_annotations new_rest seq add_tag seqlen
			  else
			    elements @ add_annotations new_rest seq add_tag seqlen
		    | _ -> failwith "Invalid Tag")
	  | Sgml.Element(l,a,content) :: rest ->
	      let sub = add_annotations content seq add_tag seqlen in
		Sgml.Element(l,a,sub) :: add_annotations rest seq add_tag seqlen
	  | Sgml.PCData c :: rest ->
	      Sgml.PCData c :: add_annotations rest seq add_tag seqlen
      and get_remaining_tokens ind beg_ind sgml_l acc seq lpos seqlen =
	if !pos < seqlen then
	  match sgml_l with
	      Sgml.Element(l,atts,tok) :: rest when 
		let cl = FeatureGen.get_inst_label (at_position seq !pos) in 
		  ((l = "lex" || l = "LEX") && (cl = ind || (lpos = 0 && cl = beg_ind))) ->
		    incr pos;
		    get_remaining_tokens ind beg_ind rest (Sgml.Element(l,atts,tok) :: acc) seq (lpos + 1) seqlen
	    | Sgml.PCData c :: rest as all ->
		let cl = FeatureGen.get_inst_label (at_position seq !pos) in
		  (* peek ahead to see if the next token is of this label type *)
		  if cl = ind || (lpos = 0 && cl = beg_ind) then
		    get_remaining_tokens ind beg_ind rest (Sgml.PCData c :: acc) seq lpos seqlen
		  else 
		    ((List.rev acc),all,lpos)
	    | Sgml.Element(l,atts,content) :: rest as all when not (l = "LEX" || l = "lex") ->
		let cl = FeatureGen.get_inst_label (at_position seq !pos) in
		  (* peek ahead to see if the next token is of this label type *)
		  if cl = ind || (lpos = 0 && cl = beg_ind) then
		    let new_content = add_annotations content seq false seqlen in
		      get_remaining_tokens ind beg_ind rest (Sgml.Element(l,atts,new_content) :: acc) seq lpos seqlen
		  else 
		    ((List.rev acc),all,lpos)
	    | rest ->
		((List.rev acc),rest,lpos)
	else
	  (List.rev acc),sgml_l,lpos in
      let rec add_annotations_sent sgml_l =
	match sgml_l with
	    [] -> []
	  | Sgml.Element(l,atts,content) :: rest when l = seq_boundary ->
	      pos := 1;  (* start at 1, not zero, because of =BEGIN= state *)
	      let s = seqs.(!seq_index) in
	      let nannots = add_annotations content s true (Array.length s) in
		incr seq_index;
		Sgml.Element(l,atts,nannots) :: add_annotations_sent rest
	  | Sgml.Element(l,a,content) :: rest ->
	      let sub = add_annotations_sent content in
		Sgml.Element(l,a,sub) :: add_annotations_sent rest
	  | Sgml.PCData x :: rest ->
	      Sgml.PCData x :: add_annotations_sent rest in
      let oc = open_out outfile in
      let nsg = add_annotations_sent [sgml] in
      let n_annots = List.hd (if print_lex then nsg else remove_lex_tags nsg) in
	Sgml.to_stream n_annots oc;
	close_out oc

    let sgml_to_tl_pairs sgml la tagset seq_boundary add_begin tri_gram =
      let ts = sgml_to_tl_pairs_with_toks sgml tagset add_begin seq_boundary tri_gram in
      let insts =
	List.map
	  (Array.map (fun {tok=t;lab=lab;aux=a} -> 
			FeatureGen.Source.create (FeatureGen.Alphabet.update la false lab) (Some t) a)) ts in
	insts
	  
    let gather_data_dir toker dir regexp_filter fset fa la tagset featureset lexicon seq_boundary add_begin tri_gram =
      let files = List.filter (fun x -> Str.string_match regexp_filter x 0 && (Str.matched_string x) = x)
		    (Array.to_list (Sys.readdir dir)) in
      let fns = Seq.get_feature_fns featureset lexicon in
      let paths = Array.of_list (List.map ((^) dir) files) in
	Array.fold_left
	  (fun c x ->
	     let sg = let s = Sgml.parse_file x in toker s in
	     let pairs = 
	       Seq.create_training_data_seqs 
		 (sgml_to_tl_pairs sg la tagset seq_boundary add_begin tri_gram) fset fa la fns in
	       List.rev_append c pairs ) [] paths
	  
    let apply_model_to_dir ?(print_lex=true) ?(tagset=None) ?(toker=default_toker) ?(odir=None) dir regexp_filter f_suffix fset la crf featureset fset_mask lexicon seq_boundary add_begin tri_gram =
      let files = List.filter (fun x -> Str.string_match regexp_filter x 0 && (Str.matched_string x) = x)
		    (Array.to_list (Sys.readdir dir)) in
      let paths = Array.of_list (List.map ((^) dir) files) in
      let inv_la = (FeatureGen.Alphabet.get_inverse la) in
      let fns = Seq.get_feature_fns featureset lexicon in
      let t1 = Sys.time () in
      let decoding_time = ref 0. in
      let cur_time = ref t1 in
	for i=0 to (Array.length paths) - 1 do
	  let sg = let s = Sgml.parse_file paths.(i) in toker s in
	    cur_time := Sys.time ();
	  let pairs = sgml_to_tl_pairs sg la (match tagset with None -> [] | Some s -> s) 
			seq_boundary false tri_gram in

	  let raw_inst_seqs = Array.of_list (Seq.create_runtime_seqs pairs fset fns) in
	  let inst_seqs = 
	    match fset_mask with 
		None -> raw_inst_seqs
	      | Some (mset,mtbl) -> 
		  Printf.printf "Augmenting data with induced features...\n"; flush stdout;
		  FeatureInduction.augment_tst_data raw_inst_seqs la mset mtbl in
	    Printf.printf "."; flush stdout;
	    let file = paths.(i) in
	      Array.iter
		(fun x ->
		   ignore (Viterbi.best_label_seq x crf)) inst_seqs;
	      decoding_time := !decoding_time +. ((Sys.time ()) -. !cur_time);
	      let ofile = 
		match odir with
		    None ->
		      file^f_suffix
		  | Some o -> o^f_suffix in
		sequences_to_sgml ~print_lex:print_lex ofile sg inst_seqs la inv_la seq_boundary add_begin
	done;
	Printf.printf "\n In-memory feature extraction and decoding completed in %f seconds" !decoding_time;
	Printf.printf "\n\nDecoding and printing completed in %f seconds \n\n" ((Sys.time ()) -. t1); flush stdout
	  
    let print_features fa_map =
      Hashtbl.iter
	(fun (i,i2,s) v ->
	   Printf.printf "\n %d %d %s -> %d" i i2 s v) fa_map
	
    let train_from_sgml 
      ?(toker=default_toker)
      ?(verbose=true)
      ?(intermediate=50)
      ?(tri_gram=false) 
      ?(i_file_root=None) 
      ?(compact=false) 
      ?(max_its=1000) 
      ?(train_type=Crf) 
      ?(tr_filter = (Str.regexp ".*")) 
      ?(add_begin=false)
      ?(input_file=false)
      seq_boundary trainpath featureset tagset lex = 
      
      let (fa: (int * int * FeatureGen.feature_name) FeatureGen.Alphabet.t) = FeatureGen.Alphabet.create () in
      let (la: tag FeatureGen.Alphabet.t) = FeatureGen.Alphabet.create () in
      let (fset: (FeatureGen.feature_name, FeatureGen.FeatureType.t) Hashtbl.t) = 
	Hashtbl.create 10 in (* hash of features *)
      let a_tr_filter = if input_file then Str.regexp (Filename.basename trainpath) else tr_filter in
      let a_train_dir = if input_file then (Filename.dirname trainpath)^"/" else trainpath in
      let (tr_seq: FeatureGen.train_data) = 	    
	Array.of_list (gather_data_dir toker a_train_dir a_tr_filter 
			 fset 
			 fa la tagset 
			 featureset lex seq_boundary add_begin tri_gram) in
	Printf.printf "\n Training with %d sequences\n" (Array.length tr_seq); flush stdout;
	Printf.printf "\n Training with %d instances\n" 
	  (let x = ref 0 in 
	     Array.iter (fun s -> x := !x + (Array.length s)) tr_seq;
	     !x); flush stdout;
	let la_fset_d = match i_file_root with Some f -> Some(f, fset, la) | None -> None in
	let la_size = FeatureGen.Alphabet.get_size la in
	let fa_size = FeatureGen.Alphabet.get_size fa in
	let crf,fset_mask = 
	  match train_type with
	      Vp ->
		VotedPerceptron.vp_train ~la_fset_d:la_fset_d ~compact:compact 
		~max_its:max_its fa_size la_size tr_seq, None
	    | Crf ->
		Crf.train_crf ~intermediate:intermediate ~la_fset_d:la_fset_d ~compact:compact 
		   ~max_its:max_its fa_size la_size tr_seq, None
	    | CrfMask(initial_model) ->
		let mask,ctable = 
		  match initial_model.Crf.fset_mask with 
		      Some(m,c) -> m,c 
		    | None -> failwith "Fset mask required for training from initial model." in
		let aug_data,nfa,nfset = FeatureInduction.augment_tr_data tr_seq la mask ctable in
		let n_la_fset_d = match i_file_root with Some f -> Some(f,nfset,la) | None -> None in
		  Crf.train_crf ~intermediate:intermediate ~la_fset_d:n_la_fset_d ~compact:compact
		    ~max_its:max_its (FeatureGen.Alphabet.get_size nfa) la_size aug_data, initial_model.Crf.fset_mask
	    | Hybrid ->
		VotedPerceptron.hybrid_train ~compact:compact ~vp_its:1 
		  ~crf_its:max_its la fset (FeatureGen.Alphabet.get_size fa) 
		  (FeatureGen.Alphabet.get_size la) tr_seq
	    | CrfInduce ->
		FeatureInduction.induce_crf_features ~intermediate:intermediate ~la_fset_d:la_fset_d 
		  ~compact:compact ~max_its:max_its fset la tr_seq
	    | Memm ->
		Memm.train_memm ~intermediate:intermediate ~la_fset_d:la_fset_d ~compact:compact 
		   ~max_its:max_its fa_size la_size tr_seq, None in
	  Printf.printf "\n Performance on training data: \n";
	  let aug_tr_data = 
	    match fset_mask with
		None -> tr_seq
	      | Some (mset,mtbl) ->
		  FeatureInduction.augment_tst_data tr_seq la mset mtbl in
	    (*
	  Array.iter
	    (fun x ->
	       let sc =  (Viterbi.best_label_seq x crf) in
		 Printf.printf "Sequence score: %f\n" sc) aug_tr_data; *)
	  Crf.create_user_model crf fset la fset_mask 
	    
    let run_on_sgml ?(print_lex=true) ?(tagset=None) ?(toker=default_toker) ?(tri_gram=false) ?(tst_filter=(Str.regexp ".*")) ?(add_begin=false) ?(input_file=false)
      ?(output_file="") seq_boundary testpath outsuffix {Crf.crf=crf; Crf.crf_fset=fset; Crf.crf_la=la; Crf.fset_mask=fset_mask} 
      featureset lex =
      let a_test_filter = if input_file then Str.regexp (Filename.basename testpath) else tst_filter in
      let a_test_dir = if input_file then (Filename.dirname testpath)^"/" else testpath in
      let act_out_dir,act_out_suffix = 
	if output_file = "" then 
	  None,outsuffix 
	else 
	  (Some ((Filename.dirname output_file)^"/")), (Filename.basename output_file) in
	apply_model_to_dir ~print_lex:print_lex ~tagset:tagset ~toker:toker ~odir:act_out_dir a_test_dir a_test_filter act_out_suffix fset la crf 
	  featureset fset_mask lex seq_boundary add_begin tri_gram
	
    let gather_tagset file =
      let ic = open_in file 
      and tagsplit = Str.regexp ":" 
      and attsplit = Str.regexp "=" 
      and tset = ref [] in
      let parse_line str =
	let els = Str.split tagsplit str in
	  match els with
	      [t] -> STag(t)
	    | t :: rest :: _ ->
		begin
		  match (Str.split attsplit rest) with
		      [at; vl] -> Tag(t,Att(at,vl))
		    | _ -> raise Invalid_target_tag end
	    | [] -> raise Invalid_target_tag in
	try
	  while true do
	    let l = (input_line ic) in
	      tset := parse_line l :: !tset
	  done;
	  !tset
	with End_of_file -> !tset
	  
  end
