(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)


(** An SGML front-end for CRFs.  Allows a CRF to be trained and applied to token sequences
  where labels are indicated with tags (Ramshaw & Marcus style) and tokens are separated by
  white space (i.e. there is a very simple tokenizer used here for now).  Implementation is a
  functor that takes a module, [Seq] of type {! Sequences.SEQ}. [Seq] specifies the feature
  extraction functions, auxillary data types and other task-specific aspects.
*)

module type SGML_FRONT_TYPE =
  sig
      (** Symbolic attribute-value pair *)
      type att = Att of string * string 

      (** Tag types *)
      type tag = STag of string | Tag of string * att | Con of tag * tag 
	
      (** Crf training method *)
      type 'a train_t = Crf | Vp | Hybrid | CrfInduce | CrfMask of 'a Crf.user_crf_model | Memm
      
      val apply_model_to_dir :
	?print_lex: bool ->
	?tagset: tag list option ->
	?toker: (Sgml.xml -> Sgml.xml) ->
	?odir:string option ->
        string ->
        Str.regexp ->
        string ->
        (FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t ->
        tag FeatureGen.Alphabet.t ->
        Crf.model ->
        string ->
        ((FeatureGen.feature_name, FeatureGen.FeatureTypeSet.elt) Hashtbl.t *
         (FeatureGen.feature_name, FeatureGen.feature_name) Hashtbl.t)
        option ->
        TextFeatures.Lexicon.t option -> string -> bool -> bool -> unit

      (** Main training call.  Creates training sequences from SGML and calls CRF trainer specified.
	[Sgmlfront.Make(Seq).train_from_sgml ?tri_gram ?i_file_root ?compact ?max_its ?train_type
	  ?tr_filter ?add_begin boundary tr_dir feature_set tag_set lexicon_opt] initiates CRF training
          over a directory containing one or more sgml files with inline tags (or tag attributes) 
          denoting labels over sequences.  [tri_gram] when true specifies that tag trigrams (instead
	of bigrams should be used.  [i_file_root] specifies the file path root to use for intermediate
	models generated during training.  [compact] specifies compact training. [max_its] is the max
	number of training iterations (default 1000).  [train_type] specifies which type of training
	(e.g. CRF, voted perceptron, CRF with feature induction).  [tr_filter] is a regular expression
	applied to the files in [tr_dir] - those files matching the regular expression are used for
	training.  [add_begin], when true, specifies that the first label (from the left) of a 
	continguous label sequence should receive a special begin label (e.g. B_NP for beginning
	of a noun phrase).  [feature_set] is a string handle denoting the specific feature set
	to use.  The allowed strings are specified in the Seq module. [tag_set] is a [tag list]
	of tags acquired by [Sgmlfront.Make(Seq).gather_tagset].  [lexicon_opt] is an optional
	lexicon to use as a source of features during training.
      *)
      val train_from_sgml :
	?toker:(Sgml.xml -> Sgml.xml) ->
	?verbose:bool ->
	?intermediate:int ->
        ?tri_gram:bool ->
        ?i_file_root:string option ->
        ?compact:bool ->
        ?max_its:int ->
        ?train_type:'a train_t ->
        ?tr_filter:Str.regexp ->
        ?add_begin:bool ->
	?input_file:bool ->
        string ->
        string ->
        string ->
        tag list -> TextFeatures.Lexicon.t option -> tag Crf.user_crf_model

      (** Main decoding call.  Creates sequences from SGML and runs against CRF model. *)	
      val run_on_sgml :
	?print_lex: bool ->
	?tagset:tag list option ->
	?toker:(Sgml.xml -> Sgml.xml) ->
        ?tri_gram:bool ->
        ?tst_filter:Str.regexp ->
        ?add_begin:bool ->
	?input_file:bool ->
	?output_file:string ->
        string ->
        string ->
        string ->
        tag Crf.user_crf_model ->
        string -> TextFeatures.Lexicon.t option -> unit

      (** Gather the specified set of tags for the task.  
	[Sgmlfront.Make(Seq).gather_tagset fname] gathers the tagset specified in the file [fname].
        See the users guide for details on the file format for specifying tagsets.
      *)
      val gather_tagset : string -> tag list
  end


module Make (Seq: Sequences.SEQ) : SGML_FRONT_TYPE
