(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** This module contains useful feature extraction functions of general utility *)


module Lexicon =
  (* A lexicon will actually store _all_ the lexicons - each word will map to the lexicon
     it belongs to.  It's basically just a wrapper of a Hash table. *)
  struct
    type t = (string, string) Hashtbl.t

    let create () = Hashtbl.create 100
    let lookup_all str lex = 
      Hashtbl.find_all lex str

    let add lex item i = Hashtbl.add lex item i

    let build_lexicon lex file =
      let fn = Filename.basename file in (* just the file suffix, not full path *)
      let ic = open_in file in
	try
	  while true do add lex (input_line ic) fn done
	with End_of_file -> close_in ic

    let build_lexicons_txt paths =
      let lex = create () in
	Array.iter (build_lexicon lex) paths;
	lex
  end

open FeatureGen

let prefixes_up_to l mx =
  let i = ref 0 in
  let prefixes = ref [] in
  let len = String.length l in
    while (!i < mx) && (!i < len) do 
      incr i;
      prefixes := (String.sub l 0 !i) :: !prefixes;
    done;
    !prefixes
    
let suffixes_up_to l mx =
  let len = String.length l in
  let i = ref len in
  let c = ref 0 in
  let suffixes = ref [] in
    while (!c < mx) && (!c < len) do 
      decr i;
      incr c;
      suffixes := (String.sub l !i !c) :: !suffixes;
    done;
    !suffixes

let affixes_f n dseq p =
  match Source.get_obs dseq.(p) with
      Some t ->
	let suffixes = (suffixes_up_to t n) in
	let prefixes = (prefixes_up_to t n) in
	let fs = ((List.map (fun x -> ("pref:"^x, 1.0)) prefixes) @
		  (List.map (fun x -> ("suf:"^x, 1.0)) suffixes)) in
	  (set_fs (fs,false))
    | None -> raise Feature_failure

	
let lexicon_membership_f ?(downcase=true) (lexicon: Lexicon.t) dseq p =
  match Source.get_obs dseq.(p) with
      Some t ->
	let lexs = Lexicon.lookup_all (String.lowercase t) lexicon in
	  if lexs = [] then
	    set_fs ([("N_l",1.0)], false)
	  else
	    let fs = List.map (fun x -> (("LEX:"^x), 1.0)) lexs in
	      set_fs (fs,false)
    | None -> raise Feature_failure

(* this should be curried on the first two arguments *)
let regexp_f_gen fname regexp dseq p =
  match Source.get_obs dseq.(p) with
      Some t ->
	if Str.string_match regexp t 0 then
	  set_fs ([(fname,1.0)],false)
	else 
	  raise Feature_failure
    | None -> raise Feature_failure

let wd_f dseq p =
  match Source.get_obs dseq.(p) with
      Some t ->
	set_fs ([(t, 1.0)], false)
    | None -> raise Feature_failure

let edge_f dseq p =
  set_fs ([("E", 1.0)], true)

let node_f dseq p =
  set_fs ([("U", 1.0)], false)


let n_gram_fn ?(edge=false) fn positions dseq p =
  let max = Array.length dseq in
  let fvs = 
    List.fold_left
      (fun c2 yp ->
	 let rp = p + yp in
	 let s = string_of_int rp in
	 let lfs = (if rp >= 0 && rp < max then match unset_fs (fn dseq rp) with fvs,_ -> fvs 
		    else if rp >= 0 then [("-END-",1.0)]
		    else [("-START-",1.0)]) in
	   (lfs,(string_of_int yp)) :: c2) [] positions in
  let fname = 
    (List.fold_left 
       (fun c (fvl,p) -> 
	  (List.fold_left 
	     (fun c1 (f,v) -> (c1^"&"^f^"("^p^")")) c fvl)) "" fvs) in
    set_fs ([(fname,1.0)],edge)

let window_fns fns window dseq p =
  let max_len = Array.length dseq in
  let fs = List.fold_left
      (fun c1 fn -> 
	(List.fold_left
	   (fun c2 rp -> 
	     let s = string_of_int rp in
	     try
	       let cp = p + rp in
	       if cp >= 0 && cp < max_len then
		 let (fvs,_) = unset_fs (fn dseq (p+rp)) in 
		 (List.map (fun (f,v) -> (("Win"^s^f),v)) fvs) @ c2
	       else c2
	     with Feature_failure -> c2) [] window)) [] fns in
  set_fs (fs,false)
  
