(**
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** This is a main routine for training segment/label CRFs over text domains *)

module MySeq = Sequences.Make(Feature_manager)
module MySgmlfront = Sgmlfront.Make(MySeq)
  
let main () =
  let tr_dir = ref ""
  and tr_regstr = ref ".*"
  and model_file = ref ""
  and tagset = ref "" 
  and lex_dir = ref ""
  and begin_encode = ref false
  and feature_set = ref ""
  and compact = ref false
  and max_its = ref 1000 
  and vp_train = ref false
  and crf_induce = ref false
  and hybrid_train = ref false
  and gp = ref 100.0
  and seq_boundary = ref "s"
  and intermediate = ref false in
  Arg.parse
    [("-input-dir", Arg.Set_string(tr_dir), "Input directory containing training files");
     ("-filter", Arg.Set_string(tr_regstr), "Input regexp filter");
     ("-model", Arg.Set_string(model_file), "Output model file");
     ("-tagset", Arg.Set_string(tagset), "Tagset to train against");
     ("-compact", Arg.Set(compact), "Compact training (slower, but conserves memory)");
     ("-gaussian-prior", Arg.Float (fun x -> Crf.invSigSquare := (1. /. (x *. x))), "Gaussian prior variance");
     ("-max-iters", Arg.Set_int (max_its), "Max number of iterations (default 1000)");
     ("-begin", Arg.Set(begin_encode), "Encode beginning of each phrase with begin label");
     ("-vp-train", Arg.Set(vp_train), " Use voted percepton training");
     ("-hybrid-train", Arg.Set(hybrid_train), " Use hybrid vp/crf training");
     ("-induce-features", Arg.Set(crf_induce), " Use Crf training with Feature Induction");
     ("-feature-set", Arg.Set_string(feature_set), "Feature set associated with task (POS, XG, XG-NOPOS, etc.)");
     ("-lexicon-dir", Arg.Set_string(lex_dir), "Directory containing lexicons");
     ("-intermediate", Arg.Set(intermediate), " Dump out intermediate modesl (every 50 iterations)");
     ("-seq-boundary", Arg.Set_string(seq_boundary), "Sequence boundary annotation")]
    (fun x -> ())
    "Usage: train-segment -input-dir <input dir> -filter <regexp> -model <model file> -lexicon-dir <lex dir>";
  let ts = MySgmlfront.gather_tagset !tagset in
  let paths = Array.map (fun x -> (!lex_dir)^x) (Sys.readdir !lex_dir) in
  let lex = TextFeatures.Lexicon.build_lexicons_txt paths in
  let inter_f = if !intermediate then Some !model_file else None in
    Printf.printf "\n Training from sgml";
    let ttype = if !hybrid_train then MySgmlfront.Hybrid else if !vp_train then MySgmlfront.Vp else if
      !crf_induce then MySgmlfront.CrfInduce else MySgmlfront.Crf in
    let user_crf = 
      MySgmlfront.train_from_sgml ~i_file_root:inter_f ~train_type:ttype ~compact:(!compact) ~max_its:(!max_its) 
       ~tr_filter:(Str.regexp (!tr_regstr)) 
       ~add_begin:(!begin_encode) !seq_boundary !tr_dir !feature_set ts (Some lex) in
  Crf.marshal_crf !model_file user_crf;;

main ();;
