(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(* Viterbi search performed to find most likely state sequence *)

exception InvalidAssignment

module SeqLab =
  struct
    type t = x option
    and x = {id: int; p: int; mutable score: float; mutable prev: t}

    (* let create id p = S(id, p, (ref (-.(max_float))), ref Null) *)
    let create id p = Some {id=id;p=p;score=(-. (max_float)); prev=None}

    let prev_seqlab = function
	Some({prev=pr}) -> pr | _ -> None

    let get_pos = function
	Some({p=p}) -> p | _ -> (-1)

    let get_label = function 
	Some({id=l}) -> l | _ -> (-1)

    let get_score = function 
	Some({score=sc}) -> sc | _ -> -. (max_float)

    let prevPos = function
	Some({prev=(Some({p=p}))}) -> p | _ -> (-1)

    let set_score sl s = match sl with Some sl -> sl.score <- s | None -> raise InvalidAssignment
			   
    let set_prev sl p = match sl with Some sl -> sl.prev <- p | None -> raise InvalidAssignment

    let copy s1 s2 =
      match s1,s2 with
	  Some(s1), Some({score=sc;prev=pr}) ->
	    s1.score <- sc; s1.prev <- pr
	| _ -> ()

    let clear = function
	Some(s) -> s.score <- (-.(max_float)); s.prev <- None | _ -> ()
	    
    let prevLabel = function
	Some({prev=(Some({id=l}))}) -> l | _ -> (-1)

    let equals s1 s2 =
      match s1,s2 with
	  Some({id=i1;p=p1;prev=pr1}), Some({id=i2;p=p2;prev=pr2}) ->
	    i1 = i2 && p1 = p2 && pr1 = pr2
	| _ -> false
  end
      

module Position =
  struct
    type t = SeqLab.t array
    let create bsize id pos = Array.init bsize (fun _ -> SeqLab.create id pos)
    let clear position =
      for i=0 to (Array.length position) - 1 do
	SeqLab.clear position.(i) done

    let size position = Array.length position

    let get position i = position.(i)

    let insert (position:t) i score prev =
      for k = (size position) - 1 downto i+1 do
	SeqLab.copy position.(k) position.(k-1) done;
      try
	SeqLab.set_score position.(i) score;
	SeqLab.set_prev position.(i) prev
      with InvalidAssignment -> ()

    let find_insert (position:t) ipos score prev =
      let halt = ref false in
      while not !halt && !ipos < size position do
	if score > (SeqLab.get_score (get position !ipos)) then (
	  insert position !ipos score prev;
	  halt := true);
	incr ipos
      done;
      !ipos
	
    let add_new (e1:t) (e2:t) score =
      let insertpos = ref 0 in
      let i = ref 0 in
	while !i < size e2 && !insertpos < size e1 do
	let sc = (SeqLab.get_score (get e2 !i)) +. score in
	  insertpos := find_insert e1 insertpos sc (get e2 !i);
	  incr i done

    let add (position:t) score =
      find_insert position (ref 0) score None

    let print (position:t) =
      for i=0 to (Array.length position) - 1 do
	let sol = get position i in
	Printf.printf "[%d %f i: %d y: %d]" i (SeqLab.get_score sol) (SeqLab.get_pos sol) (SeqLab.get_label sol);
      done;
      Printf.printf "\n"

    let get_scores (position:t) =
      Array.init (Array.length position) (fun i -> (SeqLab.get_score (get position i)))

  end

let viterbi_search ?(beamsize=1) data_seq crf calc_score =
  let num_Y = crf.Crf.nls in
  let mi = Array.make_matrix num_Y num_Y 0. in
  let ri = Array.make num_Y 0. in
  let cor_score = ref 0. in
  let final_seqlab = Position.create beamsize 0 0 in
  let winning_label = 
    Array.init num_Y 
      (fun yi -> Array.init (Array.length data_seq) 
	  (fun l -> Position.create (if l=0 then 1 else beamsize) yi l)) in
  for i=0 to (Array.length data_seq) - 1 do
    let (inst_features, size) = FeatureGen.get_compiled_vec data_seq.(i) in
    Crf.computeLogMi crf.Crf.parameters crf.Crf.nfs num_Y inst_features size mi ri false;
    for yi = 0 to num_Y - 1 do
      Position.clear winning_label.(yi).(i) done;
    for yi = 0 to num_Y - 1 do
      if i > 0 then 
	for yp = 0 to num_Y - 1 do
	  let vl = mi.(yp).(yi) +. ri.(yi) in
	  Position.add_new winning_label.(yi).(i) winning_label.(yp).(i-1) vl done
      else 
	ignore (Position.add winning_label.(yi).(i) ri.(yi)) done;
  done;
  Position.clear final_seqlab;
  for yi=0 to num_Y - 1 do
    Position.add_new final_seqlab winning_label.(yi).((Array.length data_seq) - 1) 0. done;
  final_seqlab

let get_top_n_label_seqs n data_seq crf =
  let (fsol:Position.t) = viterbi_search ~beamsize:n data_seq crf false in
  let n_best_seqs = Array.make_matrix n (Array.length data_seq) 0 in
    for i = 0 to n - 1 do
      let nth_best = ref (Position.get fsol i) in
	while not (!nth_best = None) do
	  n_best_seqs.(i).(SeqLab.get_pos !nth_best) <- (SeqLab.get_label !nth_best);
	  nth_best := (SeqLab.prev_seqlab !nth_best)
	done
    done;
    n_best_seqs, (Position.get_scores fsol)


(* this destructively modifies the data_seq crf labels to the viterbi solution *)
let best_label_seq data_seq crf =
  let (fsol:Position.t) = viterbi_search data_seq crf false in
  let (ybest:SeqLab.t ref) = ref (Position.get fsol 0) in
  let num_diffs = ref 0 in
    while not (!ybest = None) do
      let nlab = SeqLab.get_label !ybest in
      let inst = data_seq.(SeqLab.get_pos !ybest) in
      let origlab = inst.FeatureGen.crf_orig_label in
      let next = SeqLab.prev_seqlab !ybest in
	if (not (nlab = origlab)) && (not (next = None)) then incr num_diffs;
	FeatureGen.set_inst_label inst nlab;
	ybest := next
    done;
    (* return the sequence label accuracy *)
    (1. -. ((float_of_int !num_diffs) /. (float_of_int (Array.length data_seq))))
