(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** Implementation of Voted Perceptron Training 
  Algorithm outline:

  for i = 0 to Epochs
    for j = 0 to Instances
      params <- constraint - viterbi/observed
      avg_params <- params/i

*)

open Bigarray

module SSet = Set.Make(String)

let vp_inst_diff params constraints data_seq num_features num_labels =
  let tr_labs = Array.map (fun i -> FeatureGen.get_inst_label i) data_seq in
  let len = Array.length data_seq in
  let fl = (float_of_int len) in
    for i = 0 to len - 1 do 
      let (inst_features, size) = FeatureGen.get_compiled_vec data_seq.(i) in
      let label = FeatureGen.get_inst_label data_seq.(i) in
	for k=0 to size - 1 do
	  let (yprev,yp,f,vl) = inst_features.(k) in
	    if (label = yp && (((i > 0) && ((FeatureGen.get_inst_label data_seq.(i-1)) = yprev)) || yprev < 0)) then 
	      constraints.{f} <- constraints.{f} +. vl;
	done;
    done;
    (* viterbi destructively modifies label values *)
    ignore (Viterbi.best_label_seq data_seq {Crf.parameters=params; Crf.nfs=num_features; 
					     Crf.nls=num_labels});
    for i = 0 to len - 1 do
      let (inst_features, size) = FeatureGen.get_compiled_vec data_seq.(i) in
      let label = FeatureGen.get_inst_label data_seq.(i) in
	for k=0 to size - 1 do
	  let (yprev,yp,f,vl) = inst_features.(k) in
	    if (label = yp && (((i > 0) && ((FeatureGen.get_inst_label data_seq.(i-1)) = yprev)) || yprev < 0)) then 
	      constraints.{f} <- constraints.{f} -. vl;
	done;
    done;
    for i = 0 to num_features - 1 do
      params.{i} <- params.{i} +. constraints.{i} /. fl
    done;
    (* now reset origianl training labels - ick... *)
    Array.iteri (fun i inst -> FeatureGen.set_inst_label inst tr_labs.(i)) data_seq
    

let vp_train ?(la_fset_d=None) ?(compact=false) ?(max_its=10) num_features num_labels tr_data =
  let size = Array.length tr_data in
  let params = Array1.create float64 c_layout num_features
  and constraints = Array1.create float64 c_layout num_features (* just to re-use code *)
  and param_avg = Array1.create float64 c_layout num_features in
  let ctime = ref (Sys.time ()) in
    if not compact then (
      Gc.compact(); (* do a major GC before compiling data *)
      FeatureGen.compile_data tr_data);
    Array1.fill params 0.;
    Array1.fill param_avg 0.;
    ctime := Sys.time ();
    for i = 0 to max_its - 1 do
      for j = 0 to size - 1 do
	let data_seq = tr_data.(j) in
	let epoch = float_of_int ((i*size + 1) + j) in
	  Array1.fill constraints 0.0;
	  vp_inst_diff params constraints data_seq num_features num_labels;
	  for k = 0 to num_features - 1 do
	    param_avg.{k} <- param_avg.{k} +. ((params.{k} -. param_avg.{k}) /. epoch) done;
      done;
      (match la_fset_d with
	   Some (f, fset, la) ->
	     Crf.marshal_crf (f^"_at_"^(string_of_int i))
	     {Crf.crf={Crf.parameters=param_avg; Crf.nfs=num_features; Crf.nls=num_labels};
	      Crf.crf_fset=fset; Crf.crf_la=la; Crf.fset_mask=None}
	 | None -> ());
      Printf.printf "\nEpoch %d finished in %f seconds\n" i (Sys.time () -. !ctime); flush stdout;
      ctime := Sys.time ();
    done;
    {Crf.parameters=param_avg; Crf.nfs=num_features; Crf.nls = num_labels}

let scale_crf_params icrf =
  let params = icrf.Crf.parameters in
  let scale = ref 1.0 in
    for i = 0 to icrf.Crf.nfs - 1 do
      if params.{i} > !scale then
	scale := params.{i};
    done;
    if !scale > 1.0 then
      for i = 0 to icrf.Crf.nfs - 1 do
	params.{i} <- params.{i} /. !scale
      done

let get_nonzero_parameter_features i_crf fset =
  let params = i_crf.Crf.parameters in
  let fname_set = Hashtbl.fold (fun k v acc -> SSet.add (FeatureGen.unset k) acc) fset SSet.empty in
    SSet.fold
      (fun fname new_set ->
	 let all_fsets = Hashtbl.find_all fset (FeatureGen.set_f fname) in
	 let set_fname = FeatureGen.set_f fname in
	   if (List.for_all
		 (fun {FeatureGen.FeatureType.fdetail=feature_set} ->
		    (FeatureGen.FeatureSet.for_all (fun (_,_,fid,_) -> params.{fid} = 0.0) feature_set)) all_fsets) then
	     new_set
	   else FeatureGen.AgglFeatureSet.add set_fname new_set) fname_set FeatureGen.AgglFeatureSet.empty
  
(* this will now use Voted Perceptron as feature selection:
    1) Run VP for one iteration through the data
    2) Create a feature mask that only includes those features with associated parameters
       having a non-zero value
    3) Run CRF training using the feature mask *)
let hybrid_train  ?(compact=false) ?(vp_its=1) ?(crf_its=50) ?(la_fset_d=None) la orig_fset num_features num_labels tr_data =
  let i_crf = vp_train ~compact:compact ~max_its:vp_its num_features num_labels tr_data in
  let nf_aggl_set = get_nonzero_parameter_features i_crf orig_fset in
  let null_conj_table = Hashtbl.create 1 in
  let new_tr_data,nfa,nfset' = FeatureInduction.get_new_tr_data tr_data la nf_aggl_set null_conj_table in
  let new_num_features = FeatureGen.Alphabet.get_size nfa in
  let new_la_fset_d = match la_fset_d with None -> None | Some(f,fset,la) -> Some(f,nfset',la) in
    (Crf.train_crf ~compact:compact ~max_its:crf_its new_num_features ~la_fset_d:new_la_fset_d
       num_labels new_tr_data),Some(nfset',null_conj_table)
