(* Perform automatic cross-validation and scoring per token/position *)

module Int = struct type t = int let compare = compare end
module ISet = Set.Make(Int)

let permute_array arr =
  for i = (Array.length arr) - 1 downto 1 do
    let j = Random.int i in
    let t = arr.(i) in
      arr.(i) <- arr.(j);
      arr.(j) <- t done

let split_array arr ratio =
  let s = (Array.length arr) in
  let m = int_of_float ((float_of_int s) *. ratio) in
  let a1 = Array.init m (fun i -> arr.(i)) in
  let a2 = Array.init (s-m) (fun i -> arr.(i+m)) in
    (a1,a2)

let get_supported_features data =
  let f = ref ISet.empty in
    Array.iter
      (fun seq ->
	 Array.iter
	 (fun inst ->
	    let v,l = FeatureGen.get_compiled_vec inst in
	      Array.iter (fun (_,_,i,_) -> 
			    f := ISet.add i !f) v) seq) data;
    !f

(* go through each feature type and its details and simply remove features that didn't appear in the training data *)
(* do this by making a new copy of the test data so original is preserved *)
let relabel_test_data data sfs = 
  Array.map
    (fun seq ->
       Array.map
       (fun inst ->
	  let uv = FeatureGen.get_uncompiled_vec inst in
	  let new_uv = 
	    FeatureGen.FeatureTypeSet.fold 
	      (fun ftype top_acc ->
		 let nset = 
		   FeatureGen.FeatureSet.fold (fun (a,b,f,c) acc ->
						   if ISet.mem f sfs then
						     FeatureGen.FeatureSet.add (a,b,f,c) acc
						   else acc) ftype.FeatureGen.FeatureType.fdetail FeatureGen.FeatureSet.empty in
		   FeatureGen.FeatureTypeSet.add {ftype with FeatureGen.FeatureType.fdetail = nset} top_acc) 
	      uv FeatureGen.FeatureTypeSet.empty in
	    {inst with FeatureGen.crf_f_vec = (FeatureGen.UserVec new_uv)}) seq) data

let xvalidate ?(num_its=500) all_data numFeatures numLabels ratio numTrials =
  let all_scores = Array.init numTrials (fun i -> 0.0) in
    for i = 0 to numTrials - 1 do
      permute_array all_data;
      let tr,tst = split_array all_data ratio in
      let supported_features = get_supported_features tr in
      let new_tst = relabel_test_data tst supported_features in
      let crf_model = Crf.train_crf ~compact:true ~max_its:num_its numFeatures numLabels tr in
      let score = ref 0. in
      let total_lengths = ref 0. in
	Array.iter
	  (fun seq ->
	     let local = Viterbi.best_label_seq seq crf_model in
	     let fl = float_of_int (Array.length seq) in
	       Printf.printf "\tSeq[length=%d] score: %f\n" (Array.length seq) local;
	       score := !score +. (local *. fl);
	       total_lengths := !total_lengths +. fl) new_tst;
	let ls = !score /. !total_lengths in
	  all_scores.(i) <- ls;
	  Printf.printf "Iteration %d: Instance accuracy: %f\n" (i+1) ls
    done;
    let s = ref 0. 
    and s' = ref 0. 
    and var = ref 0. in
      Array.iter (fun ss -> s := !s +. ss) all_scores;
      Array.iter (fun ss -> s' := !s' +. (ss *. ss)) all_scores;
      let mean = !s /. (float_of_int numTrials) in
      let var = (!s' /. (float_of_int numTrials)) -. (mean ** 2.) in
      Printf.printf "\n Average Accuracy: %f\n" mean;
      Printf.printf "\n Standard Deviation (Variance): %f (%f)\n" (sqrt var) var

