(* Calcul de la distance d'édition entre deux mots *)
(* par programmation dynamique *)
(* $Id: levenstein.ml,v 1.1 2005/03/22 20:12:56 wellner Exp $ *)
(* Berke Durak *)
(* 20000817 *)
(* Modulification 20010528 *)

(* distance_d'edition d s t *)
(* calcule la distance d'edition entre les deux mots s et t *)
(* les mots sont représentés par des tableaux de type arbitraire *)
(* d est la fonction de cout des opérations de base (sur les lettres) *)
(* s et t sont du type 'a array *)
(* d doit etre du type 'a option -> 'a option -> int *)
(* m est la longueur du mot s, n est la longueur du mot t, et l'algorithme *)
(* s'éxécute en un temps d'ordre O(m.n) et bouffe de l'espace d'ordre O(m.n) *)

(*******************************************************************************)
(* Modified by Ben Wellner in 2004/2005 to gather back operations performed
   in computing the Edit distance.
   - Added trace-back for gathering operations
   - Removed the "example" code here, in particular code that requires the preprocessor *)

(* NOTES: Use of functors makes this slighly slower than it could be otherwise, but
   the flexibility seems worth it *)

let high_cost = 500.0

module type COSTS =
  sig
	type t
	val min : t -> t -> t
	val sum : t -> t -> t
	val zero : t
  end

module type WORDS =
  sig
	module K:COSTS
	type t
	type letter
	val length : t -> int
	val get : t -> int -> letter
	val distance : letter option -> letter option -> K.t
  end
	  
module Levenstein(A:WORDS) =
  struct
	module K = A.K
	let distance s t =
	  let m = A.length s and
		  n = A.length t in
	  let a = Array.make ((m + 1) * (n + 1)) K.zero in
	  let koy i j x =
		begin
		  assert (0 <= i && i <= m && 0 <= j && j <= n);
		  a.((n + 1) * i + j) <- x
		end
	  and al i j =
		begin
		  assert (0 <= i && i <= m && 0 <= j && j <= n);
		  a.((n + 1) * i + j)
		end
	  in
	  begin
		koy 0 0 K.zero;
		for i = 1 to m do
		  koy i 0 (K.sum (al (i - 1) 0) (A.distance None (Some(A.get s (i - 1)))))
		done;
		for j = 1 to n do
		  koy 0 j (K.sum (al 0 (j - 1)) (A.distance (Some(A.get t (j - 1))) None))
		done;
		for i = 0 to m - 1 do
		  for j = 0 to n - 1 do
			let x1 = K.sum (al i       (j + 1)) (A.distance (Some (A.get s i)) None)
			and x2 = K.sum (al (i + 1) j      ) (A.distance (Some (A.get t j)) None)
			and x3 = K.sum (al i       j      )       (A.distance (Some (A.get s i)) (Some (A.get t j)))
			in
			let x = K.min x1 (K.min x2 x3) in
			koy (i + 1) (j + 1) x
		  done
		done;
		(al m n, a)
	  end
  end

let row y0 x0 w u1 u2 =
  let m = String.length u1 in
  let r = ref [] in
  for i = 0 to m - 1 do
	r := (y0, x0 +. (float_of_int i) *. w, 0.0, u1.[i])::!r;
	r := (y0, x0 +. (float_of_int i) *. w, 1.0, u2.[i])::!r
  done;
  !r

let qwerty_description = List.concat [
  row 0.0 0.0 1.7461
	"`1234567890-="
	"~!@#$%^&*()_+";
  [1.5,1.0,0.0,'\t'];
  row 1.5 2.8 1.7461 "qwertyuiop[]" "QWERTYUIOP{}";
  row 3.0 3.4 1.7461 "asdfghjkl;'\\" "ASDFGHJKL:\"|";
  row 4.5 2.8 1.7461 "<zxcvbnm,./" ">ZXCVBNM<>?";
  [6.0,12.5,0.0,' ']
]

let qwerty_dummy = (20.0,4.0,0.5)

let qwerty_map =
  let a = Array.make 256 qwerty_dummy in
  List.iter (fun (x,y,z,c) -> a.(Char.code c) <- (x,y,z)) qwerty_description;
  a

let euclidian_distance_squared (x1,y1,z1) (x2,y2,z2) =
  let f a1 a2 = (a1 -. a2) *. (a1 -. a2) in
  sqrt ((f x1 x2) +. (f y1 y2) +. (f z1 z2))

let qwerty_insertion_cost = 7.0
let qwerty_deletion_cost = 5.0

let qwerty_distance c1 c2 =
  let p c = qwerty_map.(Char.code c) in
  match (c1,c2) with
	None,None -> 0.0
  |	Some(_),None -> qwerty_deletion_cost
  |	None,Some(_) -> qwerty_insertion_cost
  |	Some(c1),Some(c2) ->
	  euclidian_distance_squared (p c1) (p c2)

let insertion_cost c =
  match c with
      _ -> 0.75

let deletion_cost c = 
  match c with
      _ -> 0.75

(* a pretty standard distance metric - case difference substitutions have less cost then
   arbitrary substitutions (i.e. h=>H is less costly then h=>g.  This is a good baseline
   alignment for learning edit-distance models with CRFs or HMMs.  In particular, it aligns
   acronyms albeit not perfectly *)
let basic_distance c1 c2 = 
  match (c1,c2) with
      None,None -> 0.0
    | Some(c),None -> deletion_cost c
    | None,Some(c) -> insertion_cost c
    | Some(c1),Some(c2) -> 
	let code1 = Char.code c1 
	and code2 = Char.code c2 in
	  if c1 = c2 then 0.0 
	  else if code1 > 64 && code2 > 64 && code1 < 123 && code2 < 123 && 
	    (((abs code1 - code2) mod 32) = 0) then 0.25
	  else 1.0

let adjustable_distance ht c1 c2 = 
  match (c1,c2) with
      None,None -> 0.0
    | Some(c),None -> Hashtbl.find ht (Some(c),None)
    | None,Some(c) -> Hashtbl.find ht (None,Some(c))
    | Some(c1),Some(c2) -> 
	if c1 = c2 then 0.0 
	else Hashtbl.find ht (Some(c1),Some(c2))

module Float_cost =
  struct
	type t = float
	let min = min
	let sum = (+.)
	let zero = 0.0
  end

module String_words =
  struct
	module K = Float_cost
	type t = string
	type letter = char
	let length = String.length
	let get = String.get
	let distance = basic_distance
	(* let distance = adjustable_distance (Hashtbl.create 10) (* qwerty_distance *) *)
  end
	
module Levenstein_string = Levenstein(String_words)

(* Added by BRW - just convert array matrix into a real matrix *)
let matrixify a d1 d2 =
  assert ((Array.length a) = (d1 * d2));
  let m = Array.make_matrix d1 d2 0. in
    for i=0 to d1 - 1 do
      for j=0 to d2 - 1 do
	m.(i).(j) <- a.((i*d2) + j) done done;
    m

(* Added function by BRW *)
let get_operations (cost_mat: float array array) s t =
  let l1 = String.length s
  and l2 = String.length t in
  let ops = ref [] in
  let maxiters = l1 + l2 in
  let iters = ref 0 in
    assert (((Array.length cost_mat.(0)) = (l2+1)) && ((Array.length cost_mat) = (l1+1)));
    let curcost = ref cost_mat.(l1).(l2)
    and i = ref l1
    and j = ref l2 in
      while (!curcost > 0. || !i > 0 || !j > 0) && !iters < maxiters do
	incr iters;
	let addc = if !i > 0 && !j > (-1) then cost_mat.(!i - 1).(!j) else high_cost
	and delc = if !j > 0 && !i > (-1) then cost_mat.(!i).(!j - 1) else high_cost
	and subc = if !j > 0 && !i > 0 then cost_mat.(!i - 1).(!j - 1) else high_cost in
	let all = [(addc,(!i-1)+(!j),"add");(delc,(!i)+(!j-1),"del");(subc,(!i-1)+(!j-1),"sub")] in
	  let sorted = 
	    (List.sort 
	       (fun (a1,a2,_) (b1,b2,_) -> 
		  let cc = (int_of_float (a1 *. high_cost)) - (int_of_float (b1 *. high_cost)) in
		    if cc = 0 then a2 - b2
		    else cc) all) in
	let (ncost,_,op) = 
	  List.hd sorted in
	  let actop = 
	    match op with 
		"add" when !i > 0 -> i := !i - 1; ("add",(Some(String.get s !i)),None)
	      | "del" when !j > 0 -> j := !j - 1; ("del",None,(Some(String.get t !j)))
	      | "sub" when !i > 0 && !j > 0 -> i := !i - 1; j := !j - 1; ("sub",(Some(String.get s !i)),(Some(String.get t !j)))
	      | _ -> failwith "Invalid match" in
	    curcost := ncost;
	    ops := actop :: !ops
      done;
      !ops

let string_dist s1 s2 = fst (Levenstein_string.distance s1 s2)

let string_dist_ops s1 s2 =
  let l1 = String.length s1
  and l2 = String.length s2 in
  let d,a = Levenstein_string.distance s1 s2 in
  let m = matrixify a (l1+1) (l2+1) in
    (d, (get_operations m s1 s2))

