(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** This module contains different array implementations for arrays of floats or ints. 
   The current implementations use the Bigarray module.  Most of the modules in this
   package that use arrays are paramterizable by array type. *)


module type ArraySig  =
  sig
    type t
    type arr
    val create : int -> arr
    val size_of : arr -> int
    val get : arr -> int -> t
    val set : arr -> int -> t -> unit
    val fill : arr -> t -> unit
    val make : int -> t -> arr
    val sub : arr -> int -> int -> arr
    val max : arr -> t
    val print : arr -> unit
  end

module type I_ArraySig = ArraySig with type t = int

module type F_ArraySig = ArraySig with type t = float

module FBigArrayImpl : F_ArraySig =
  struct
    open Bigarray
    type t = float
    type arr = (float, Bigarray.float64_elt, Bigarray.c_layout) Bigarray.Array1.t

    let create size = Array1.create float64 c_layout size

    let size_of array = Array1.dim array
    let get array ind = array.{ind}
    let set array ind v = array.{ind} <- v
    let fill array v = Array1.fill array v
    let sub array ind len = Array1.sub array ind len
    let make size v = 
      let a = create size in
      fill a v;
      a
    let max array =
      let cm = ref (get array 0) in
      for i=0 to (size_of array) - 1 do
	if (get array i) > !cm then
	  cm := (get array i);
      done;
      !cm

    let print array =
      print_string "\n";
      for i=0 to (size_of array) - 1 do
	print_string (" " ^ (string_of_float (get array i)))
      done

  end    

module IBigArrayImpl : I_ArraySig =
  struct
    open Bigarray
    type t = int
    type arr = (int, Bigarray.int_elt, Bigarray.c_layout) Bigarray.Array1.t

    let create size : arr = Array1.create int c_layout size
    let size_of array : int = Array1.dim array
    let get array ind : t = array.{ind}
    let set array ind v : unit = array.{ind} <- v
    let fill array v : unit = Array1.fill array v
    let sub array ind len = Array1.sub array ind len
    let make size v = 
      let a = create size in
      fill a v;
      a
    let max array =
      let cm = ref (get array 0) in
      for i=0 to (size_of array) - 1 do
	if (get array i) > !cm then
	  cm := (get array i);
      done;
      !cm

    let print array =
      print_string "\n";
      for i=0 to (size_of array) - 1 do
	print_string (" " ^ (string_of_int (get array i)))
      done
  end    
