(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(** This module contains instance types as well as modules/functions for manipulating
   instances.  Many of the tools here will be utilized for mapping user-level instances
   which will likely have symbolic form, into numeric instances used by the machine
   learning code *)


module Int =
struct
  type t = int
  let compare x y = x - y
end

module ISet = Set.Make (Int)

module FType =
struct
  type t = string * float
  let compare (s,_) (t,_)  = if s < t then 1 else if s > t then -1 else 0
end

module FSet = Set.Make(FType)


type 'a raw_instance = {mutable label: int;
			f_vector: Vectors.sparse_vector;
			source: 'a} (* source represents information to save of same type as info *)

(* user_instance  is a parameterized type that may vary from domain to domain *)
(* really used to build feature vectors for the above raw_instance type *)
type 'a user_instance = {lab: string;
			 mutable features: FSet.t;
			 info: 'a}

type alphabet = {map: (string,int) Hashtbl.t;
		 mutable size: int;
		 counts: (int,int) Hashtbl.t}

type inv_alphabet = {inv_map: (int,string) Hashtbl.t;
		     mutable inv_size: int}

type 'a instance_set = {mutable r_insts: 'a raw_instance list;
			falpha: alphabet;
			lalpha: alphabet;
			mutable fixed: bool}

let create_alphabet () = {map=(Hashtbl.create 10); size=0; counts=(Hashtbl.create 10)}

let copy_inst_list li =
  List.map (fun inst -> {inst with features=(FSet.empty)}) li

let rrr = Str.regexp ":"
let rr = Str.regexp " "
let ret = Str.regexp "\n"

let print_user_inst ?(add_str="") i ostr =
  Printf.fprintf ostr "%s" i.lab;
  FSet.iter (fun (s,v) -> 
	       if v > 0.0 then 
		 Printf.fprintf ostr " %s:%f" 
		   (Str.global_replace ret "" 
		      (Str.global_replace rr "" (Str.global_replace rrr "_" s))) v) i.features;
  Printf.fprintf ostr "%s\n" add_str

let permute_array arr =
  for i = (Array.length arr) - 1 downto 1 do
    let j = Random.int i in
    let t = arr.(i) in
      arr.(i) <- arr.(j);
      arr.(j) <- t done;
  arr

let split_array arr ratio =
  let s = (Array.length arr) in
  let m = int_of_float ((float_of_int s) *. ratio) in
  let a1 = Array.init m (fun i -> arr.(i)) in
  let a2 = Array.init (s-m) (fun i -> arr.(i+m)) in
    (a1,a2)

let split_set ?(rand=true) ?(f1=false) ?(f2=true) set ratio =
  let ar = Array.of_list set.r_insts in
  let a1,a2 = split_array (if rand then (permute_array ar) else ar) ratio in
  let tr_set = {{set with r_insts=(Array.to_list a1)} with fixed=f1} in
  let tst_set = {{set with r_insts=(Array.to_list a2)} with fixed=f2} in
    (tr_set,tst_set)

(* creation functions *)
let create_raw int sp_vec src = {label=int; f_vector=sp_vec; source=src}

let create_instance_set 
  ?(fa={map = Hashtbl.create 10; size = 0; counts=Hashtbl.create 10}) 
  ?(la={map = Hashtbl.create 2; size = 0; counts=Hashtbl.create 10}) fix =
  {r_insts = []; 
   falpha = fa;
   lalpha = la;
   fixed = fix}

let create_user_inst labl data =
  {lab=labl; features=FSet.empty; info=data}

(* adder functions *)
let ts_add (ts: 'a instance_set) (ri: 'a raw_instance) =
  ts.r_insts <- ri :: ts.r_insts

let add_feature u_inst feature value =
  u_inst.features <- FSet.add (feature,value) u_inst.features
 (*  u_inst.features <- ((feature,value) :: u_inst.features) *)

(** This takes a list of strings and makes an atomic feature out of them by concatenation *)
let add_feature_l u_inst f_list value =
  let feature = "("^(String.concat "," f_list)^")" in
    add_feature u_inst feature value

(* accessor functions *)
let get_f_vector ri = ri.f_vector
let get_label ri = ri.label

let update_alphabet (fix:bool) (alphabet:alphabet) (item:string) =
  let ind = 
    try 
      Hashtbl.find alphabet.map item
    with Not_found ->
      if not fix then (
	let s = alphabet.size in
	  Hashtbl.add alphabet.map item s;
	  alphabet.size <- s + 1;
	  s )
      else (* with fixed alphabet, if not found, return -1 *)
	-1 in
    if not fix then (
      let cc = try Hashtbl.find alphabet.counts ind with Not_found -> 1 in
	Hashtbl.replace alphabet.counts ind (cc+1) );
    ind

(*
let rec get_feature_set u_insts cur_fs =
  match u_insts with 
    [] -> cur_fs
  | i :: is ->
      let new_fs = 
	List.fold_right FSet.add (List.map (fun x -> match x with (s,_) -> s) i.features) cur_fs in
      get_candidate_features is new_fs
*)
(* creates a raw_instance out of user_inst and updates alphabets if needed *)

let rec intern_features ?(inc=1) features f_alphabet fix =
  match features with
    [] -> []
  | (f,v) :: fs ->
      let ind = update_alphabet fix f_alphabet f in
	if ind < 0 then intern_features ~inc:inc fs f_alphabet fix
	else (ind+inc,v) :: intern_features ~inc:inc fs f_alphabet fix 

(* get a single raw instance from a user instance, provided the alphabets *)
let single_unlabeled_inst (inst:'a user_instance) falph =
  let sfeatures = intern_features (FSet.elements inst.features) falph true in
  let (inds,vals) = List.split sfeatures in
  {label = (-1); f_vector = (Vectors.l_create (0 :: inds) (1. :: vals)); source=inst.info}

let intern_inst ?(add_dummy=true) (src_fn: 'a -> 'b) 
    (inst: 'a user_instance) (cur_insts: 'a instance_set) : 'a raw_instance =
  let f_alphabet = cur_insts.falpha 
  and l_alphabet = cur_insts.lalpha
  and fix = cur_insts.fixed in
  let sfeatures = intern_features ~inc:(if add_dummy then 1 else 0) (FSet.elements inst.features) f_alphabet fix in
  let inds,vals = List.split sfeatures in
  let lab_ind = update_alphabet fix l_alphabet inst.lab in
  (* add in dummy feature directly in sparse vector *)
    if add_dummy then
      {label=lab_ind; f_vector=(Vectors.l_create (0 :: inds) (1. :: vals)); source=(src_fn inst.info)}
    else {label=lab_ind; f_vector=(Vectors.l_create inds vals); source=(src_fn inst.info)}

(* an optional function src_fn allows the user to provide control over what source information
   is saved in the raw_instance record - default is to include the user instance *)
let intern_insts_inc ?(add_dummies=true) ?(src_fn)
    (u_insts: 'a user_instance list) 
    (cur_insts: 'a instance_set) =
  match src_fn with
    None ->
      List.iter (fun x -> ts_add cur_insts 
		   (intern_inst ~add_dummy:add_dummies (fun y -> y) x cur_insts)) u_insts
  | Some fn ->
      List.iter (fun x -> ts_add cur_insts (intern_inst ~add_dummy:add_dummies fn x cur_insts)) u_insts

let filter_vec inds vls imap =
  let ni = ref [] in
  let nv = ref [] in
    for i = 0 to (Array.length inds) - 1 do
      try 
	let ai = inds.(i) - 1 in (* remember shift of 1 due to dummy feature *)
	  if ai >= 0 then
	    ni := ((Hashtbl.find imap ai) + 1) :: !ni (* and back again *)
	  else 
	    ni := 0 :: !ni;
	  nv := vls.(i) :: !nv
      with Not_found -> ()
    done;
    ((List.rev !ni), (List.rev !nv))

(* This is a fairly crude way of pruning features that occur infrequently *)
let commonify_instances (insts: 'a instance_set) mincnt =
  let cnts = insts.falpha.counts in
  let fmap = insts.falpha.map in
  let newmap = Hashtbl.create 10 in
  let imap = Hashtbl.create 10 in
  let count = ref 0 in
  let filtered = ref 0 in
    Hashtbl.iter 
      (fun k v -> 
	 let vl = Hashtbl.find cnts v in 
	   if vl >= mincnt then  (
	     Hashtbl.add newmap k !count;
	     Hashtbl.add imap v !count;
	     incr count )
	   else incr filtered) fmap;
    Printf.printf "Removed %d features based on infrequent counts\n" !filtered;
    Printf.printf "   %d features remaining\n" !count;
      { r_insts = 
	List.rev_map
	  (fun (rinst: 'a raw_instance) ->
	     let ni,nv = filter_vec rinst.f_vector.Vectors.indices rinst.f_vector.Vectors.values imap in
	       {rinst with f_vector=(Vectors.l_create ni nv)}) insts.r_insts;
      falpha = {map=newmap; size = !count; counts=cnts};
      lalpha = insts.lalpha;
      fixed = insts.fixed }


(** This function will take an instance list interpreted as a sequence of instances (for a CRF)
    and will ensure that the instance ordering is maintained as opposed to the above method of
    incremental addition.  This approach is also faster and perhaps should be used universally *)

let intern_inst_seq ?(src_fn) (u_insts: 'a user_instance list) (cur_insts: 'a instance_set) =
  match src_fn with
    None ->
      let raw_insts = List.map (fun x -> intern_inst (fun y -> y) x cur_insts) u_insts in
      cur_insts.r_insts <- cur_insts.r_insts @ raw_insts
  | Some fn ->
      let raw_insts = List.map (fun x -> intern_inst fn x cur_insts) u_insts in
      cur_insts.r_insts <- cur_insts.r_insts @ raw_insts
    
let invert_alphabet alphabet =
  let inv_ht = Hashtbl.create alphabet.size in
  let a_ht = alphabet.map in
  Hashtbl.iter
    (fun k v ->
      Hashtbl.add inv_ht v k)
    a_ht;
  {inv_map=inv_ht; inv_size=alphabet.size}
      
