(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)


(** This module contains the core LBFGS optimization code adapted from the C-implementation
    which is in the public domain.  Below is the relevent reference.
       for     unconstrained optimization problems
       alg     limited memory BFGS method
       by      J. Nocedal
       contact nocedal at eecs.nwu.edu
       ref     D. C. Liu and J. Nocedal, ``On the limited memory BFGS method for
       ,       large scale optimization methods'' Mathematical Programming 45
       ,       (1989), pp. 503-528.
       ,       (Postscript file of this paper is available via anonymous ftp
       ,       to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_um.)  *)

(** This module type provides an interface for the Opt module which contains
    the LBFGS optimization code including the line search code in Mcsrch *)

open Bigarray

module type OptSig =
  sig
    type ibarray = (int, Bigarray.int_elt, Bigarray.c_layout) Bigarray.Array1.t
    type rbarray = (float, Bigarray.float64_elt, Bigarray.c_layout) Bigarray.Array1.t

    module type McsrchSig = 
      sig
	val init : unit -> unit
	val mcsrch : int -> rbarray -> float ref -> rbarray -> rbarray
	  -> int -> float ref -> float -> float -> int -> int ref -> int ref -> rbarray -> unit
      end
    module Mcsrch : McsrchSig
    module type LbfgsSig = 
      sig
	val init : unit -> unit
	val lbfgs : int -> int -> rbarray -> float ref -> rbarray -> bool ref -> rbarray -> int array
	  -> float -> float -> int ref -> unit
      end
    module type TestSig =
      sig
	val test : unit -> unit
      end
    module Test : TestSig
    module Lbfgs : LbfgsSig
  end

(** This module contains all the code relevant to the optimization including the line
    search code.  The general mechanism here uses reverse communication, so the function
    lbfgs should be called repeatedly by the application, providing new function and gradient
    values for each call based on the new values returned by the previous lbfgs iteration. *)
module Opt =
  struct

    type ibarray = (int, Bigarray.int_elt, Bigarray.c_layout) Bigarray.Array1.t
    type rbarray = (float, Bigarray.float64_elt, Bigarray.c_layout) Bigarray.Array1.t

    let l_gtol = ref 0.99
    and l_stpmin = 1e-10
    and l_stpmax = 1e20

    module type McsrchSig = 
      sig
	val init : unit -> unit
	val mcsrch : int -> rbarray -> float ref -> rbarray -> rbarray
	  -> int -> float ref -> float -> float -> int -> int ref -> int ref -> rbarray -> unit
      end

    (** This is the module that implements the line search. *)
    module Mcsrch : McsrchSig =
    struct

      let infoc = ref 0
      and j = ref 0
      and dg = ref 0.
      and dgm = ref 0.
      and dginit = ref 0.
      and dgtest = ref 0.
      and dgx = ref 0.
      and dgxm = ref 0.
      and dgy = ref 0.
      and dgym = ref 0.
      and finit = ref 0.
      and ftest1 = ref 0.
      and fm = ref 0.
      and fx = ref 0.
      and fxm = ref 0.
      and fy = ref 0.
      and fym = ref 0.
      and p5 = ref 0.
      and p66 = ref 0.
      and stx = ref 0.
      and sty = ref 0.
      and stmin = ref 0.
      and stmax = ref 0.
      and width = ref 0.
      and width1 = ref 0.
      and xtrapf = ref 0.
      and brackt = ref true
      and stage1 = ref false 
	
      exception Return

      let init () =
	infoc := 0; j := 0; dg := 0.; dgm := 0.; dginit := 0.; dgtest := 0.; dgx := 0.; dgxm := 0.;
	dgy := 0.; dgym := 0.; finit := 0.; fm := 0.; fx := 0.; fxm := 0.; fy := 0.; fym := 0.; p5 := 0.;
	p66 := 0.; stx := 0.; sty := 0.; stmin := 0.; stmax := 0.; width := 0.; width1 := 0.; xtrapf := 0.;
	brackt := true; stage1 := false
	
      let sqr x = x *. x
		      
      let max3 x y z =  if x < y then if y < z then z else y else if x < z then z else x
	
      let mcsrch n (x:rbarray) f (g:rbarray) (s:rbarray) is0 stp ftol xtol maxfev info nfev (wa:rbarray) =
	
	let mcstep stx fx dx sty fy dy stp fp dp brackt stpmin stpmax info =
	  let bound = ref true
	  and gamma = ref 0.
	  and p = ref 0.
	  and q = ref 0.
	  and r = ref 0.
	  and s = ref 0.
	  and sgnd = ref 0.
	  and stpc = ref 0.
	  and stpf = ref 0.
	  and stpq = ref 0.
	  and theta = ref 0.
	  in
	    info := 0;
	    try 
	      if (  (!brackt && ( (!stp <= (min !stx !sty)) || (!stp >= (max !stx !sty))) )
		    || (!dx *. (!stp -. !stx)) >= 0. || !stpmax < !stpmin) then (
		print_string "\n left mcstep early...";
		raise Return );
	      sgnd := !dp *. (!dx /. (abs_float !dx));
	      if (!fp > !fx) then (
		info := 1;
		bound := true;
		theta := (3. *. (!fx -. !fp) /. (!stp -. !stx)) +. !dx +. !dp;
		s := max3 (abs_float !theta) (abs_float !dx) (abs_float !dp);
		gamma := !s *. (sqrt ((sqr (!theta /. !s)) -. (!dx /. !s) *. (!dp /. !s)));
		if (!stp < !stx) then gamma := -. (!gamma);
		p := (!gamma -. !dx) +. !theta;
		q := ( (!gamma -. !dx) +. !gamma) +. !dp;
		r := !p /. !q;
		stpc := !stx +. !r *. (!stp -. !stx);
		stpq := !stx +. ( ( !dx /. ((!fx -. !fp) /. (!stp -. !stx) +. !dx)) /. 2.) *. (!stp -. !stx);
		if (abs_float (!stpc -. !stx)) < (abs_float (!stpq -. !stx)) then
		  stpf := !stpc
		else
		  stpf := !stpc +. (!stpq -. !stpc) /. 2.;
		brackt := true; )
	      else if !sgnd < 0. then (
		info := 2;
		bound := false;
		theta := 3. *. (!fx -. !fp) /. (!stp -. !stx) +. !dx +. !dp;
		s := max3 (abs_float !theta) (abs_float !dx) (abs_float !dp);
		gamma := !s *. sqrt(sqr(!theta /. !s) -. (!dx /. !s) *. (!dp /. !s));
		if (!stp > !stx) then gamma := -. (!gamma);
		p := (!gamma -. !dp) +. !theta;
		q := ( (!gamma -. !dp) +. !gamma) +. !dx;
		r := !p /. !q;
		stpc := !stp +. !r *. (!stx -. !stp);
		stpq := !stp +. (!dp /. (!dp -. !dx)) *. (!stx -. !stp);
		if (abs_float (!stpc -. !stp)) > (abs_float (!stpq -. !stp)) then
		  stpf := !stpc
		else
		  stpf := !stpq;
		brackt := true )
	      else if (abs_float !dp) < (abs_float !dx) then (
		info := 3;
		bound := true;
		theta := 3. *. (!fx -. !fp) /. (!stp -. !stx) +. !dx +. !dp;
		s := max3 (abs_float !theta) (abs_float !dx) (abs_float !dp);
		gamma := !s *. sqrt(sqr(!theta /. !s) -. (!dx /. !s) *. (!dp /. !s));
		if (!stp > !stx) then gamma := -. (!gamma);
		p := (!gamma -. !dp) +. !theta;
		q := (!gamma +. (!dx -. !dp)) +. !gamma;
		r := !p /. !q;
		if (!r < 0. && !gamma <> 0.) then 
		  stpc := !stp +. !r *. (!stx -. !stp)
		else if !stp > !stx then
		  stpc := !stpmax
		else
		  stpc := !stpmin;
		stpq := !stp +. (!dp /. (!dp -. !dx)) *. (!stx -. !stp);
		if !brackt then
		  if (abs_float (!stp -. !stpc)) < (abs_float (!stp -. !stpq)) then
		    stpf := !stpc
		  else
		    stpf := !stpq
		else
		  if (abs_float (!stp -. !stpc)) > (abs_float (!stp -. !stpq)) then
		    stpf := !stpc
		  else
		    stpf := !stpq; )
	      else (
		info := 4;
		bound := false;
		if !brackt then (
		  theta := 3. *. (!fp -. !fy) /. (!sty -. !stp) +. !dy +. !dp;
		  s := max3 (abs_float !theta) (abs_float !dy) (abs_float !dp);
		  gamma := !s *. sqrt(sqr(!theta /. !s) -. (!dy /. !s) *. (!dp /. !s));
		  if (!stp > !sty) then gamma := -. (!gamma);
		  p := !gamma -. !dp +. !theta;
		  q := ( (!gamma -. !dp) +. !gamma) +. !dy;
		  r := !p /. !q;
		  stpc := !stp +. !r *. (!sty -. !stp);
		  stpf := !stpc )
		else if !stp > !stx then
		  stpf := !stpmax
		else 
		  stpf := !stpmin; );
	      if !fp > !fx then (
		sty := !stp;
		fy := !fp;
		dy := !dp )
	      else (
		if (!sgnd < 0.) then (
		  sty := !stx;
		  fy := !fx;
		  dy := !dx );
		stx := !stp;
		fx := !fp;
		dx := !dp );
	      stpf := min !stpmax !stpf;
	      stpf := max !stpmin !stpf;
	      stp := !stpf;
	      if (!brackt && !bound) then 
		if (!sty > !stx) then
		  stp := min (!stx +. 0.66 *. (!sty -. !stx)) !stp
		else
		  stp := max (!stx +. 0.66 *. (!sty -. !stx)) !stp;
	      raise Return
	    with Return -> () in
	  p5 := 0.5;
	  p66 := 0.55;
	  xtrapf := 4.;
	  try 
	    if (!info <> (-1)) then (
	      infoc := 1;
	      if (n <= 0 || !stp <= 0. || ftol < 0. || !l_gtol < 0. || xtol < 0. || l_stpmin < 0.
		  || l_stpmax < l_stpmin || maxfev < 0) then 
		raise Return;
	      dginit := 0.;
	      for j=1 to n do
		dginit := !dginit +. g.{j-1} *. s.{is0 + j - 1} done;
	      if !dginit >= 0. then (
		print_string "The seach direction is not a descent direction";
		raise Return );
	      
	      brackt := false;
	      stage1 := true;
	      nfev := 0;
	      finit := !f;
	      dgtest := ftol *. !dginit;
	      width := l_stpmax -. l_stpmin;
	      for j=1 to n do
		wa.{j-1} <- x.{j-1} done;
	      stx := 0.;
	      fx := !finit;
	      dgx := !dginit;
	      sty := 0.;
	      fy := !finit;
	      dgy := !dginit );
	    while true do
	      if (!info) <> -1 then (
		if !brackt then (
                  stmin := min !stx !sty;
                  stmax := max !stx !sty )
		else (
		  stmin := !stx;
		  stmax := !stp +. !xtrapf *. (!stp -. !stx) );
		stp := max !stp l_stpmin;
		stp := min !stp l_stpmax;
		if (  (!brackt && (!stp <= !stmin || !stp >= !stmax))
		      || !nfev >= (maxfev - 1)
		      || !infoc = 0
		      || (!brackt && ((!stmax -. !stmin) <= (xtol *. !stmax))) ) then (
		  stp := !stx );
		for j=1 to n do
		  let sel = s.{is0 + j - 1} in
		  let fwa = wa.{j-1} in
		  let toset = (fwa +. !stp *. sel) in
		    x.{j-1} <- toset done;
		info := -1;
		raise Return);
	      info := 0;
	      nfev := !nfev + 1;
	      dg := 0.;
	      for j=1 to n do
		dg := !dg +. g.{j-1} *. s.{is0 + j - 1} done;
	      (* Printf.printf "\n dg: %f" !dg; *)
	      ftest1 := !finit +. !stp *. !dgtest;
	      if ( (!brackt && ((!stp <= !stmin) || (!stp >= !stmax)))
		   || (!infoc = 0) ) then ( 
		info := 6 );
	      
	      if ( !stp == l_stpmax && !f <= !ftest1 && !dg <= !dgtest ) then info := 5;
	      if ( !stp == l_stpmin && (!f > !ftest1 || !dg >= !dgtest) ) then info := 4;
	      if ( !nfev >= maxfev ) then info := 3;
	      if ( !brackt && ((!stmax -. !stmin) <= (xtol *. !stmax)) ) then info := 2;
	      if ( !f <= !ftest1 && ((abs_float !dg) <= (!l_gtol *. (-. !dginit))) ) then info := 1;
	      
	      if !info <> 0 then
		raise Return;
	      
	      if (!stage1 && !f <= !ftest1 && !dg >= ((min ftol !l_gtol) *. !dginit)) then stage1 := false;
	      
	      if (!stage1 && !f <= !fx && !f > !ftest1) then (
		fm := !f -. !stp *. !dgtest;
		fxm := !fx -. !stx *. !dgtest;
		fym := !fy -. !sty *. !dgtest;
		dgm := !dg -. !dgtest;
		dgxm := !dgx -. !dgtest;
		dgym := !dgy -. !dgtest;
		mcstep stx fxm dgxm sty fym dgym stp fm dgm brackt stmin stmax infoc;
		fx := !fxm +. !stx *. !dgtest;
		fy := !fym +. !sty *. !dgtest;
		dgx := !dgxm +. !dgtest;
		dgy := !dgym +. !dgtest )
	      else (
		mcstep stx fx dgx sty fy dgy stp f dg brackt stmin stmax infoc );
	      
	      (* Printf.printf "\nstep length from mcstep: %f" !stp; *)
	      
	      if !brackt then (
		if ((abs_float (!sty -. !stx)) >= (!p66 *. !width1)) then
		  stp := !stx +. !p5 *. (!sty -. !stx);
		width1 := !width;
		width := abs_float (!sty -. !stx) )
	    done
	  with Return -> ()
    end
      
    module type LbfgsSig = 
    sig
      val init : unit -> unit
      val lbfgs : int -> int -> rbarray -> float ref -> rbarray -> bool ref -> rbarray -> int array
	-> float -> float -> int ref -> unit
    end
      
    (** This module implements the main lbfgs optimization code *)
    module Lbfgs : LbfgsSig = 
    struct
      let gnorm = ref 0.
      and stp1 = ref 0.
      and ftol = ref 0.
      and stp = ref 0.
      and ys = ref 0.
      and yy = ref 0.
      and sq = ref 0.
      and yr = ref 0.
      and beta = ref 0.
      and xnorm = ref 0.
      and iter = ref 0
      and nfun = ref 0
      and point = ref 0
      and ispt = ref 0
      and iypt = ref 0
      and maxfev = ref 0
      and bound = ref 0
      and npt = ref 0
      and cp = ref 0
      and info = ref 0
      and inmc = ref 0
      and iycn = ref 0
      and iscn = ref 0
      and nfev = ref 0
      and finish = ref false
      and iprint = ref 0
      and w_flag = ref false
      and w = ref (Array1.create float64 c_layout 0) (* pointer to an array *)
		
      exception IFlagException of string
      exception Return
	
      let ddot n (dx:rbarray) ix0 incx (dy:rbarray) iy0 incy =
	let dtemp = ref 0.
	and ix = ref 0
	and iy = ref 0
	and m1 = ref 0
	and mp1 = ref 0;
	in
	  try 
	    if n <= 0 then raise Return;
	    
	    if not(incx = 1 && incy = 1) then (
	      ix := 1;
	      iy := 1;
	      if incx < 0 then 
		ix := (1 - n) * incx + 1;
	      if incy < 0 then
		iy := (1 - n) * incy + 1;
	      for i=1 to n do
		dtemp := !dtemp +. dx.{ix0 + !ix - 1} *. dy.{iy0 + !iy - 1};
		ix := !ix + incx;
		iy := !iy + incy;
	      done;
	      raise Return );
	    m1 := n mod 5;
	    if !m1 <> 0 then (
	      for i=1 to !m1 do
		dtemp := !dtemp +. dx.{ix0 + i - 1} *. dy.{iy0 + i - 1} done;
	      if n < 5 then
		raise Return );
	    mp1 := !m1 + 1;
	    let t1 = ref !mp1 in
	      while (!t1 <= n) do
		dtemp := !dtemp +. dx.{ix0 + !t1 - 1} *. dy.{iy0 + !t1 - 1}
		+. dx.{ix0 + !t1} *. dy.{iy0 + !t1}
		+. dx.{ix0 + !t1 + 1} *. dy.{iy0 + !t1 + 1}
		+. dx.{ix0 + !t1 + 2} *. dy.{iy0 + !t1 + 2}
		+. dx.{ix0 + !t1 + 3} *. dy.{iy0 + !t1 + 3};
		t1:= !t1 + 5
	      done; 
	      raise Return
	  with Return -> !dtemp
	    
      let lb1 iprint iter nfun gnorm n m x f g stp finish =
	if iter = 0 then (
	  print_string "*************************************************\n";
	  Printf.printf " n = %d number of corrections = %d\n\n     initial values" n m;
	  Printf.printf " f = %f    gnorm =  %f\n" f gnorm;
	  if iprint.(1) >= 1 then (
	    print_string " vector x =";
	    for i=1 to n do
	      Printf.printf "  %f" x.{i-1};
	    done;
	    print_string "\n";
	    print_string " gradient vector g =";
	    for i=1 to n do
	      Printf.printf "  %f" g.{i-1};
	    done;
	  );
	  print_string "*************************************************";
	  print_string "\n\ti\tnfn\tfunc\tgnorm\tsteplength";
	)
	else (
	  if not( (iprint.(0) = 0) && (iter <> 1 && not(finish))) then (
	    if (iprint.(0) <> 0) then (
	      if (((iter - 1) mod iprint.(0)) = 0) || finish then (
		if (iprint.(1) > 1 && iter > 1) then
		  print_string "\n\ti\tnfn\tfunc\tgnorm\tsteplength";
		print_string "\n";
		Printf.printf "\t%d\t%d\t%f\t%f\t%f" iter nfun f gnorm stp ) )
	    else (
	      print_string "\n";
	      if (iprint.(1) > 1 && finish) then
		print_string "\n\ti\tnfn\t\tfunc\t\tgnorm\tsteplength\n";
	      Printf.printf "\t%d\t%d\t%f\t%f\t%f" iter nfun f gnorm stp );
	    if (iprint.(1) = 2 || iprint.(1) = 3) then (
	      if (finish) then
		print_string " final point x = "
	      else
		print_string " vector x = ";
	      for i=1 to n do
		Printf.printf " %f" x.{i-1} done;
	      print_string "\n";
	      if iprint.(1) = 3 then (
		print_string " gradient vector g = ";
		for i=1 to n do
		  Printf.printf " %f" x.{i-1} done;
		print_string "\n" )
	    );
	    if (finish) then
	      print_string "\n\n The minimization terminated without detecting errors. iflag = 0";
	  );
	  flush stdout )
	  
      let daxpy n da (dx:rbarray) ix0 incx (dy:rbarray) iy0 incy =
	let ix = ref 0
	and iy = ref 0
	and m = ref 0
	and mp1 = ref 0
	in
	  try 
	    if n <= 0 then raise Return;
	    if da = 0. then raise Return;
	    if incx = 1 && incy =1 then
	      for i=0 to (n-1) do
		dy.{i + iy0} <- (dy.{i + iy0} +. da *. dx.{i + ix0}) done
	    else (
	      ix := 0;
	      iy := 0;
	      if (incx < 0) then ix := (1 - n) * incx;
	      if (incy < 0) then iy := (1 - n) * incy;
	      for i=0 to n-1 do
		dy.{iy0 + !iy} <- (dy.{iy0 + !iy} +. da *. dx.{ix0 + !ix});
		ix := !ix + incx;
		iy := !iy + incy
	      done;
	      raise Return );
	  with Return -> ()
	    
      let init () =
	w_flag := false; (* this should be set to false so that new workspace is created *)
	gnorm := 0.; stp1 := 0.; ftol := 0.; stp := 0.; ys := 0.; yy := 0.; sq := 0.; yr := 0.; beta := 0.;
	xnorm := 0.; iter := 0; nfun := 0; point := 0; ispt := 0; iypt := 0; maxfev := 0; bound := 0;
	npt := 0; cp := 0; info := 0; inmc := 0; iycn := 0; iscn := 0; nfev := 0; finish := false;
	iprint := 0; w := (Array1.create float64 c_layout 0);
	Mcsrch.init ()
	  
	  
	  
      (** This is the main function for lbfgs.  It should be recalled completely by
	application code with the current function value f and gradient vector g *)
      let lbfgs n m (x:rbarray) f (g:rbarray) diagco (diag:rbarray) iprint eps xtol iflag =
	let execute_entire_while_loop = ref false in
	  (* original code has a possibly emtpy w and fills it. We're assuming it is empty *)
          (* XXX - we're recreating this on each lbfgs iteration, clean up *)
	  (* let solution_cache = Array1.create n 0. in *)
	  if !w_flag = false then (
	    w := Array1.create float64 c_layout (n * (2 * m + 1) + 2 * m);
	    Array1.fill !w 0.;
	    w_flag := true );
	  try 
	    if !iflag = 0 then (
	      (* do something with solution_cache, figure out logic to work around *)
	      iter := 0;
	      if (n <= 0 || m <= 0) then (
		iflag := -3;
		raise (IFlagException "n or m less than or equal to 0") );
	      if (!l_gtol <= 0.0001) then (
		print_string "LBFGS.lbfgs: gtol is less than or equal to 0.0001. It has been reset to 0.9.";
		l_gtol := 0.99 );
	      nfun := 1;
	      point := 0;
	      finish := false;
	      if (!diagco) then
		for i=1 to n do 
		  if diag.{i-1} <= 0. then (
		    iflag := -2;
		    raise (IFlagException "Diagonal element less than 0") )
		done
	      else
		for i=1 to n do
		  diag.{i-1} <- 1. done;
	      ispt := n + 2 * m;
	      iypt := !ispt + n * m;
	      
	      for i=1 to n do
		(!w).{!ispt + i - 1} <- (-.(g.{i-1} *. diag.{i-1})) done;
	      
	      gnorm := sqrt ( ddot n g 0 1 g 0 1 );
	      stp1 := 1. /. !gnorm;
	      ftol := 0.0000001;
	      maxfev := 20;  (* 50 *)
	      if iprint.(0) >= 0 then (
		print_string "\n LB1";
		lb1 iprint !iter !nfun !gnorm n m x !f g !stp !finish );
	      execute_entire_while_loop := true;
	    );
	    while true do
	      if !execute_entire_while_loop then (
		iter := !iter + 1;
		info := 0;
		bound := !iter - 1;
		if (!iter <> 1) then (
		  if (!iter > m) then bound := m;
		  ys := ddot n !w (!iypt + !npt) 1 !w (!ispt + !npt) 1;
		  if not(!diagco) then (
		    yy := ddot n !w (!iypt + !npt) 1 !w (!iypt + !npt) 1;
		    for i=1 to n do
		      diag.{i-1} <- (!ys /. !yy) done )
		  else (
		    iflag := 2;
		    raise Return ) ) );
	      if (!execute_entire_while_loop || !iflag = 2) then (
		if (!iter <> 1) then (
		  if (!diagco) then 
		    for i=1 to n do
		      if diag.{i-1} <= 0. then (
			iflag := -2;
			raise (IFlagException "Diagonal element less than or equal to 0") )
		    done;
		  cp := !point;
		  if (!point = 0) then cp := m;
		  (!w).{n + !cp -1} <- (1. /. !ys);
		  for i=1 to n do
		    (!w).{i-1} <- (-.(g.{i-1})) done;
		  cp := !point;
		  for i=1 to !bound do
		    cp := !cp - 1;
		    if (!cp = (-1)) then cp := m - 1;
		    sq := ddot n !w (!ispt + !cp * n) 1 !w 0 1;
		    inmc := n + m + !cp + 1;
		    iycn := !iypt + !cp * n;
		    (!w).{!inmc - 1} <- (((!w).{n + !cp}) *. !sq);
		    daxpy n (-.(!w.{!inmc - 1})) !w !iycn 1 !w 0 1
		  done;
		  
		  for i=1 to n do
		    (!w).{i-1} <- (diag.{i-1} *. (!w).{i-1}) done;
		  
		  for i=1 to !bound do
		    yr := ddot n !w (!iypt + !cp * n) 1 !w 0 1;
		    beta := (!w).{n + !cp} *. !yr;
		    inmc := n + m + !cp + 1;
		    beta := (!w).{!inmc - 1} -. !beta;
		    iscn := !ispt + !cp * n;
		    daxpy n !beta !w !iscn 1 !w 0 1;
		    cp := !cp + 1;
		    if (!cp = m) then cp := 0
		  done;
		  
		  for i=1 to n do
		    (!w).{!ispt + !point * n + i - 1} <- (!w).{i-1} done;
		);
		nfev := 0;
		stp := 1.;
		if (!iter = 1) then stp := !stp1;
		
		for i=1 to n do
		  (!w).{i-1} <- g.{i-1} done
	      );
	      (* LINE SEARCH *)
	      Mcsrch.mcsrch n x f g !w (!ispt + !point * n) stp !ftol xtol !maxfev info nfev diag;
	      if !info = (-1) then (
		iflag := 1;
		raise Return );
	      if (!info <> 1) then (
		iflag := -1;
		raise (IFlagException (string_of_int !info)));
	      nfun := !nfun + !nfev;
	      npt := !point * n;
	      for i=1 to n do
		(!w).{!ispt + !npt + i - 1} <- (!stp *. (!w).{!ispt + !npt + i - 1});
		(* At.set !w (!ispt + !npt + i - 1) !stp; *)
		(!w).{!iypt + !npt + i - 1} <- (g.{i-1} -. (!w).{i-1})
	      done;
	      point := !point + 1;
	      
	      if (!point = m) then point := 0;
	      
	      gnorm := sqrt (ddot n g 0 1 g 0 1);
	      xnorm := sqrt (ddot n x 0 1 x 0 1);
	      xnorm := max 1.0 !xnorm;
	      
	      if (!gnorm /. !xnorm) <= eps then finish := true;
	      
	      if (iprint.(0) >= 0) then
		lb1 iprint !iter !nfun !gnorm n m x !f g !stp !finish;
	      
	      (* COPY x into solution_cache here or figure out alternative *)
	      if !finish then (
		iflag := 0;
		raise Return );
	      execute_entire_while_loop := true
	    done
	  with 
	      Return  -> ()
	    | (IFlagException x) -> Printf.printf "\n Exiting: %s" x
    end
      
    module type TestSig =
    sig
      val test : unit -> unit
    end
      
    module Test : TestSig =
    struct
      let ndim = 2000
      and msave = 7
      exception Return
	
      let test () =
	let x = Array1.create float64 c_layout ndim;
	and g = Array1.create float64 c_layout ndim;
	and diag = Array1.create float64 c_layout ndim;
	and n = 100
	and m = 5
	and iprint = Array.make 2 0
	and diagco = ref false
	and eps = 1.0e-2
	and xtol = 1.0e-16
	and icall = ref 0
	and iflag = ref 0
	and it1 = ref 1
	and f = ref 0.
	and t1 = ref 0.
	and t2 = ref 0.
	in
	  Array1.fill x 0.;
	  Array1.fill g 0.;
	  Array1.fill diag 0.;
	  iprint.(0) <- 1;
	  iprint.(1) <- 0;
	  while !it1 <= n do
	    x.{!it1 - 1} <- (-.(1.2e0));
	    x.{!it1} <- 1.e0;
	    it1 := !it1 + 2;
	  done;
	  try 
	    while true do
	      f := 0.;
	      let it2 = ref 1 in
		while !it2 <= n do
		  t1 := 1.e0 -. x.{!it2-1};
		  t2 := 1.e1 *. x.{!it2} -. x.{!it2-1} *. x.{!it2-1};
		  g.{!it2} <- (2.e1 *. !t2);
		  g.{!it2-1} <- (-.(2.e0)) *. x.{!it2-1} *. g.{!it2} +. !t1;
		  (*print_string ("\n -> "^ (string_of_float (!t1 *. !t1))); *)
		  f := !f +. (!t1 *. !t1) +. (!t2 *. !t2);
		  it2 := !it2 + 2
		done;
		Lbfgs.lbfgs n m x f g diagco diag iprint eps xtol iflag;
		if (!iflag = 0 || !icall > 400) then (
		  Printf.printf "\n... reached maximum number of iterations allowed" ; flush stdout;
		  raise Return );
		icall := !icall + 1;
	    done
	  with Return -> ()
	    
    end
  end
    
