(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(** This module contains the core Maximum Entropy training and decoding code.
    It makes use of the Lbfgs.Opt module for parameter estimation *)
open Bigarray
open Vectors
open Str

module MLO = Lbfgs.Opt

let gaussian_prior = ref 100.0

type 'a il = 'a Instance.raw_instance array

type 'a grad_args = { ilist: 'a il;
		      num_features: int;
		      constraints: Vectors.rbarray;
		      num_labels: int}

and  model = {parameters: Vectors.rbarray;
	      nfs: int;
	      nls: int}

(** This record represents the full user-level model - i.e. it includes the 
    mappings from feature names to ids *)
type user_model = { um_model: model;
		    um_falpha:  Instance.alphabet;
		    um_lalpha:  Instance.alphabet }

exception Return

let class_scores_unnormalized num_labels num_features parameters sparse_vector : float array =
  let scores = Array.make num_labels 0.0 in
    for i=0 to num_labels-1 do
      scores.(i) <- (Vectors.ds_dot_product (i * num_features) parameters sparse_vector);
    done; 
    scores

let class_scores_norm num_labels num_features parameters sparse_vector : float array =
  let sum = ref 0.0 in
  let scores = class_scores_unnormalized num_labels num_features parameters sparse_vector in
  let max = Array.fold_left (fun x y -> if x>y then x else y) neg_infinity scores in
    for i=0 to num_labels-1 do
      scores.(i) <- (exp (scores.(i) -. max));
      sum := !sum +. scores.(i);
    done;
    if !sum > 0.0 then
      for i=0 to num_labels-1 do
	scores.(i) <- (scores.(i) /. !sum);
      done
    else (
      print_string "\n Warning: scores 0.0\n";
      flush stdout;
      for i=0 to num_labels-1 do
	scores.(i) <- 0.0;
      done; );
    scores

let get_best_class model r_inst =
  flush stdout;
  let fv = (Instance.get_f_vector r_inst) in
  let scores = (class_scores_norm model.nls model.nfs 
		  model.parameters fv) in
  let bind = ref (-1) in
  let bscr = ref (-1.0) in
  for i=0 to (Array.length scores) - 1 do
    if scores.(i) > !bscr then (
      bind := i;
      bscr := scores.(i) ) done;
  (!bind,!bscr)

let get_distributions model r_ilist =
  Array.map 
    (fun x -> (class_scores_norm model.nls model.nfs model.parameters (Instance.get_f_vector x))) 
    r_ilist

(* label and score pairs *)
let classify_instances_raw model r_ilist =
  Array.map (get_best_class model) r_ilist

(* just get the best label *)
let classify_instances model r_ilist =
  Array.map (fun x -> match (get_best_class model x) with y,z -> y) r_ilist

let score_arrays a1 a2 =
  let s = ref 0. in
  let len = (Array.length a1) in
  for i=0 to len - 1 do
    if a1.(i) = a2.(i) then
      s := !s +. 1.; done;
  !s /. (float_of_int len)

let evaluate_instances_x (model: model) r_ilist =
  let gold = Array.map (Instance.get_label) r_ilist in
  let response = classify_instances model r_ilist in
  score_arrays gold response

let evaluate_instances (model: model) (iset: 'a Instance.instance_set) =
  evaluate_instances_x  model (Array.of_list iset.Instance.r_insts)

let conf_matrix gold response size =
  assert ((Array.length gold) = (Array.length response));
  let conf_mat = Array.make_matrix size size 0. in
  for i=0 to (Array.length gold) - 1 do
    let g = gold.(i) in
    let r = response.(i) in
      if g >= 0 && r >= 0 then
	conf_mat.(g).(r) <- conf_mat.(g).(r) +. 1. done;
  conf_mat

let get_confusion_matrix model r_ilist =
  let gold = Array.map (Instance.get_label) r_ilist
  and response = classify_instances model r_ilist in
    conf_matrix gold response model.nls

let sum_array (rarr: float array) = 
  let s = ref 0. in
  for i=0 to (Array.length rarr) - 1 do
    s := !s +. rarr.(i) done;
  !s

let compute_recalls conf_mat =
  let ln = (Array.length conf_mat) in
  let rs = Array.make ln 0. in
  for i=0 to ln - 1 do
    rs.(i) <- conf_mat.(i).(i) /. (sum_array conf_mat.(i)) done;
  rs
    
let compute_precisions conf_mat =
  let ln = (Array.length conf_mat) in
  let ps = Array.make ln 0. in
  for i=0 to ln - 1 do
    let den = ref 0. in
    for j=0 to ln - 1 do
      den := !den +. conf_mat.(j).(i) done;
    ps.(i) <- conf_mat.(i).(i) /. !den done;
  ps

let f_measure p r = 2. /. ((1. /. p) +. (1. /. r))

let create_results_list rs ps cf mapping =
  let rl = Array.to_list rs
  and pl = Array.to_list ps in
  let rec create_list rli pli i =
    match rli, pli with
      r :: rls, p :: pls -> 
	begin
	try
	  let lab = Hashtbl.find mapping i in
	  let size = int_of_float (sum_array cf.(i)) in
	    (lab, p, r, (f_measure p r), size) :: create_list rls pls (i + 1)
	with Not_found -> create_list rls pls (i + 1)
	end
    | [] , [] -> []
    | [] , _ :: _ -> []
    | _ :: _ , [] -> [] in
  create_list rl pl 0

let print_confusion_matrix cf mapping =
  let len = Array.length cf in
    Printf.printf "\n%10s" "";
    for i=0 to len - 1 do
      Printf.printf "%10s" (Hashtbl.find mapping i) done;
    Printf.printf "\n\n";
    for i=0 to len - 1 do
      Printf.printf "\n%10s" (Hashtbl.find mapping i);
      for j=0 to len - 1 do
	Printf.printf "%10.1f" cf.(i).(j)
      done
    done;
    Printf.printf "\n\n\n"

let print_confusion_matrix_tabs cf mapping =
  let len = Array.length cf in
    Printf.printf "\n\t";
    for i=0 to len - 1 do
      Printf.printf "\t%s" (Hashtbl.find mapping i) done;
    Printf.printf "\n\n";
    for i=0 to len - 1 do
      Printf.printf "\n\t%s" (Hashtbl.find mapping i);
      for j=0 to len - 1 do
	Printf.printf "\t%1.1f" cf.(i).(j)
      done
    done;
    Printf.printf "\n\n\n"

let get_precision_recall model r_ilist mapping =
  let cf = get_confusion_matrix model r_ilist in
  let rs = compute_recalls cf in
  let ps = compute_precisions cf in
  let results_list = 
    List.sort 
      (fun (_,_,_,_,x) (_,_,_,_,y) -> x - y) (* sort by size *)
      (create_results_list rs ps cf mapping) in
    results_list

let pretty_print_precision_recall ?(tabs=false) model r_ilist mapping =
  let results_list = get_precision_recall model r_ilist mapping in
  let cf = get_confusion_matrix model r_ilist in
    Printf.printf "\n\n\n============================ Results ==========================\n\n";
    Printf.printf "\n%40s\tPrecision\tRecall\t\tF-Measure\n" "";
    List.iter 
      (fun (lab, psi, rsi, fi, size) ->
	 Printf.printf "\n%40s\t%8f\t%8f\t%8f\t%8d" lab psi rsi fi size)
      results_list;
    Printf.printf "\n\n\n===============================================================\n\n";
    Printf.printf "\n\n\n\n";
    if tabs then
      print_confusion_matrix_tabs cf mapping
    else
      print_confusion_matrix cf mapping

let precision_recall_x pos_label (model: model) r_ilist =
  let gold = Array.map (Instance.get_label) r_ilist
  and response = classify_instances model r_ilist
  and misses = ref 0
  and errors = ref 0 
  and num_pos = ref 0 in
  Array.iteri 
    (fun i v -> 
      if v = pos_label then (
	if not(response.(i) = v) then misses := !misses + 1;
	num_pos := !num_pos + 1)
      else if (response.(i) = pos_label) then errors := !errors + 1) gold;
    (*
  Printf.printf "\n number correct: %d" ((Array.length gold) - (!misses + !errors));
  Printf.printf "\n number incorrect: %d" (!errors + !misses);
  Printf.printf "\n total number: %d" (Array.length gold); *)
  ((1. -. (float_of_int !errors) /. ((float_of_int (Array.length gold)) -. (float_of_int !num_pos))), 
   (1. -. (float_of_int !misses) /. (float_of_int !num_pos)))

let get_gradient  (parameters:Vectors.rbarray) (gradient:Vectors.rbarray) 
    (iarray:'a il) num_features (constraints:Vectors.rbarray) num_labels =
  let cost = ref 0.0 in
    Array1.fill gradient 0.0;
    for i = 0 to (Array.length iarray) - 1 do
      let sp = Instance.get_f_vector iarray.(i) in
      let lb = Instance.get_label iarray.(i) in (* the true label for this instance *)
      let scores = class_scores_norm num_labels num_features parameters sp in
	cost := !cost -. log(scores.(lb));
	for j=0 to num_labels-1 do
	  Vectors.plus_equals scores.(j) num_features j sp gradient;
	done
    done;
    for i=0 to (num_features * num_labels) - 1 do
      cost := !cost +. (parameters.{i} *. parameters.{i} /. (2. *. !gaussian_prior)) done;
    Vectors.times_equals (-.(1.0)) num_features num_labels gradient; (* point opposite direction *) 
    Vectors.plus_equals_m 1.0 num_features num_labels constraints gradient;
    Vectors.times_equals (-.(1.0)) num_features num_labels gradient; (* point opposite direction *) 
    Vectors.plus_equals_m (-.(1.0) /. !gaussian_prior) num_features num_labels parameters gradient;
    !cost

(* this creates the constraint array and initializes it *)
let compute_constraints (ilist:'a il) num_features num_labels =
  let constraints = Array1.create float64 c_layout (num_features * num_labels) in
  Array1.fill constraints 0.;
  Array.iter
    (fun x ->
      let sp = Instance.get_f_vector x in
      let lab = Instance.get_label x in
	Vectors.plus_equals 1.0 num_features lab sp constraints) ilist;
  constraints

let train_lbfgs_x ?(no_print=false) ?(max_iters=1000) ilist num_features num_labels =
  Printf.printf "\n Training with %d features and %d labels\n" num_features num_labels;
  let iprint = Array.make 2 0
  and params = Array1.create float64 c_layout (num_features * num_labels)
  and gradient = Array1.create float64 c_layout (num_features * num_labels)
  and diagco = ref false
  and diag = Array1.create float64 c_layout (num_features * num_labels)
  and eps = 1.0e-5
  and xtol = 4.9e-300
  and icall = ref 0
  and m = 7
  and iflag = ref 0 in
  let iter = ref 0 in
  let n = num_labels * num_features in
  let f = ref 0. in
  let constraints = compute_constraints ilist num_features num_labels in
    Array1.fill params 0.;
    Array1.fill gradient 0.;
    Array1.fill diag 0.;
    if no_print then 
      iprint.(0) <- 0
    else
      iprint.(0) <- 1;
    iprint.(1) <- (-1);
  f := get_gradient params gradient ilist num_features constraints num_labels;
  MLO.Lbfgs.init (); (* initialize lbfgs routine *)
  try
    while (!iter < max_iters) do
      MLO.Lbfgs.lbfgs n m params f gradient diagco diag iprint eps xtol iflag;
      if (!iflag <= 0) then
	raise Return;
      f := get_gradient params gradient ilist num_features constraints num_labels;
      iter := !iter + 1
    done;
    {parameters=params; nfs=num_features; nls=num_labels}
  with Return -> {parameters=params; nfs=num_features; nls=num_labels}

let train_lbfgs ?(no_print=false) (ts: 'a Instance.instance_set) =
  let ilist = (Array.of_list ts.Instance.r_insts)
  and num_features = ts.Instance.falpha.Instance.size + 1
  and num_labels = ts.Instance.lalpha.Instance.size in
    train_lbfgs_x ilist ~no_print:no_print num_features num_labels
    
