(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(** This is main program that will decode a set of provided instances. *)
open Mxutils


let get_model model_file = 
  let in_ch = open_in model_file in
  let (model:Maxent.user_model) = Marshal.from_channel in_ch in
  close_in in_ch;
  model

let write_predictions predictions gold iprotos file mapping =
  let o_ch = open_out file in
  Array.iteri 
    (fun i (el,sc) -> 
       Printf.fprintf o_ch "%40s\t%10f" (Hashtbl.find mapping el) sc;
       (try
	  if gold.(i)=el then 
	     Printf.fprintf o_ch "\tCorrect" 
	  else ( 
	     let v = Hashtbl.find mapping gold.(i) in
	       Printf.fprintf o_ch "\tIncorrect\t%40s" v);
	  Printf.fprintf o_ch "\t%s\n" iprotos.(i).GenMe.comment
	with Not_found -> Printf.fprintf o_ch "\tIncorrect%40s\n" "UNK" )
    )
    (Array.of_list (List.rev (Array.to_list predictions)));
  close_out o_ch

let main () =
  let input_model = ref ""
  and input_file = ref "" 
  and verbose = ref false
  and scale = ref false
  and output_file = ref "" 
  and distributions = ref false in
  Arg.parse 
    [("-model",Arg.Set_string(input_model)," Input model file");
     ("-input", Arg.Set_string(input_file), " Input data file");
     ("-verbose", Arg.Set(verbose), " Confusion matrix and per-class f-measure provided");
     ("-scale", Arg.Set(scale), " Scale values");
     ("-distributions", Arg.Set(distributions), " Output distributions over labels, not just the best one ");
     ("-output",Arg.Set_string(output_file), "Output file for predicted instances")]
    (fun x -> ())
    "Usage: mxtest -model <model file> -input <input data> -output <output predictions>";
    let umodel = get_model !input_model in
    let model = umodel.Maxent.um_model in
    let la = umodel.Maxent.um_lalpha and fa = umodel.Maxent.um_falpha in
    let nfs = model.Maxent.nfs in
    let nls = model.Maxent.nls in
    let iprotos = GenMe.protoseqs_from_vectors !input_file in
    let iset,_,_ = GenMe.get_inst_set ~runtime:true ~fa:(Some fa) ~la:(Some la) iprotos in 
    let ilist = Array.of_list iset.Instance.r_insts in
    let gold = Array.map (Instance.get_label) ilist in
    let predicted = Maxent.classify_instances_raw model ilist in
    let score:float = Maxent.score_arrays gold (Array.map (fun (x,_) -> x) predicted) in
    let ht = (Instance.invert_alphabet la).Instance.inv_map in
    let gold_rev = Array.of_list (List.rev (Array.to_list gold)) in
      Printf.printf "\n Accuracy on test data: %f\n" score;
      Maxent.pretty_print_precision_recall ~tabs:true model ilist ht;
      if !distributions then 
	let label_dist = Maxent.get_distributions model ilist in
	let och = open_out !output_file in
	  for i = 0 to nls - 1 do
	    Printf.fprintf och "%s\t" (Hashtbl.find ht i); done;
	  Printf.fprintf och "\n";
	  Array.iter (fun ar -> Array.iter (Printf.fprintf och "%f\t") ar; Printf.fprintf och "\n") label_dist;
	  close_out och
      else
	write_predictions predicted gold_rev (Array.of_list iprotos) (!output_file) ht;
      Printf.printf "\n Execution time (CPU): %f seconds\n\n" (Sys.time ());;

main ();;
