(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(** This is a main program that will train a maximum entropy file. *)

open Bigarray
open Mxutils

module Int = struct type t = int let compare = compare end
module ISet = Set.Make(Int)

let write_parameters file model =
  let o_ch = open_out file in
  let params = model.Maxent.parameters in
  let ln = Array1.dim params in
    for i=0 to ln-1 do
      Printf.fprintf o_ch "%d: %f\n" i params.{i} done;
    close_out o_ch;;

let print_model file model = 
  let o_ch = open_out file in
    Printf.fprintf o_ch "%d\n" model.Maxent.nfs;
    Printf.fprintf o_ch "%d\n" model.Maxent.nls;
    for i=0 to (Array1.dim model.Maxent.parameters) - 1 do
      Printf.fprintf o_ch "%f\n" model.Maxent.parameters.{i} done;
    close_out o_ch;;

let marshal_model file model =
  let o_ch = open_out file in
    Marshal.to_channel o_ch model [];
    close_out o_ch;;

let permute_array arr =
  for i = (Array.length arr) - 1 downto 1 do
    let j = Random.int i in
    let t = arr.(i) in
      arr.(i) <- arr.(j);
      arr.(j) <- t done

let split_array arr ratio =
  let s = (Array.length arr) in
  let m = int_of_float ((float_of_int s) *. ratio) in
  let a1 = Array.init m (fun i -> arr.(i)) in
  let a2 = Array.init (s-m) (fun i -> arr.(i+m)) in
    (a1,a2)

let avg ar = (Array.fold_right (+.) ar 0.) /. (float_of_int (Array.length ar))
let fmeas p r = (2. *. p *. r) /. (p +. r)

let xvalidate_x ilist nfs nls invla ratio numTrials =
  let all_scores = Array.init numTrials (fun i -> 0.0) in
  let all_recalls = Array.make_matrix nls numTrials 0. in
  let all_precs = Array.make_matrix nls numTrials 0. in
  let all_fmes = Array.make_matrix nls numTrials 0. in
    for i = 0 to numTrials - 1 do
      permute_array ilist;
      let tr,tst = split_array ilist ratio in
      let model = Maxent.train_lbfgs_x tr nfs nls in
	Maxent.pretty_print_precision_recall model tst invla.Instance.inv_map;
	for j = 0 to nls - 1 do
	  let p,r = Maxent.precision_recall_x j model tst in
	    all_recalls.(j).(i) <- if r = r then r else 0.0;
	    all_precs.(j).(i) <- p;
	    let fm = fmeas p r in
	      all_fmes.(j).(i) <- if fm = fm then fm else 0.0
	done;
	all_scores.(i) <- Maxent.evaluate_instances_x model tst 
    done;
    print_endline "Finished runs"; flush stdout;
    let tot = Array.fold_right (+.) all_scores 0. in
    let tot' = Array.fold_right (fun a t -> t +. (a *. a)) all_scores 0. in
    let mean = tot /. (float_of_int numTrials) in
    let var = (tot' /. (float_of_int numTrials)) -. (mean ** 2.0) in
      Printf.printf "\nExpected accuracy, %f , variance ( %f ) \n\n" mean var; flush stdout;
      for i = 0 to nls - 1 do
	let avg_prec = avg all_precs.(i)
	and avg_rec = avg all_recalls.(i)
	and avg_fmes = avg all_fmes.(i) in
	  Printf.printf "For class: %s\n" (Hashtbl.find invla.Instance.inv_map i);
	  Printf.printf "\t Precision: %f\n" avg_prec;
	  Printf.printf "\t Recall: %f\n" avg_rec;
	  Printf.printf "\t F-measure: %f\n" avg_fmes
      done

let xvalidate ?(min_cnt=1) maxits protos ratio numtrials =
  let parray = Array.of_list protos in
  let all_scores = Array.init numtrials (fun i -> 0.0) in
    for i = 0 to numtrials - 1 do
      permute_array parray;
      let tr,tst = split_array parray ratio in
      let tr_iset,_,la = GenMe.get_inst_set (Array.to_list tr) in
      let a_tr_iset = if min_cnt > 1 then Instance.commonify_instances tr_iset min_cnt else tr_iset in
      let a_tr_ilist = Array.of_list a_tr_iset.Instance.r_insts in
      let fa = a_tr_iset.Instance.falpha in
      let tst_iset,_,_ = GenMe.get_inst_set ~runtime:true ~fa:(Some fa) ~la:(Some la) (Array.to_list tst) in
      let a_tst_iset = if min_cnt > 1 then Instance.commonify_instances tst_iset min_cnt else tst_iset in
      let a_tst_ilist = Array.of_list a_tst_iset.Instance.r_insts in
      let nfs = fa.Instance.size + 1 and nls = la.Instance.size in
      let model = Maxent.train_lbfgs_x ~max_iters:maxits a_tr_ilist nfs nls in
      let invla = Instance.invert_alphabet la in
      let score = Maxent.evaluate_instances_x model a_tst_ilist in
	Maxent.pretty_print_precision_recall ~tabs:true model a_tst_ilist invla.Instance.inv_map;
	all_scores.(i) <- score;
    done;
    print_endline "Finished runs"; flush stdout;
    let tot = Array.fold_right (+.) all_scores 0. in
    let tot' = Array.fold_right (fun a t -> t +. (a *. a)) all_scores 0. in
    let mean = tot /. (float_of_int numtrials) in
    let var = (tot' /. (float_of_int numtrials)) -. (mean ** 2.0) in
      Array.iter (fun x -> Printf.printf " (%f)" x) all_scores;
      Printf.printf "\nExpected accuracy, %f , variance ( %f ) \n\n" mean var; flush stdout

(* change this so that it will either train or test by argument flag *)
let main () =
  let input_file = ref ""
  and output_file = ref ""
  and scale = ref false
  and verbose = ref false
  and max_iters = ref 1000
  and print_model_file = ref "" 
  and min_cnt = ref 1
  and xval = ref (-1) in
  Arg.parse 
    [("-input",Arg.Set_string(input_file)," Input file in sparse-vector format");
     ("-prior-variance",Arg.Float(fun x -> Maxent.gaussian_prior := x), "Set Gaussian Prior Variance (default 100.0)");
     ("-print-model-file", Arg.Set_string(print_model_file), "Print model in standard format");
     ("-scale", Arg.Set(scale), " Auto-scale feature values");
     ("-xval", Arg.Set_int(xval), " Perform n-fold X-validation");
     ("-max-iters", Arg.Set_int(max_iters), " Maximum number of iterations (default 1000)");
     ("-verbose",Arg.Set(verbose), " Verbose printing during training");
     ("-min-cnt",Arg.Set_int(min_cnt), " Minimum count required to include feature");
     ("-output",Arg.Set_string(output_file), "Output file for learned parameters")]
    (fun x -> ())
    "Usage: mxtrain -input <input file> -output <output file>";
  (* let (ilist, _, nfs, nls, scale_table) = get_instances_from_file ~norm:(!scale) None !input_file in *)
    let protos = GenMe.protoseqs_from_vectors !input_file in
    if !xval < 2 then (
      let iset,_,la = GenMe.get_inst_set ~runtime:false protos in
      let inv_la = Instance.invert_alphabet la in
      let a_iset = if !min_cnt > 1 then Instance.commonify_instances iset !min_cnt else iset in
      let a_ilist = Array.of_list a_iset.Instance.r_insts in
      let fa = a_iset.Instance.falpha in
      let nfs = fa.Instance.size + 1 and nls = la.Instance.size in
	Printf.printf "\nTraining with %d features and %d labels\n" nfs nls; flush stdout;
	let model = Maxent.train_lbfgs_x ~no_print:(not (!verbose)) ~max_iters:(!max_iters) a_ilist nfs nls in
	let score = Maxent.evaluate_instances_x model a_ilist in
	  Printf.printf "\n Accuracy on training data: %f\n" score;
	  if model.Maxent.nls = 2 then (
	    let (prec,recall) = Maxent.precision_recall_x 1 model a_ilist in
	      Printf.printf "\n Precision: %f Recall: %f" prec recall);
	  marshal_model (!output_file) {Maxent.um_model=model;Maxent.um_falpha=fa;Maxent.um_lalpha=la};
	  if not (!print_model_file = "")  then print_model !print_model_file model;
	  Printf.printf "\n Execution time (CPU): %f seconds\n\n" (Sys.time ()))
    else (
      Printf.printf "Performing random-resample xvalidation\n"; flush stdout;
      xvalidate ~min_cnt:!min_cnt !max_iters protos (1. -. (1. /. (float_of_int !xval))) !xval);;

main ();;
  

