(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(** This module is a small set of utilities used for developing application front-ends to the
    underlying Maximum Entropy code **)

open Bigarray

let split_reg = Str.regexp "[ \t]+"
let comment_reg = Str.regexp "^#"


let rec get_features_and_comments line acc =
  match line with
      a :: r as all ->
	if Str.string_match comment_reg a 0 then
	  (List.rev acc),(String.concat " " all)
	else get_features_and_comments r (a :: acc)
    | [] -> (List.rev acc), ""

let make_inst ?(norm=false) ?(fixed=false) max_fvals feature_line max_feature max_label num =
  let label = int_of_string (List.hd feature_line) in
  let features,comment = get_features_and_comments (List.tl feature_line) [] in
  let nfeatures = List.length features in
  let a_features = Array.of_list features in
  let inds = Array.init nfeatures (fun i -> 0) in
  let vals = Array.init nfeatures (fun i -> 0.) in
    if (label > !max_label) && (not fixed) then max_label := label;
    if (label <= !max_label) then 
      (      
	Array.iteri 
	       (fun i fea ->
		  let pair = Str.split (Str.regexp "[\\:]+") fea in
		  let ind = (int_of_string (List.hd pair)) in
		  let vl = float_of_string (List.nth pair 1) in
		    if (not fixed) then (
		      if (ind >= (!max_feature)) then 
			max_feature := ind;
		      inds.(i) <- ind;
		      vals.(i) <- vl)
		    else
		      if (ind < !max_feature) then (
			inds.(i) <- ind;
			vals.(i) <- vl);
		    if norm then (
		      let ab_val = abs_float vl in
		      let cv = try Hashtbl.find max_fvals ind with Not_found -> 0.0 in
			if (not fixed) && (ab_val > cv) then
			  Hashtbl.replace max_fvals ind ab_val)) a_features;
	(Instance.create_raw label (Vectors.create inds vals) 0), comment)
    else failwith "Invalid Instance"

let string_to_instance s =
  let max_feature = ref 0
  and max_label = ref 0 in
    fst (make_inst (Hashtbl.create 10) (Str.split split_reg s) max_feature max_label 0)

let string_array_to_instances ?(maxf=0) ?(maxl=0) sarray =
  let fixed = maxf > 0 && maxl > 0 in
  let max_feature = ref maxf
  and max_label = ref maxl in
  let insts = 
    Array.map 
      (fun s -> 
	 fst 
	 (make_inst ~fixed:fixed (Hashtbl.create 10) (Str.split split_reg s) max_feature max_label 0)) 
      sarray in
    (insts, !max_feature + 1, !max_label+1)

let get_instances_from_file ?(norm=false) ?(fixed=false) ?(nfs=0) ?(nls=0) max_fvals filename =
  let a_max_fvals = match max_fvals with None -> Hashtbl.create 10 | Some h -> h in
  let channel = open_in filename in
  let max_feature = ref (nfs-1) in
  let max_label = ref (nls-1) in
  let rec read_file num = 
    try 
      let l = Str.split (Str.regexp "[ \t]+") (input_line channel) in
	match l with 
	    a :: b ->
	      let inst = make_inst ~norm:norm ~fixed:fixed a_max_fvals  l
			   max_feature max_label num in
		inst :: (read_file (num+1))
	  | _ -> read_file num
    with End_of_file ->  close_in channel ; [] in
  let insts,comments = List.split (read_file 0) in
  let a_insts = Array.of_list insts in
  let a_comments = Array.of_list comments in
    if norm then
      Array.iter (fun inst -> 
		    let sp = Instance.get_f_vector inst in
		    let inds = Vectors.get_indices sp 
		    and vls = Vectors.get_values sp in
		      Array.iteri
			(fun i ind ->
			   vls.(i) <- vls.(i) /. 
			   (try Hashtbl.find a_max_fvals ind with Not_found  
				-> 1.0))
			inds) a_insts;
    (a_insts, a_comments, !max_feature+1, !max_label+1, a_max_fvals);;
