(*
  Copyright The MITRE Corporation 2005.  All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(**
  This module contains various data types and functions for 
  dealing with sparse and dense arrays and matrices
*)

open Bigarray

type rbarray = (float, Bigarray.float64_elt, Bigarray.c_layout) Bigarray.Array1.t
type sparse_vector = {size: int;
		      indices: int array;
		      values: float array}
type dense_vector = rbarray

let get_indices sp = sp.indices
let get_values sp = sp.values

let print_sp_vec sp_vec =
  print_string "\n indices: ";
  for i=0 to (Array.length sp_vec.indices) - 1 do
    Printf.printf "%d " sp_vec.indices.(i) done;
  print_string "\n values: ";
  for i=0 to (Array.length sp_vec.values) - 1 do
    Printf.printf "%f " sp_vec.values.(i) done;
  print_string "\n"

    
    (* This adds the first matrix m1 into the second m2 *)
let plus_equals_m factor num_columns num_rows (m1:dense_vector) (m2:dense_vector) =
  for i=0 to (num_columns*num_rows) - 1 do
    m2.{i} <- ((m1.{i} *. factor) +. m2.{i})
  done

    (* This adds the sparse vector sp_vec into the second d_vec *)
let plus_equals factor num_entries index (sp_vec:sparse_vector) (d_vec:dense_vector) =
  let shift = num_entries * index
  in
  for i=0 to sp_vec.size - 1 do
    let ind = shift + (sp_vec.indices.(i)) in 
    d_vec.{ind} <- ((d_vec.{ind}) +. (sp_vec.values.(i) *. factor));
  done

let times_equals factor num_entries num_rows (d_vec: dense_vector) =
  for i=0 to (num_entries * num_rows) - 1 do
    d_vec.{i} <- (d_vec.{i} *. factor)
  done

(* This will dot product two SPARSE vectors *)
(* The entries must be ordered - i.e. the elements of the indices 
   array must be strictly ascending *)
let sp_dot_product v1 v2 =
  let s = ref 0.0
  and j = ref 0
  in
  if v1.size < v2.size then 
    for i = 0 to v1.size - 1 do
      let m = v1.indices.(i) in begin
	while !j < v2.size && v2.indices.(!j) < m do
	  j := !j + 1
	done;
	if (v2.indices.(!j)) = m then
	  s := !s +. v1.values.(i) *. v2.values.(!j)
      end
    done
  else
    for i = 0 to v2.size - 1 do
      let m = v2.indices.(i) in begin
	while !j < v1.size && v1.indices.(!j) < m do
	  j := !j + 1
	done;
	if v1.indices.(!j) = m then
	  s := !s +. v2.values.(i) *. v1.values.(!j);
      end
    done;
  !s

let l_create indices values = 
  let s = List.length indices in
    {size=s; indices=(Array.of_list indices); values=(Array.of_list values)}
    
let create indices values =
  let s = Array.length indices in
    {size=s; indices=indices; values=values}


let ds_dot_product offset (d1:dense_vector) (v2:sparse_vector) =
  try 
    let s = ref 0.0
    in
    for i = 0 to v2.size - 1 do
      s := !s +. v2.values.(i) *. d1.{v2.indices.(i) + offset}
    done;
    !s
  with (Invalid_argument x) ->
    Printf.printf "\n Error with offset: %d" offset;
    Printf.printf "sparse vector:";
    print_sp_vec v2;
    Printf.printf "\n dense vector length: %d" (Array1.dim d1);
    raise (Invalid_argument x)
